
import java.awt.*;
import java.awt.geom.*;
import java.text.*;
import java.util.*;
import javax.swing.*;


/**
 * An amino acid (letter) above a colored triangle with point in (x, y)
 */
class AAMark extends Mark {
  final static private int[] rxs = {  -6,   4, 0 };
  final static private int[] lxs = {  -4,   6, 0 };
  final static private int[] ys =  { -10, -10, 0 };
  final static private GeneralPath ltriangle = new GeneralPath();
  final static private GeneralPath rtriangle = new GeneralPath();
  final static int radius = 6;
  final static DecimalFormat fmt3 = new DecimalFormat("0.000");
  boolean left; //Is this to the left of the start of this sequence?
  boolean best;
  AminoAcid acid;

  static {
    ltriangle.moveTo(lxs[0],ys[0]);
    rtriangle.moveTo(rxs[0],ys[0]);
    for (int i=1; i<lxs.length; i++) {
      ltriangle.lineTo(lxs[i],ys[i]);
      rtriangle.lineTo(rxs[i],ys[i]);
    }
    ltriangle.closePath();
    rtriangle.closePath();
  }
  
  public AAMark(AminoAcid acid, int kind, double x, double y, Peak peak, 
		AbstractSequence seq, boolean left) {
    this.acid = acid; 
    this.x = x; this.y = y; this.peak = peak;
    this.seq = seq; this.left = left;
    this.kind = kind; 
  } 

  //A new interpretation of mark
  public AAMark(AAMark aamark, int kind) {
    this(aamark.acid, kind, aamark.x, aamark.y, aamark.peak,
	 aamark.seq, aamark.left);
    this.score = aamark.score;
  } 

  //"Same" mark connected to different peak
  public AAMark(Peak peak, AAMark aamark) {
    this(aamark.acid, aamark.kind, aamark.x, peak.y, peak,
	 aamark.seq, aamark.left);
  } 

  public void setBest(boolean best) {
    this.best = best;
  }

  void drawInner(Graphics2D g, AbstractCoordinates coor, 
		 double tolerance, boolean showErrors) {
    AffineTransform t = AffineTransform.getTranslateInstance(coor.fromX(x),
							     coor.fromY(y));
    GeneralPath triangle = left ? ltriangle : rtriangle;
    //The colors should be variables/constants defined centrally
    g.setColor(kind == Mark.ACCEPTED ? Color.green : Color.blue);
    g.fill(triangle.createTransformedShape(t));
    
    int gwidth = g.getFontMetrics().stringWidth(acid.abbr);
    g.drawString(acid.abbr, 
		 coor.fromX(x) - gwidth/2 + (left?2:-2), 
		 coor.fromY(y)-12);
    int yoffset = (left?1:-1);
    if (kind == Mark.ACCEPTED) {
      //Selected, draw a line to show precisely which peak is the one
      g.draw(new Line2D.Float(coor.fromX(x), coor.fromY(y)+yoffset,
			      coor.fromX(peak.x), coor.fromY(y)+yoffset));
      if (showErrors) {
	String ellol = fmt3.format(error());
	gwidth =  g.getFontMetrics().stringWidth(ellol);
	g.drawString(ellol,coor.fromX(x) - gwidth/2 + (left?2:-2), 
		     coor.fromY(y)-24);
      }
    }
    else { 
      //Potential: draw a line to show the valid range of masses for
      //this amino acid
      g.draw(new Line2D.Float(coor.fromX(x-tolerance),
			      coor.fromY(y)+yoffset,
			      coor.fromX(x+tolerance),
			      coor.fromY(y)+yoffset));

    }
    if (best) {
      g.setColor(Color.red);
      g.draw(new Ellipse2D.Float(coor.fromX(x)-radius, coor.fromY(y)-radius, 
				 2*radius, 2*radius)); 
    }
  }

  public String getAbbr() { 
    return acid.abbr;
  }

  public void remove() {
    if (kind != Mark.ACCEPTED)
      throw new IllegalArgumentException("Remove unaccepted mark");
    LinkedList list = seq.getAccepted(left);
    //This should be unnecessary:
    if (!list.contains(this))
      throw new IllegalArgumentException("Mark not in sequence");
    Mark m;
    do {
      m = (Mark)list.removeFirst();
    } while (m != this);
  }

  public void accept() {
    if (kind != Mark.POTENTIAL)
      throw new IllegalArgumentException("Accept unpotential mark");
    LinkedList list = seq.getAccepted(left);
    list.addFirst(new AAMark(this, Mark.ACCEPTED));
  }

}
