
import java.text.*;
import java.util.*;


/**
 * AminoAcid
 *
 * Data about a single amino acid
 */
class AminoAcid {
  final public String name;	// Full name               
  final public String trigram;	// Three-letter name       
  final public String abbr;	// One-letter name (possibly starred)
  final public double monoW;	// Monoisotopic mass (in amu = Dalton) 

  final public ChemicalFormula formula;

  public AminoAcid(String name, String trigram, String abbr, double monoW,
		   String formula) {
    this.name = name;
    this.trigram = trigram;
    this.abbr = abbr;
    this.monoW = monoW;
    this.formula = new ChemicalFormula(formula);
  }
  
  public String toString() {
    return abbr;
  }

/**
 *  Data about all amino acids
 */

  private static String[] names;	// Full name
  private static String[] trigrams;	// Three-letter name
  private static String[] abbrs;	// One-letter name (possibly starred)
  private static double[] monoWs;	// Monoisotopic mass (amu)
  private static String[] formulas;     // Chemical Formula

  // All amino acids, also sorted by name, trigram, abbr, monoW

  static private AminoAcid[] aa;		
  public static final SortedSet /* of AminoAcid */ aaByName;
  public static final SortedSet /* of AminoAcid */ aaByTrigram;
  public static final SortedSet /* of AminoAcid */ aaByAbbr;
  public static final SortedSet /* of AminoAcid */ aaByMonoW;

  // Maps for fast retrieval of amino acids

  private static HashMap /* from String to AminoAcid */ nameTable;
  private static HashMap /* from String to AminoAcid */ trigramTable;
  private static HashMap /* from String to AminoAcid */ abbrTable;

  static {
    names = new String[] {
      "Glycine", "Alanine", "Serine", "Proline", "Valine",
      "Threonine", "Cysteine", /*"Isoleucine",*/ "Leucine", "Asparagine", 
      "Aspartic acid", "Glutamine", "Lysine", "Glutamic acid", "Methionine", 
      "Histidine", "Methionine ox.", "Phenylalanine", "Arginine", 
        "Carbamidomethylcysteine",
      "Tyrosine", "Cysteine **", "Tryptophan", "Tryptophan ox."
    };
    
    trigrams = new String[] {
      "Gly", "Ala", "Ser", "Pro", "Val", 
      "Thr", "Cys", /*"Ile", */"Leu", "Asn", 
      "Asp", "Gln", "Lys", "Glu", "Met", 
      "His", "Mox", "Phe", "Arg", "Cys*",
      "Tyr", "Cys**", "Trp", "Wox"
    };
    
    abbrs = new String[] {
      "G", "A", "S", "P", "V", 
      "T", "C", /*"I", */"L", "N", 
      "D", "Q", "K", "E", "M", 
      "H", "M*", "F", "R", "C*", 
      "Y", "C**", "W", "W*"
    };

    monoWs = new double[] {
       57.02147,  71.03712,  87.03203,  97.05277,  99.06842, 
       101.04768, 103.00919, /*113.08407, */113.08407, 114.04293,
      115.02695, 128.05858, 128.09497, 129.04260, 131.04049,
      137.05891, 147.03581, 147.06842, 156.10112, 160.03065,
      163.06333, 174.04630, 186.07932, 218.06914
    };

    //NBNBNB the formula for M* does not add
    //up to the right monoW within round-off errors!!!!
    formulas = new String[] {
      "C2H3NO", "C3H5NO", "C3H5NO2", "C5H7NO", "C5H9NO", 
      "C4H7NO2", "C3H5NOS", /*"C6H11NO",*/ "C6H11NO", "C4H6N2O2", 
      "C4H5NO3", "C5H8N2O2", "C6H12N2O", "C5H7NO3", "C5H9NOS", 
      "C6H7N3O", "C5H9NOS+O", "C9H9NO", "C6H12N4O", "C3H5NOS+C2H3NO", 
      "C9H9NO2", "C3H5NOS+C3H5NO", "C11H10N2O", "C11H10N2O+O2"
    };

    aa = new AminoAcid[names.length];

    nameTable = new HashMap();
    trigramTable = new HashMap();
    abbrTable = new HashMap();

    for (int i=0; i<names.length; i++) {
      aa[i] = new AminoAcid(names[i], trigrams[i], abbrs[i], monoWs[i], 
			    formulas[i]);    
      nameTable.put(names[i], aa[i]);
      trigramTable.put(trigrams[i], aa[i]);
      abbrTable.put(abbrs[i], aa[i]);
    }     
    
    List /* of AminoAcid */ aaList = Arrays.asList(aa);

    TreeSet tmp;

    tmp = 
      new TreeSet(new Comparator() {
	  public int compare(Object o1, Object o2) 
	  { return ((AminoAcid)o1).name.compareTo(((AminoAcid)o2).name); }
	});
    tmp.addAll(aaList);
    aaByName = Collections.unmodifiableSortedSet(tmp);
    
    tmp= 
      new TreeSet(new Comparator() {
	  public int compare(Object o1, Object o2) 
	  { return ((AminoAcid)o1).trigram.compareTo(((AminoAcid)o2).trigram);}
	});
    tmp.addAll(aaList);
    aaByTrigram = Collections.unmodifiableSortedSet(tmp);    

    tmp = 
      new TreeSet(new Comparator() {
	  public int compare(Object o1, Object o2) 
	  { return ((AminoAcid)o1).abbr.compareTo(((AminoAcid)o2).abbr); }
	});
    tmp.addAll(aaList);
    aaByAbbr = Collections.unmodifiableSortedSet(tmp);

    tmp = 
      new TreeSet(new Comparator() {
	  public int compare(Object o1, Object o2) {
	    double w1 = ((AminoAcid)o1).monoW, w2 = ((AminoAcid)o2).monoW;
	    if (w1 < w2) return -1;
	    else if (w1 > w2) return +1;
	    else return 0; 
	  }
	});
    tmp.addAll(aaList);
    aaByMonoW = Collections.unmodifiableSortedSet(tmp);

  } 

  public static final AminoAcid getAA(char abbr) {
    return getAA(String.valueOf(abbr));
  }

  public static final AminoAcid getAA(String s) {
    int len = s.length();
    if (len == 1 || len > 1 && s.charAt(len-1) == '*')
      return (AminoAcid)abbrTable.get(s);
    else if (s.length() == 3) 
      return (AminoAcid)trigramTable.get(s);
    else
      return (AminoAcid)nameTable.get(s);
  }

  //Format a String with (single) amino acids matching a given mass
  //within tolerance. This is for a specific use in
  //Controler.mousePosition(). 
  public static String getMatchesFormatted(double w, double tolerance) {
    w = Math.abs(w);
    StringBuffer res = new StringBuffer();
    boolean foundSome = false;
    //Perhaps we should use a precomputed TreeMap(monoW->AminoAcid)
    //together with TreeMap.subMap(). Even better would be
    //TreeMap(w->result) with keys at all values monoWs[a]+-tolerance,
    //but in that case the TreeMap would have to be recomputed for
    //each new tolerance (a rare occasion). 
    for (int a=0; a<monoWs.length; a++)
      if (Math.abs(w-monoWs[a])<tolerance) {
	if (foundSome)
	  res.append(",");
	else
	  res.append(" (");
	res.append(abbrs[a]);
	foundSome = true;
      }
    if (foundSome)
      res.append(")");
    return res.toString();
  }

  //Find sequences of amino acids with a total mass matching w within
  //tolerance.  At w=400 there are 1091 different sequences. At w=500
  //the number is 19047. (The count is wildly oscillating though).
  public static Collection getMatches(double w, double tolerance) { 
    LinkedList res = new LinkedList();
    //getMatches(res, new int[monoWs.length], 0, w, tolerance);
    getMatches(res, new LinkedList(), w, tolerance);
    return res;
  }
  //The private getMatches() recursively calls itself for each amino
  //acid added to the potential list and subtracted from the taget mass w.
  private static void getMatches(Collection out, 
				  LinkedList acidlist,
				  double w, double tolerance) {
    for (int a=0; a<monoWs.length; a++) {
      if (monoWs[a] > w+tolerance)
	return;
      LinkedList newlist = new LinkedList(acidlist);
      newlist.add(aa[a]);
      if (monoWs[a] >= w-tolerance) 
	out.add(newlist);
      else 
	getMatches(out, newlist, w-monoWs[a], tolerance);
    } 
  }
  public static void main(String[] args) {
    DecimalFormat fmt6 = new DecimalFormat("0.000000");
    //Check all amino acids:
    Iterator aaiter = aaByMonoW.iterator();
    while (aaiter.hasNext()) {
      AminoAcid acid = (AminoAcid)(aaiter.next());
      System.out.println(acid.abbr+" "+
			 fmt6.format(acid.monoW)+" "
	 		 +fmt6.format(acid.formula.monoW)+" "
		 	 +fmt6.format(acid.monoW-acid.formula.monoW));
    } 


    if (args.length > 0) {
      double w = Double.parseDouble(args[0]);
      Collection coll = getMatches(w, 0.5);
      System.out.println(w+" "+coll.size());
      //System.out.println(coll);
      Iterator iter = coll.iterator();
      while (iter.hasNext()) {
	Iterator aiter = ((LinkedList)(iter.next())).iterator();
	//Compute and print the mass of each match. It turns out that
	//(at least for w<600) almost all masses fall within an
	//interval of length 0.1 although we collect all peptide
	//masses inside an interval of length 1.
	//
	//We also compute the number of C atoms in order to check
	//correlation of mass and C-count.
	double m = 0.0;
	ChemicalFormula f = new ChemicalFormula("");
	while (aiter.hasNext()) {
	  AminoAcid a = (AminoAcid)(aiter.next());
	  m += a.monoW;
	  f = f.add(a.formula);
	}
	System.out.println(m+" "+(f.getCount(6)/w));
	
      }
    }
  }
}








