
import java.text.*;
import java.util.*;

/**
 *  ChemicalFormula computes masses of molecules given by a molecular
 *  formula.
 */

class ChemicalFormula {
  //Note: this implementation is perhaps too closely coupled to the
  //implementation of Element
  private int[] cofz =  new int[Element.Z_LIMIT]; // count by atomic number
  //this is a compressed  representation of the same data: 
  private int[] count;  // cofz[usedZ[i]] == count[i] when i<count.length
  private int[] usedZ;  // usedZ[] is the Z for which cofz != 0
  //
  public final double monoW;
  //
  private static int[] s2c(String formula) {
    int[] res =  new int[Element.Z_LIMIT];
    int i=0; 
    int sign = 1;
    while (i<formula.length()) {
      //First read past the sign if any
      if (formula.charAt(i)=='+' || formula.charAt(i) == '-') { 
	sign = formula.charAt(i) == '+' ? 1 : -1;
	i++;
      }
      //
      int j = i+1;
      while (j<formula.length() && Character.isLowerCase(formula.charAt(j)))
	j++;
      int Z = Element.get(formula.substring(i,j)).Z;
      i = j;
      while (j<formula.length() && Character.isDigit(formula.charAt(j))) 
	j++;
      res[Z] += sign*(i==j ? 1 : Integer.parseInt(formula.substring(i,j)));
      i=j;
    }
    return res;
  }
  ChemicalFormula(String formula) {
    this(s2c(formula));
  }
  ChemicalFormula(int[] c) { //c[Z] is the # of Z atoms in formula
    int numused = 0;
    double mass = 0.0;
    for (int z=1; z<c.length; z++) 
      if (c[z] != 0) {
	numused++;
	mass += c[z] * Element.elements[z].monoW;
      }
    monoW = mass;
    count = new int[numused];
    usedZ = new int[numused];
    numused = 0;
    for (int z=1; z<Element.Z_LIMIT; z++) 
      if (c[z] != 0) {
	count[numused] = c[z];
	usedZ[numused] = z;
	numused++;
	cofz[z] = c[z];
      }
    
    
  }

  public String toString() {
    StringBuffer res = new StringBuffer();
    boolean hasNegativeCounts = false;
    for (int i=0; i<usedZ.length; i++) 
      if (count[i]>1)
	res.append(Element.elements[usedZ[i]].symbol+count[i]);
      else if (count[i]==1)
	res.append(Element.elements[usedZ[i]].symbol);
      else 
	hasNegativeCounts = true;
    if (hasNegativeCounts) {
      res.append("-");
      for (int i=0; i<usedZ.length; i++) 
	if (count[i]<-1)
	  res.append(Element.elements[usedZ[i]].symbol+(-count[i]));
	else if (count[i]==-1)
	  res.append(Element.elements[usedZ[i]].symbol);
    }
    return res.toString();
  }

  public int getCount(int Z) {
    return cofz[Z];
  }

  public ChemicalFormula add(ChemicalFormula other) {
    int[] newcofz = new int[Element.Z_LIMIT];
    for (int i=0; i<newcofz.length; i++)
      newcofz[i] = cofz[i]+other.cofz[i];
    return new ChemicalFormula(newcofz);
  }

  public ChemicalFormula add(String formula) {
    return add(new ChemicalFormula(formula));
  }

  public ChemicalFormula add(ChemicalFormula[] others) {
    int[] newcofz = new int[Element.Z_LIMIT];
    for (int i=0; i<count.length; i++)
      newcofz[usedZ[i]] = count[i];
    for (int j=0; j<others.length; j++) {
      int[] ocount = others[j].count;
      int[] ousedZ = others[j].usedZ;
      int olength = ocount.length;
      for (int i=0; i<olength; i++)
	newcofz[ousedZ[i]] += ocount[i];
    }
    return new ChemicalFormula(newcofz);
  }
  
  //Instead of computing with the binomialdistibution, we just simulate 

  //www.matrixscience.com has a reference to an algorithm for such a thing :
  //Yergey, JA, A general approach to calculating isotopic
  //distributions for mass spectrometry. Int. J. Mass Spectrom. Ion
  //Phys., 52 337-49 (1983). But we haven't read that.

  public double[][] toIsotopicSpectrum(int numsamples, Random rnd) {
    double resolution = 0.1; //How do we model the instrument's resolution?
    TreeMap tree = new TreeMap();
    for (int i=0; i<numsamples; i++) {
      double mass = 0.0;
      for (int j=0; j<count.length; j++) {
	Element elem = Element.elements[usedZ[j]];
	for (int k=0; k<count[j]; k++)
	  mass += elem.randomIsotope(rnd).W;
      }
      //So, now we have the mass of a particular isotopic
      //composition. If there already is another peak nearby, we merge
      //with it immediately. This keeps tree small at all times. Note
      //that for any reasonably sized peptide fragment there will with
      //very high probability be at most a handfull of "strange"
      //isotopes used and so mass-monoW will be very close to an
      //integer. It follows that the merging procedure applied here
      //will not depend on the order in which the masses are added and
      //will in fact produce the right spectrum. 

      //So, perhaps we should just compute mass-monoW and add its
      //fractional part to a bin indexed by the integer part. At the
      //end we would have two arrays: count[] and offsetsum[] from
      //which we could easily compute the peaks and counts.

      //Note that the above argument really depends on the isotope
      //mass(differenc)es being very close to integer values. For
      //heavier elements it will break down.
      Double mAss = new Double(mass);
      SortedMap foo = tree.headMap(mAss);
      double bestdist = resolution;
      double lowerdist;
      double upperdist;
      Double match = null;
      if (!foo.isEmpty()) {
	Object lowerpeak = foo.lastKey();
	lowerdist = mass - ((Double)lowerpeak).doubleValue();
	if (lowerdist<resolution) {
	  match = (Double)lowerpeak;
	  bestdist = lowerdist;
	}
      }
      foo = tree.tailMap(mAss);
      if (!foo.isEmpty()) {
	Object upperpeak = foo.firstKey();
	upperdist = ((Double)upperpeak).doubleValue() - mass;
	if (upperdist<bestdist) 
	  match = (Double)upperpeak;
      }
      int newcount = 1;
      if (match != null) {
	newcount += ((Integer)(tree.remove(match))).intValue();
	mAss = new Double(((newcount-1)*match.doubleValue()+mass)/newcount);
      }
      tree.put(mAss, new Integer(newcount));
    } 
    
    int numpeaks = tree.size();
    double[][] res = new double[numpeaks][2];
    Iterator iter = tree.keySet().iterator();
    int i=0;
    while (iter.hasNext()) {
      Double key = (Double)(iter.next()); 
      res[i][0] = key.doubleValue();
      res[i][1] = ((Integer)(tree.get(key))).intValue();
      //System.out.println(res[i][0]+" "+res[i][1]);
      i++;
    }
    return res;
  }

  public static ChemicalFormula H = new ChemicalFormula("H");
  public static ChemicalFormula OH = new ChemicalFormula("OH");
  
  public static void main(String[] args) {
    System.out.println(new ChemicalFormula("NaOH+C2H5OH-H2O"));
    System.out.println(new ChemicalFormula("NaOH-H2O"));
  }
}

