import java.awt.*;
import java.awt.event.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;

//import javax.swing.event.*;

/**
 *   Controller acts as a bridge between view and model classes
 *   (although this program doesn't really fit into MVC pattern)
 */
class Controller extends AbstractController {
  private final static DecimalFormat fmt2 = new DecimalFormat("0.00");
  private final static DecimalFormat fmt3 = new DecimalFormat("0.000");

  //
  private LockableJFrame mainFrame;
  private LockableJFrame dpparmsFrame;
  private DebugOut debugOut;

  //
  private SpectrumWindow sw;
  private ProposalPanel propP;
  private Spectrum spectrum;
  private Sequences sequences;

  //These are so far accessible from other classes
  //SequencingParameters seqParms;
  //boolean bwdSequence;		// Show accepted sequence backwards



  private Observable spectrumObservable;
  private Observable fileObservable;

  //Various UI components created in Sequencer.main():
  private JTextField xCoor;
  private JTextField yCoor;
  private JTextField lDist;
  private JTextField rDist;
  private JTextField msgField;
  private JTextField ionTypeMatch;

  private JTextField seqOut;
  private JTextField mDelta;

  private UndoStack undoStack = new UndoStack();

  //------------------------------------------------------------------
  //Constructor

  Controller(JTextField lDist, JTextField rDist, 
	     JTextField xCoor, JTextField yCoor,
	     JTextField msgField, JTextField ionTypeMatch,
	     JTextField seqOut, JTextField mDelta,
	     Observable fileObservable,
	     LockableJFrame mainFrame) {
    this.lDist        = lDist;
    this.rDist        = rDist;
    this.xCoor        = xCoor;
    this.yCoor        = yCoor;
    this.msgField     = msgField;
    this.ionTypeMatch = ionTypeMatch;
    this.seqOut = seqOut;
    this.mDelta = mDelta;
    this.fileObservable = fileObservable;
    this.mainFrame    = mainFrame;
    
    seqParms = new SequencingParameters();

    //Note: this is only for resetting DP data on spectrum changes
    spectrumObservable = new Observable() {
	public void notifyObservers() {
	  setChanged();
	  super.notifyObservers();
	}
      };
    //This forwards fileRead notifications to those listening for
    //spectrumchange notifications.
    fileObservable.addObserver(new Observer() {
	public void update(Observable o, Object arg) {
	  spectrumObservable.notifyObservers();
	}
      });
    
    debugOut = new DebugOut();
  }

  //-------------------------------------------------------------
  //SpectrumWindow IF

  public void registerSpectrumWindow(Object /*SpectrumWindow*/ sw) {
    this.sw = (SpectrumWindow)sw; //Silly, to allow AbstractController
				  //to be compiled without compiling
				  //SpectrumWindow
  }

  public void repaintSpectrumWindow() {
    sw.repaint();
  }


  //-------------------------------------------------------------
  // ProposalPanel IF

  public void registerProposalPanel(ProposalPanel propP) {
    this.propP = propP;
  }

  public void showProposal(int npi) {
    if (propP !=null)
      propP.showProposal(npi);
  }

  public boolean proposalValid() {
    return propP==null ? false : propP.proposalValid();
  }

  public Proposal currentProposal(boolean left) {
    return propP==null ? null : propP.currentProposal(left);
  }

  public void resetProposal() {
    if (propP !=null)
      propP.resetProposal();
    
  }

  //--------------------------------------------------------------
  //Spectrum IF

  //For the moment, this is called from SpectrumWindow
  public void setSpectrum(Spectrum spectrum) {
    this.spectrum = spectrum;
  }

  public Spectrum getSpectrum() {
    return spectrum;
  }

  public String getSpectrumDescription() {
    return spectrum==null ? "" : spectrum.getDescription();
  }
  
  
  public void addSpectrumObserver(Observer obs) {
    spectrumObservable.addObserver(obs);
  }

  public void setSpectrumChanged() {
    spectrumObservable.notifyObservers();
  }


  //---------------------------------------------------------------
  //Sequences IF

  public Sequences newSequences() {
    return sequences = new Sequences(this);
  }

  public void showSeq() {
    if (sequences != null)
      sequences.showSeq(seqOut, sw.getLog());
  }

  public String getSeqOut() {
    return seqOut.getText();
  }

  public void setDelta(String del) {
    mDelta.setText(del);
  }

  //----------------------------------------------------------------
  // SeqParms

  public void setDPParmsFrame(DPParmsFrame dppf) {
    dpparmsFrame = dppf;
  }


  private Observable seqParmsObservable = new Observable() {
      public void notifyObservers(Object o) {
	setChanged();
	super.notifyObservers(o);
      }
    };

  public void addSeqParmsObserver(Observer obs) {
    seqParmsObservable.addObserver(obs);
  }

  public void setSeqParmsChanged(Object arg) {
    seqParmsObservable.notifyObservers(arg);
    if (sw !=null && sequences != null) {
      sw.setParams();
      sw.setPotential();
      showSeq();
      sw.repaint();
    }
  }

  //--------------------------------------------------------------------
  // Text output

  public void message(String msg) {
    msgField.setText(msg);
    debugOut.message(msg, false);
  }

  public void debugmessage(String msg) {
    debugOut.message(msg);
  }
  public void debugmessage(String msg, boolean popup) {
    debugOut.message(msg, popup);
  }
  

  public void mousePosition() {
    lDist.setText("");
    rDist.setText("");
    xCoor.setText("");
    yCoor.setText("");
    ionTypeMatch.setText("");
  }

  public void mousePosition(double x, double rel_y) {
    if (sequences != null) {
      double ld = sequences.getDistance(x,true);
      double rd = sequences.getDistance(x,false);
      lDist.setText(fmt2.format(ld)+
		    AminoAcid.getMatchesFormatted(ld,seqParms.tolerance));
      rDist.setText(fmt2.format(rd)+
		    AminoAcid.getMatchesFormatted(rd,seqParms.tolerance));
      xCoor.setText(fmt3.format(x));
      yCoor.setText(fmt3.format(100*rel_y)+'%');
      ionTypeMatch.setText(spectrum.getIonTypeAltsFmt(x, 
						      seqParms.tolerance,
						      seqParms.grass));
    }
  }

  //---------------------------------------------------------------
  //Utilities for running computation in other thread

  public void lockGUI() {
    mainFrame.lockGUI();
    dpparmsFrame.lockGUI();
  }

  public void unlockGUI() {
    mainFrame.unlockGUI();
    dpparmsFrame.unlockGUI();
  }

  /**
   * Inner class Controller.UndoStack 
   *
   * Rudimentary support for undo. Should be inner in
   * Controller. Right now UndoStack does not use fields or methods
   * from SpectrumWindow (that is not available in Controller), but
   * Actions in SpectrumWindow use methods from UndoStack - in
   * particular checkPoint() and clear().
   */

  //At this point we do not: 
  //
  // 2. save state accross setSpectrum(). One might wish to do so for
  // reinterpretOne() and perhaps mirror()), but in this case, the saved
  // marks in sequences.save() may contain stale Peak references
  // unless we also save the complete spectrum. 
  //
  // 3. save and restore candidates. Now the last candidates computed
  // are retained in undo! Packaging the results from DP into an
  // object should make it easy to remedy this.
  //
  // 4. Use thew javax.swing.undo framwork, which seems to use similar
  // ideas.
  //
  // 5. Save zoom and translate state

  class UndoStack {
    private class UndoItem {
      Object sequencesState;
      SequencingParameters sp;
      UndoItem(Object sequencesState, SequencingParameters sp) {
	this.sequencesState = sequencesState;
	this.sp = sp;
      }
    }
    private LinkedList undoStack = new LinkedList();
    private Action undoAction = new AbstractAction("Undo") {
	{
	  setEnabled(false);
	}
	public void actionPerformed(ActionEvent e) {
	  undo();
	}
      };
    private Action redoAction = new AbstractAction("Redo") {
	{
	  setEnabled(false);
	}
	public void actionPerformed(ActionEvent e) {
	  redo();
	}
      };
    private int current = -1;
    void checkPoint() {
      for (int i=undoStack.size()-1; i>current; i--)
	undoStack.remove(i);
      undoStack.addLast(new UndoItem(sequences.save(), 
				     seqParms.save()));
      //if (undoStack.size()>100)
      //undoStack.removeFirst();
      current = undoStack.size()-1;
      if (redoAction.isEnabled())
	redoAction.setEnabled(false);
      if (current == 1)
	undoAction.setEnabled(true);
    }
    void clear() {
      undoStack.clear();
      current = -1;
      if (undoAction.isEnabled())
	undoAction.setEnabled(false);
      if (redoAction.isEnabled())
	redoAction.setEnabled(false);
    }
    private void redo() {
      if (current<undoStack.size()-1) {
	restore(++current);
	if (!undoAction.isEnabled())
	  undoAction.setEnabled(true);
      }
      else
	message("No more redo");
      if (current == undoStack.size()-1)
	redoAction.setEnabled(false);
    }
    private void undo() {
      if (current>0 ) { 
	restore(--current);
	if (!redoAction.isEnabled())
	  redoAction.setEnabled(true);
      }
      else 
	message("No more undo");
      if (current == 0)
	undoAction.setEnabled(false);
    }
    private void restore(int i) {
      sequences.clear();
      sequences.restore(((UndoItem)undoStack.get(i)).sequencesState);
      seqParms.restore(((UndoItem)undoStack.get(i)).sp);
      setSeqParmsChanged(UndoStack.this);
    }
  }

  public void checkPoint() {
    undoStack.checkPoint();
  }

  public void clearUndoStack() {
    undoStack.clear();
  }

  public Action getUndoAction() {
    return undoStack.undoAction;
  }
  public Action getRedoAction() {
    return undoStack.redoAction;
  }

}



/**
 * A wrapper for LogFrame
 */
interface DebugOutInterface {
  //Should message popup on each message?
  void setPopup(boolean b);
  //Should messages be ignored?
  void setEnabled(boolean b);
  void message(String s);
  void message(String msg, boolean doPopup);
  void show();
  }

/**
 * DebugOut as a wrapper for LogFrame
 */

class DebugOut implements DebugOutInterface {
  private LogFrame debugFrame;
  private boolean isEnabled = true;
  private boolean doPopup = true;
  private boolean exitOnClose;
  DebugOut() {
    this.exitOnClose =  false;
  }
  private DebugOut(boolean exitOnClose) {
    this.exitOnClose =  exitOnClose;
  }
  public static void main(String[] args) {
    DebugOut dbo = new DebugOut(true);
    dbo.setEnabled(true);
    dbo.message("hej");
    dbo.message("hejhej", true);
    for (int i=0; i< 55; i++)
      dbo.message(""+i);
    dbo.message("hejhej");
  }
  public void setPopup(boolean b) {
    doPopup = b;
  }
  public void setEnabled(boolean b){
    isEnabled = b;
  }
  public void message(String msg) {
    message(msg, doPopup);
  }
  public void message(String msg, boolean doPopup) {
    if (isEnabled) {
      if (debugFrame == null)
	debugFrame = new LogFrame("Debug messages", exitOnClose);
      debugFrame.message(msg,doPopup);
    }
  }
  public void show() {
    if (debugFrame == null)
      debugFrame = new LogFrame("Debug messages", exitOnClose);
    debugFrame.popup();
  }
}



/**
 * class ProposalPanel
 *
 */

class ProposalPanel extends JPanel {
  private JButton updateB, hideB, globalLeftB, globalRightB;
  private JScrollBar leftSb, rightSb;
  private JTextField tf, lrf, rrf;
  
  private Controller controller;
  private Spectrum spectrum;
  
  private boolean showCandidate;        // Show DP candidate sequence
  private Proposal leftCandidate, rightCandidate;
  int current_peak_index = -1;          // where to start a local candidate
  
  ProposalAlgo prop;
  
  public boolean proposalValid() {
    return showCandidate;
  }
  
  public Proposal currentProposal(boolean left) {
    return left ? leftCandidate : rightCandidate;
  }
  
  public void resetProposal() {
    doReset();
  }
  
  private void doReset() {
    showCandidate = false;
    leftCandidate = null;
    rightCandidate = null;
    current_peak_index = -1;
    leftSb.setEnabled(false);
    rightSb.setEnabled(false);
    updateTexts();
  }
  
  void updateTexts() {
    if (!showCandidate) {
      lrf.setText("");
      rrf.setText("");
      tf.setText("");
      return;
    }
    lrf.setText(leftSb.isEnabled() ? (leftSb.getValue()+1)+"" : "");
    rrf.setText(rightSb.isEnabled() ? (rightSb.getValue()+1)+"" : "");

    if (current_peak_index > -1) {
      tf.setText(prop.getSequence(current_peak_index,
				  leftSb.isEnabled()?leftSb.getValue():0,
				  rightSb.isEnabled()?rightSb.getValue():0,
				  controller.bwdSequence));
    }
    else if (leftSb.isEnabled()) {
      tf.setText(prop.getSequence(-1,true,leftSb.getValue(),
				  controller.bwdSequence));
    }
    else if (rightSb.isEnabled()) {
      tf.setText(prop.getSequence(-1,false,rightSb.getValue(),
				  controller.bwdSequence));
    }
    else {
      tf.setText("");
    }
  }

  private void setCandidates() {
    if (!showCandidate)
      return;
    if (current_peak_index > -1) {
      if (leftSb.isEnabled()) {
	leftCandidate = prop.getProposal(current_peak_index,
					 true,
					 leftSb.getValue());
      }
      else {
	leftCandidate = null;
      }
      if (rightSb.isEnabled()) {
	rightCandidate = prop.getProposal(current_peak_index,
					  false,
					  rightSb.getValue());
      }
      else {
	rightCandidate = null;
      }
    }
    else if (leftSb.isEnabled()) {
      leftCandidate = prop.getProposal(-1, true, leftSb.getValue());
      rightCandidate = null;
    }
    else if (rightSb.isEnabled()) {
      leftCandidate = null;
      rightCandidate = prop.getProposal(-1, false, rightSb.getValue());
    }
    updateTexts();
    controller.repaintSpectrumWindow();
  }

  private void showLeftGlobal() {
    doReset();
    showCandidate = true;
    initSb(true,prop.getProposalCount(-1,true));
    setCandidates();
  }
    
  private void showRightGlobal() {
    doReset();
    showCandidate = true;
    initSb(false,prop.getProposalCount(-1,false));
    setCandidates();
  }
    
  private Observer spectrumObserver = new Observer() {
      public void update(Observable o, Object arg) {
	spectrum = controller.getSpectrum();
	doReset(); 
	updateB.setEnabled(true);
	globalLeftB.setEnabled(false);
	globalRightB.setEnabled(false);
	hideB.setEnabled(false);
      }
    };
    
  public void showProposal(int current_peak_index) {
    if (!showCandidate || current_peak_index <0 || prop==null)
      return;
    this.current_peak_index = current_peak_index;
    //Left side:
    int max_rank = prop.getProposalCount(current_peak_index,true);
    initSb(true, max_rank);
    //Right side:
    max_rank = prop.getProposalCount(current_peak_index,false);
    initSb(false, max_rank);
    setCandidates();
  }
    
  void initSb(boolean left, int max_rank) {
    JScrollBar sb = left ? leftSb : rightSb;
    if (max_rank>-1) {
      sb.setEnabled(false);
      sb.setValue(0);
      sb.setMaximum(max_rank);
      sb.setEnabled(true);
    }
    else 
      sb.setEnabled(false);
  }

  ProposalPanel(Controller controller_in) {
    controller = controller_in;
    controller.registerProposalPanel(this);
    setLayout(new BorderLayout());
    Box nP = new Box(BoxLayout.X_AXIS);
    nP.add(new JLabel("Computed proposals:"));
    nP.add(updateB = new JButton("Compute"));
    nP.add(globalLeftB = new JButton("Show left global"));
    nP.add(globalRightB = new JButton("Show right global"));
    nP.add(hideB = new JButton("Hide proposals"));
    add(nP,BorderLayout.NORTH);
    Box sP = new Box(BoxLayout.X_AXIS);
    sP.add(new JLabel("L rank:"));
    sP.add(lrf = new JTextField(2));
    sP.add(leftSb = new JScrollBar(JScrollBar.HORIZONTAL,0,1,0,1));
    sP.add(tf = new JTextField(40));
    sP.add(rightSb = new JScrollBar(JScrollBar.HORIZONTAL,0,1,0,1));
    sP.add(new JLabel("R rank:"));
    sP.add(rrf = new JTextField(2));
    lrf.setMaximumSize(new Dimension(2,20));
    rrf.setMaximumSize(new Dimension(2,20));
    add(sP,BorderLayout.SOUTH);
    lrf.setEditable(false);
    rrf.setEditable(false);
    tf.setEditable(false);
    updateB.setEnabled(false); 
    globalLeftB.setEnabled(false);
    globalRightB.setEnabled(false);
    hideB.setEnabled(false);
    doReset();
    controller.addSpectrumObserver(spectrumObserver);
      
    updateB.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  final Score score = new SeqScore(spectrum.maxAbundance,
					   spectrum.totalMass,
					   controller.seqParms);
	  doReset();
	  final ProgressMonitor monitor = 
	    new ProgressMonitor(ProposalPanel.this,
				"Computing",
				null,
				0,
				2*spectrum.obs.length);
	  monitor.setMillisToDecideToPopup(250);
	  monitor.setMillisToPopup(500);
	  //The methods of feedback is called form the doCompute
	  //worker thread that should not call mathods that update
	  //the GUI. Therefore it must use SwingUtilities.invokeLater()
	  final ProposalAlgo.Feedback feedback = new ProposalAlgo.Feedback() {
	      public boolean isCanceled() {
		return monitor.isCanceled();
	      }
	      public void setMaximum(final int m) {
		SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
		      monitor.setMaximum(m);
		    }
		  });
	      }
	      public void setProgress(final int nv) {
		SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
		      monitor.setProgress(nv);
		    }
		  });
	      }
	    };
	  //Silly local class for communicating elapsed time from the
	  //worker thread to doShow(). There can be no overlapping
	  //acces by the two threads, so we don't need 'synchronized'(?)
	  class Tyme {long startTime, endTime; };
	  final Tyme tyme = new Tyme();
	  //
	  //doShow is invoked on the main AWT event thread by
	  //doCompute() when it is done
	  final Runnable doShow = new Runnable() {
	      public void run() {
		if (monitor.isCanceled())
		  prop=null;
		monitor.close();
		if (prop!=null) {
		  if (prop.getProposalCount(-1,true) > 0 ) {
		    globalLeftB.setEnabled(true);
		    showCandidate = true;
		  }
		  if (prop.getProposalCount(-1,false) > 0 ) {
		    globalRightB.setEnabled(true);
		    showCandidate = true;
		    showRightGlobal();
		  }
		  if (showCandidate)
		    hideB.setEnabled(true);
		  controller.debugmessage("DenovoDP elapsed time:" +
					  (tyme.endTime-
					   tyme.startTime)/1000.0, false);
		}
		else {
		  globalLeftB.setEnabled(false);
		  globalRightB.setEnabled(false);
		  hideB.setEnabled(false);
		  doReset();
		}
		controller.unlockGUI();
	      }
	    };
	  //This will run in another thread (than the AWT dispatch thread)
	  Runnable doCompute = new Runnable() {
	      public void run() {
		tyme.startTime = System.currentTimeMillis();
		prop = new DenovoDP(spectrum, score, feedback);
		tyme.endTime = System.currentTimeMillis();
		SwingUtilities.invokeLater(doShow);
	      }
	    };
	  controller.lockGUI();
	  new Thread(doCompute).start();
	}
      });

    globalLeftB.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  showLeftGlobal();
	}
      });

    globalRightB.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  showRightGlobal();
	}
      });

    hideB.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  doReset();
	  controller.repaintSpectrumWindow();
	}
      });

    AdjustmentListener selSol = new AdjustmentListener() {
	public void adjustmentValueChanged(AdjustmentEvent e) {
	  setCandidates();
	}
      };
      
    leftSb.addAdjustmentListener(selSol);
    rightSb.addAdjustmentListener(selSol);
  }
}


/**
 * Window for changing sequencing parameters
 */
class DPParmsFrame extends LockableJFrame {
  private final static DecimalFormat fmt3 = new DecimalFormat("0.000");
  private DPSlider 
    betaSb, toleranceSb, grassSb, 
    nuSb, sigmaSb, lambdaSb, krpenaltySb;
  private JCheckBox
    adCb, ieCb;
  private JPanel outerP;
  private JButton setB, resetB, defaultsB;
  private GridBagConstraints lblC,sbC,txtC;

  class DPSlider {
    private JSlider sb;
    private JTextField tf;
    double bot, top;
    int marks;
    DPSlider(String label, final double bot, 
		final double top, final int marks) {
      this.bot = bot;
      this.top = top;
      this.marks = marks;
      outerP.add(new JLabel(label), lblC);
      outerP.add(sb = new JSlider(JSlider.HORIZONTAL,0,marks,0),sbC);
      //sb.setPaintTicks(true);
      //sb.setMajorTickSpacing(marks/10);
      Hashtable lbltbl = new Hashtable(2);
      lbltbl.put(new Integer(0), new JLabel(fmt3.format(bot)));
      lbltbl.put(new Integer(marks), new JLabel(fmt3.format(top)));
      sb.setLabelTable(lbltbl);
      sb.setPaintLabels(true);
      tf = new JTextField(5) {
	  //public boolean isFocusTraversable() {
	  //  return false;
	  //}
	};
      outerP.add(tf, txtC);
      tf.setEditable(false);
      sb.addChangeListener(new ChangeListener() {
	  public void stateChanged(ChangeEvent e) {
	    tf.setText(fmt3.format(bot+ sb.getValue()*(top-bot)/marks));
	    setB.setEnabled(true);
	    resetB.setEnabled(true);
	    defaultsB.setEnabled(true);
	  }
	});
    }
    double getValue() {
      return fmt3.parse(tf.getText(), new ParsePosition(0)).doubleValue();
    }
    void setValue(double val) {
      sb.setValue((int) ( marks * (val - bot) / (top-bot)));
      tf.setText(fmt3.format(val));
    }
    void setEnabled(boolean b) {
      sb.setEnabled(b);
    }
  }
    
  DPParmsFrame(final Controller controller) {
    super("KVL AminoSequencer: Dynamic programming parameters");
    setDefaultCloseOperation(HIDE_ON_CLOSE);
    controller.setDPParmsFrame(this);
    controller.addSeqParmsObserver(new Observer() {
	public void update(Observable o, Object arg) {
	  if (arg != DPParmsFrame.this) {
	    //Note: we reset non-set changes
	    setValues(controller.seqParms);
	  }
	}
      });
    outerP = new JPanel();
    outerP.setLayout(new GridBagLayout());
    lblC = new GridBagConstraints();
    lblC.weightx = 0.1;
    lblC.anchor = GridBagConstraints.WEST;
    lblC.insets = new Insets(1, 1, 1, 1);
    lblC.gridwidth = GridBagConstraints.WEST;
      
    sbC = (GridBagConstraints) lblC.clone();
    sbC.weightx = 1.0;
    sbC.fill = GridBagConstraints.HORIZONTAL;

    txtC = (GridBagConstraints) lblC.clone();
    txtC.anchor = GridBagConstraints.EAST;
    txtC.gridwidth =  GridBagConstraints.REMAINDER;

      
    toleranceSb = new DPSlider("mass tolerance", 0.1, 1.5, 140);
    //
    ActionListener cbListener = new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  setB.setEnabled(true);
	  resetB.setEnabled(true);
	  defaultsB.setEnabled(true);
	}
      };
    outerP.add(new JLabel("independent errors"), lblC);
    outerP.add(ieCb = new JCheckBox(), sbC);
    ieCb.addActionListener(cbListener);
    outerP.add(new JLabel(""), txtC);
    //
    outerP.add(new JLabel("allow duplicates if mirrored"), lblC);
    outerP.add(adCb = new JCheckBox(), sbC);
    adCb.addActionListener(cbListener);
    outerP.add(new JLabel(""), txtC);
    //
    grassSb = new DPSlider("abundance threshold", 0.001, 0.100, 99);
    sigmaSb = new DPSlider("\u03c3 (mass std. error)", 
			      0.1, 1.0, 90);
    nuSb = new DPSlider("\u03bd (noise level)", 
			   0.001, 1.00, 999);
    krpenaltySb = new DPSlider("\u03b3 (confidence of K/R (only) at the end)", 
			       0.0, 1.00, 100);
    lambdaSb = new DPSlider("\u03bb (weighting of size vs. penalty)", 
			    0.00, 2.0, 100);
    betaSb = new DPSlider("\u03b2 (weighting of count vs length)", 
			     0.0, 1.0, 100);
    JPanel cP = new JPanel();
    cP.setLayout(new FlowLayout());
    cP.add(setB = new JButton("Set"));
    cP.add(resetB = new JButton("Reset"));
    cP.add(defaultsB = new JButton("Defaults"));
    outerP.add(cP,lblC);
    JPanel foo = new JPanel();
    foo.setPreferredSize(new Dimension(200,1));
    outerP.add(foo,sbC);
    //outerP.setBorder(BorderFactory.createEtchedBorder());
    //outerP.setPreferredSize(new Dimension(600,300));
    setDefaults();
    setB.setEnabled(false);
    resetB.setEnabled(false);
    getContentPane().add(outerP);

    setB.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent ev) {
	  SequencingParameters seqp = controller.seqParms;
	  seqp.beta =  betaSb.getValue();
	  seqp.tolerance =  toleranceSb.getValue();
	  seqp.grass =  grassSb.getValue();
	  seqp.nu =  nuSb.getValue();
	  seqp.sigma =  sigmaSb.getValue();
	  seqp.lambda =  lambdaSb.getValue();
	  seqp.krpenalty = krpenaltySb.getValue();
	  seqp.independent_errors = ieCb.isSelected();
	  seqp.allowDuplicatePeaks = adCb.isSelected();
	  controller.setSeqParmsChanged(DPParmsFrame.this);
	  setB.setEnabled(false);
	  resetB.setEnabled(false);
	}
      });

    resetB.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent ev) {
	  setValues(controller.seqParms);
	  setB.setEnabled(false);
	  resetB.setEnabled(false);
	}
      });
      
    defaultsB.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent ev) {
	  setDefaults();
	  setB.setEnabled(true);
	  resetB.setEnabled(true);
	}
      });

  }

  void setValues(SequencingParameters seqParms) {
    toleranceSb.setValue(seqParms.tolerance);      
    grassSb.setValue(seqParms.grass);
    nuSb.setValue(seqParms.nu);
    sigmaSb.setValue(seqParms.sigma);
    lambdaSb.setValue(seqParms.lambda);
    betaSb.setValue(seqParms.beta);
    krpenaltySb.setValue(seqParms.krpenalty);
    ieCb.setSelected(seqParms.independent_errors);
    adCb.setSelected(seqParms.allowDuplicatePeaks);
  }

  void setDefaults() {
    setValues(new SequencingParameters()); 
    defaultsB.setEnabled(false);
  }
}
 
