
import java.util.*;

/**
 * A concrete proposal algorithm
 */
class DenovoDP implements ProposalAlgo {
  //Perform global dynamic programming.
  //
  //ridiculous NB: we assume mass_tolerance < lowest amino acid mass

  private int tbCount = 10; //How many tracebacks do we remember?

  //Main datastructure created by constructor
  private Endpoint[][] leftTracebacks;
  private Endpoint[][] rightTracebacks;
  private Endpoint[] topLeftArray;
  private Endpoint[] topRightArray;

  private int lastEPid; //To make Endpoint comparisons correctly
  
  //reference to the spectrum 
  private Spectrum spectrum;

  private static AminoAcid[] sortedAcids;
  static {
    sortedAcids = 
      (AminoAcid[]) AminoAcid.aaByMonoW.toArray(new AminoAcid[0]);
  }

  DenovoDP(Spectrum spectrum, Score score, Feedback feedback) {
    this.spectrum = spectrum;
    //renormalize
    boolean independent_errors = score.getIndependentErrors();
    boolean pruneDuplicatePeaks = 
      spectrum.isMirrored && !score.getAllowDuplicatePeaks();
    double a_limit = spectrum.maxAbundance*score.getGrass();
    double mass_tolerance = score.getTolerance();
    double d_mass_tolerance = 
      independent_errors ? mass_tolerance : 2.0*mass_tolerance;
    Peak[] obs = spectrum.obs;
    leftTracebacks = new Endpoint[obs.length][];
    rightTracebacks = new Endpoint[obs.length][];
    //
    //
    EndpointQ topLeft = new EndpointQ(tbCount);
    EndpointQ topRight = new EndpointQ(tbCount);
    //
    int progress = 0; 
    if (feedback != null)
      feedback.setMaximum(2*obs.length);
    for (int side=-1; side<=1;side+=2) {
      boolean rightsweep = side == 1; 
      //
      int index_start = rightsweep ? 0 : obs.length-1;
      int index_end =  rightsweep ? obs.length-1 : 0;
      //Optimization: first possible index at which to find a candidate peak
      int index_limit = index_start; 
      //
      //Below, 
      // i is an index into the peak array
      // a is an index into the amino acid array
      // j is an index into the peak array for a candidate amino acid
      for (int i=index_start; side*(index_end-i)>=0; i += side) {
	if (feedback!= null) {
	  if (feedback.isCanceled())
	    return;
	  feedback.setProgress(progress++);
	}
	//Immediately ignore low and ignored peaks
	if (obs[i].y <= a_limit || obs[i].ignore)
	  continue;
	//Optimization: first possible index at which to find a
	//candidate peak for current amino acid
	int local_index_limit = index_limit;
	//
	//
	EndpointQ topTbs = new EndpointQ(tbCount);
	//add a null traceback corresponding to starting here
	{
	  double nullscore = score.peakscore(obs[i]);
	  //Assume the mass of this peak is exact
	  Endpoint nulltb = new Endpoint(obs[i].x,i,null,null,nullscore);
	  topTbs.add(nulltb); 
	}
	//Always take amino acids in order of decreasing mass
	for (int a=sortedAcids.length-1; a>=0; a--) {
	  AminoAcid acid = sortedAcids[a];
	  double monoW = acid.monoW;

	  //Search for the nearest possible peak for this candidate acid.
	  //
	  double w = obs[i].x - side * (monoW + d_mass_tolerance);
	  if (side*(obs[local_index_limit].x - w) >= 0) {
	    //Found
	  }
	  else {
	    //Binary search 
	    //
	    //Here obs[local_index_limit].x is farer from obs[i].x than
	    //w and obs[i].x is nearer than w. We will invariantly have
	    //obs[far].x farer than w and obs[near].x at least as near as w.
	    int far = local_index_limit;
	    int near = i;
	    do {
	      int r = (far+near)/2;
	      if (side*(obs[r].x - w) <0 ) //
		far = r;
	      else 
		near = r;
	    } while ( side*(near - far) > 1 );
	    local_index_limit = near;//First index of peak at least as near as w
	    if (a==sortedAcids.length-1)
	      index_limit = local_index_limit;//optimize search start for next i
	  }
	  //Check candidate positions (linear search for best
	  //candidate). We accept all peaks in tolerance interval with
	  //abundance above threshold
	  //
	  //
	  int j = local_index_limit;
	  while (side*(obs[j].x-w) <= 2*d_mass_tolerance) {
	    if (obs[j].y > a_limit && !obs[j].ignore) {
	      Endpoint[] preds = 
		rightsweep ? leftTracebacks[j] : rightTracebacks[j];
	      if (preds != null ) 
		for (int pred=0; pred<preds.length; pred++) {
		  Endpoint ep = preds[pred];
		  if (pruneDuplicatePeaks && ep.gibtEs(obs[i].orig_index))
		    continue;
		  //THIS IS DIFFERENT FROM THE COMPUTATION IN Sequences!!!
		  double predictedmass = 
		    (independent_errors ? obs[j].x : ep.predictedmass) + 
		    side*monoW;
		  //NBNBNBNBNB
		  // double mass_error = Math.round(predictedmass) - obs[i].x;
		  double mass_error = predictedmass - obs[i].x;
		  //This may only happen if
		  //d_mass_tolerance>mass_tolerance, i.e. if
		  //independent_errors==false:
		  if (Math.abs(mass_error)>mass_tolerance)
		    continue;
		  Peak olepeak;
		  double newscore = 
		    ep.score + score.stepscore(acid, mass_error,obs[i],obs[j]);
		  topTbs.add(new Endpoint(predictedmass,
					  i, ep, acid, newscore));
		}
	    }
	    j += side;
	  }
	}
	//Tried all amino acids
	// Remember best tracebacks globally
	Endpoint[] preds = topTbs.toArray();
	if (rightsweep) 
	  topLeft.addAll(leftTracebacks[i] = preds);
	else 
	  topRight.addAll(rightTracebacks[i] = preds);	
      }
    }
    topLeftArray = topLeft.toArray();
    topRightArray = topRight.toArray();
  }
  
  public String toString() {
    StringBuffer res = new StringBuffer();
    Endpoint[] t = topLeftArray;
    for (int seq=0; seq<(t == null?0:t.length); seq++) 
      res.append("Left Solution "+(seq+1)+": "+
			 t[seq].toAASeq(true) + "\n");
    t = topRightArray;
    for (int seq=0; seq<(t == null ? 0 : t.length); seq++) 
      res.append("Right Solution "+(seq+1)+": "+
			 t[seq].toAASeq(true) + "\n");
    return res.toString();
  }
  public Proposal getProposal(int i, boolean left, int rank) {
    if (i < 0)
      return (left ? topLeftArray : topRightArray)[rank].candidateList();
    else 
      return (left?leftTracebacks:rightTracebacks)[i][rank].candidateList();
  }

  public String getSequence(int current_peak, 
			     int current_left_rank, 
			     int current_right_rank, 
			     boolean reversed) {
    StringList res;
    if (leftTracebacks[current_peak].length > current_left_rank) {
      res = leftTracebacks[current_peak][current_left_rank].toAAList();
    } else 
      res = new StringList();
    res.addLast("<>");
    if (rightTracebacks[current_peak].length > current_right_rank) {
      res.addAll(rightTracebacks[current_peak][current_right_rank].toAAList());
    }
    return res.toStringBuffer(reversed).toString();
  }

  public String getSequence(int i,boolean left,int rank,boolean bwdSequence){
    if (i < 0)
      return (left ? topLeftArray : topRightArray)[rank].toAASeq(bwdSequence);
    else 
      return (left?leftTracebacks:rightTracebacks)[i][rank].toAASeq(bwdSequence);
  }

  public int getProposalCount(int i, boolean left) {
    if (i < 0)
      return (left ? topLeftArray : topRightArray).length;
    else 
      return (left ? leftTracebacks[i] : rightTracebacks[i]).length;
  }

  public static void main(String[] args) throws java.io.IOException {
    // Usage: java Denovo filename mass_tolerance
    String filename = args.length >0 ? args[0] : "Qtof_IPDGFAGAQGGITFR.dta";
    Spectrum s = Spectrum.readFile(filename);
    Score score = new SeqScore(s.maxAbundance,s.totalMass,
			       new SequencingParameters());
    System.out.println((new DenovoDP(s,score,null)).toString());
  }

  /**
   * Inner class DenovoDP.Endpoint 
   *
   * A simple one-way linked list describing a possible sequence
   * ending at spectrum.obs[this.id]. 
   */

  private class Endpoint implements Comparable {
    private int id;               //To make ordering total
    private double predictedmass; //
    private int index;            //Index into spectrum.obs
    private Endpoint prev;        //Previous Endpoint
    private AminoAcid acid;       //The acid of the interval there
    private double score;
    Endpoint(double pm, int i, Endpoint ep, AminoAcid a, double s) {
      id = lastEPid++;
      predictedmass = pm;
      index = i;
      prev = ep;
      acid = a;
      score = s;
    }
    // Utility function for mirrored spectra. Tells whether orig_index
    // appears as spectrum.obs[index].orig_index in this sequence.
    boolean gibtEs(int orig_index) {
      Endpoint ep = this;
      //A linear search in this linked list may seem inefficient, but
      //running the constructor with and without allowDuplicatePeaks
      //indicates that this will not be a bottleneck of the dynamic
      //programming. Anyway, using other algorithms would incur setup
      //penalties of extra datastructures and we have a lot of short
      //lists to search, each one only a few times.
      while (ep != null) {
      if (spectrum.obs[ep.index].orig_index == orig_index)
	return true;
	ep = ep.prev;
      }
      return false;
    }

    // We do not explicitly remember which way we are headed, but it
    // is simple to infer from the data:
    private boolean isLeftWards() {
      if (prev == null)
	return true; //answer is arbitrary
      return index>prev.index;
    }
    public int compareTo(Object o) {
      //Invert the sign so that the higher scoring elements comes first
      double oscore = ((Endpoint) o).score;
      //if the scores happens to be equal, prefer (arbitrarily) the
      //element created first
      return score<oscore ? 1 : score>oscore ? -1 : id - ((Endpoint) o).id;
    }
    public String toString() {
      StringBuffer res = 
	new StringBuffer(score + "("+spectrum.obs[index].x+")");
      Endpoint ep = prev;
      AminoAcid a = acid;
      while (ep != null) {
	res.append(a.abbr + "("+spectrum.obs[ep.index].x+")");
	a = ep.acid;
	ep = ep.prev;
      }
      return res.toString();
    }
    public String toAASeq (boolean bwdSequence) {
      return toAAList().toStringBuffer(bwdSequence).toString();
    }
    
    public StringList toAAList() {
      boolean isLeftWards = isLeftWards();
      StringList res = new StringList();
      Endpoint ep = prev;
      AminoAcid a = acid;
      while (ep != null) {
	if (isLeftWards)
	  res.addFirst(a.abbr);
	else 
	  res.addLast(a.abbr);
	a = ep.acid;
	ep = ep.prev;
      }
      return res;
    }

    public Proposal candidateList() {
      LinkedList res = new LinkedList();
      //Fist candidate has no amino acid:
      res.add(new AcidWithPeak(spectrum.obs[index],null));
      Endpoint ep = prev;
      AminoAcid a = acid;
      while (ep != null) {
	res.add(new AcidWithPeak(spectrum.obs[ep.index],a));
	a = ep.acid;
	ep = ep.prev;
      }
      return new Proposal(res, score);
    }

  }

  /**
   * DenovoDP.EndpointQ, Utility subclass of TreeSet for easy
   * bookkeeping
   */
  private class EndpointQ {
    private TreeSet tree;
    private int maxRank;
    EndpointQ(int N) {
      maxRank = N;
      tree = new TreeSet();
    }
    void add(Endpoint tb) {
      if (tree.size() == maxRank) {
	Endpoint last = (Endpoint) tree.last();
	if (tb.compareTo(last) < 0) 
	  tree.remove(last);
	else 
	  return;
      }
      tree.add(tb);
    } 
    void addAll(Endpoint[] tbs) {
      if (tbs !=null)
	for (int i=0; i<tbs.length; i++)
	  add(tbs[i]);
    }
    Endpoint[] toArray() {
      if (tree.size() == 0) 
      return new Endpoint[0];
      Endpoint[] res = new Endpoint[tree.size()];
      tree.toArray(res);
      return res;
    }
    Iterator iterator() {
      return tree.iterator();
    }
  }
}
