
import java.util.*;

/**
 * Data about a single ion type
 *
 * The m/z of an ion is (M + offset)/charge, where M is the sum of
 * the constituent neutral amino acid residue masses
 */

//Note: there should be a reference to the Mascot web site and the
//article by Roepstorff et al.

class IonType {
  final public String abbr;
  final public double offset;
  final public double charge;
  final public boolean cterminal;
  final public String txt;               //For help screen
  final public ChemicalFormula formula;  //Of terminal group+modification
  private IonType(String abbr, double charge, boolean cterminal, 
		  ChemicalFormula formula, String txt) {
    this.abbr = abbr;
    this.charge = charge;
    this.offset = formula.monoW;
    this.cterminal = cterminal;
    this.formula = formula;
    this.txt = txt; 
  }
  public String toString() {
    return abbr;
  }
  private IonType doublecharged(String txt) {
    //assert charge==1;
    return new IonType(abbr+"++", 2, cterminal, formula.add("H"), txt);
  }
  private IonType translate(String abbr, String offset, String txt) {
    ChemicalFormula form = new  ChemicalFormula(offset);
    return new IonType(abbr, charge, cterminal, formula.add(form), txt);
  }
  public final static IonType Y = 
    new IonType("Y", 1, true, new ChemicalFormula("OH+H2"), "[C]+[M]+H2");
  public final static IonType B = 
    new IonType("B", 1, false, new ChemicalFormula("H"), "[N]+[M]");


  /**
   * Data about all ion types
   */
  private final static IonType A = B.translate("A", "-CO", "B - CO");
  private final static IonType[] allTypes = new IonType[] { 
    Y, 
    Y.translate("Y*", "-NH3", "Y - NH3"), //=Z
    Y.doublecharged("(Y+H)/2"),
    Y.translate("X", "CO-H2", "Y - H2 + CO"),
    B,
    B.translate("B*", "-NH3", "B - NH3"),
    B.doublecharged("(B+H)/2"),
    A,
    A.translate("A*","-NH3","A - NH3"),
    A.doublecharged("(A+H)/2"),
    B.translate("C", "NH3", "B + NH3"),
  };
  private final static HashMap /*from String to IonType*/ abbrToIonType;
  static {
    abbrToIonType = new HashMap();
    for (int i=0; i<allTypes.length; i++) 
      abbrToIonType.put(allTypes[i].abbr, allTypes[i]);
  }
  private static Collection /*of IonType*/ list =
    Collections.unmodifiableList(Arrays.asList(allTypes));

  public static Iterator iterator() {
    return list.iterator();
  }

  public static int getCount() {
    return allTypes.length;
  }

  public static IonType get(String abbr) {
    return (IonType)abbrToIonType.get(abbr);
  }

  //this is the (constant) differential kvotient of mapToY wrt. mz
  //public double getFactor() {
  //  return cterminal ? charge : -charge;
  //}

  //sum of constituent amino acids to ion m/z
  public double toMZ(double m) {
    return (m + offset)/charge;
  }
  public double fromMZ(double mz) {
    return mz * charge - offset;
  }

  public final static double terminalGroupMasses =  
    ChemicalFormula.OH.monoW +  // mass of C-terminal group
    ChemicalFormula.H.monoW;    // mass of N-terminal group

  //What is the m/z of the correspondig ion of another type? 
  /*
  private double corresponding(double mz, IonType newType, double totalMass) {
    //totalMass is the total mass of the neutral peptide: the sum of
    //the neutral amino acid residues plus the C-terminal OH group and
    //the N-terminal H group.
    if (cterminal == newType.cterminal)
      return newType.toMZ(fromMZ(mz));
    else 
      return newType.toMZ(totalMass - terminalGroupMasses - fromMZ(mz));
  }
  */
  public double mapFromY(double mz, double totalMass) {
    if (cterminal)
      return  toMZ(Y.fromMZ(mz));
    else
      return  toMZ(totalMass - terminalGroupMasses-Y.fromMZ(mz));
  }
  public double mapToY(double mz, double totalMass) {
    if (cterminal)
      return  Y.toMZ(fromMZ(mz));
    else
      return  Y.toMZ(totalMass - terminalGroupMasses-fromMZ(mz));
  }
}

