
import java.awt.*;
import java.text.*;
import java.util.*;
import javax.swing.*;


/**
 * A sequence of proposed sequences
 */

class Sequences {
  private /*final*/ static double gTolerance;   // in pixels squared
  private final static DecimalFormat fmt1 = new DecimalFormat("0.0");
  private final static DecimalFormat fmt3 = new DecimalFormat("0.000");

  // Using a linkedlist is OK because its length will be at most 20 
  private LinkedList /* of Sequence */ sequences;
  // Invariants: 
  // (1) no two sequences overlap, but possibly touch in one accepted peak;
  // (2) sequences are ordered by increasing leftEnd
  // (3) no sequence isEmpty()

  // Note: A peak may at the same time be an endpoint of one sequence
  // and a startpoint of another. If this is the case, the latter
  // marking must have taken place first. 

  // To have fast lookup whether a peak is valid in a doubled
  // spectrum. Maintained by fixup(), which is called by setPotential().
  IndexSet used_peaks;

  // To quickly position a mass relative to the sequences. Maintained
  // by fixup(). Searched all the time by mouseMotion handler, so for
  // performance reasons searching could remember last
  // position. (Actually, we will normally have at most two sequences
  // and in that case this optimization might be counterproductive).
  double[] endpoints;

  // Some rendering parameters inherited from SpectrumWindow

  private boolean showErrors;
  private Spectrum spectrum;
  private AbstractController controller;

  public Sequences(AbstractController controller) {
    this.controller = controller;
    sequences = new LinkedList();
    used_peaks = new IndexSet();
    fixup();
  }

  /**
   * Fixup used_peaks and endpoints[]
   */
  private void fixup() {
    used_peaks.clear();
    endpoints = new double[2*sequences.size()];
    Iterator iter = sequences.iterator();
    int i=0;
    while (iter.hasNext()) {
      Sequence seq = (Sequence)iter.next();
      seq.updateUsage();
      endpoints[i++] = seq.leftEnd();
      endpoints[i++] = seq.rightEnd();
    }
  }
  /**
   * Is the sequences list empty?
   */
  public boolean isEmpty() {
    return sequences.isEmpty();
  }
  /**
   * Empty sequences list!
   */
  public void clear() {
    sequences.clear();
    fixup();
  }
  

  //------------------------------------------------------------------------
  //Support for undo

  /**
   * Save copies of sequences for undo
   */
  public Object save() {
    LinkedList res = new LinkedList();
    Iterator iter = sequences.iterator();
    while (iter.hasNext()) 
      res.add(((Sequence)iter.next()).save());
    return res;
  }
  /**
   * Restore copies of sequences
   */
  public void restore(Object ll) {
    Iterator iter = ((LinkedList)ll).iterator();
    while (iter.hasNext()) {
      Sequence seq = new Sequence(used_peaks);
      seq.restore((LinkedList[])iter.next());
      sequences.add(seq);
    }
    //Caller will call setParams(...) and setPotential() etc.
  }

  //-------------------------------------------------------------------------
  //Utility methods for the position of masses/peaks relative to
  //sequences and the usage of peaks in sequences

  /**
   *  The position of a mass relative to the sequence coming from the
   *  side given by left.
   *
   *  The last indices found should be remembered for efficient
   *  searching. Oth. there should rarely be more than two sequences,
   *  so such an optimization might be superfluous.
   */
  double getDistance(double x, boolean left) {
    //Note: endpoints is always sorted. At most three consequtive
    //elements may be equal.
    int below = -1, above = -1;
    for (int i=0; i<endpoints.length; i++) {
      if (endpoints[i] == x) 
	return 0; //This can't happen
      if (endpoints[i] <= x)
	below = i;
      if (above<0 && endpoints[i] >= x)
	above = i;
    }
    //No hit, so we are strictly inside or outside:
    if ((left && above<0) || (!left && below<0)) 
      return 1.0/0.0; 
    if (left)
      return above%2==1 ? 0 : endpoints[above]-x;
    else
      return below%2==0 ? 0 : x-endpoints[below];
      
  }

  /**
   *  Find the index in 0..sequences.size() where mass x should be
   *  inserted; the least i s.t. i==sequences.size() or
   *  x<sequences.get(i).leftEnd()
   */
  private int indexOf(double x) {
    int i = 0;
    while (i < sequences.size() 
	   && x >= ((Sequence)sequences.get(i)).leftEnd())
      i++;
    return i;
  }

  /**
   *  Return the sequence if p is inside one, null else: Consider using
   *  endpoints[].
   */
  private Sequence inside(double x) {
    int i = 0;
    while (i < sequences.size() 
	   && x >= ((Sequence)sequences.get(i)).leftEnd())
      i++;
    if ( i == 0 )
      return null;
    Sequence res = (Sequence)sequences.get(i-1);
    return res.rightEnd()<x ? null : res;
  } 

  /**
   *  The neighbour sequence having x strictly to the side given by
   *  left. Return null if x is deep inside a sequence or if no such
   *  sequence exists.  Consider using endpoints[]
   */
  private Sequence neighbourSequence(double x, boolean left) {
    if (left) {
      int i = 0;
      while (i < sequences.size()) {
	Sequence seq = (Sequence)sequences.get(i);
	if (x < seq.leftEnd())
	  return seq;
	if (x < seq.rightEnd())
	  return null;
	i++;
      }
      return null;
    }
    else { 
      int i = sequences.size()-1;
      while (i > -1 ) {
	Sequence seq = (Sequence)sequences.get(i);
	if (x > seq.rightEnd())
	  return seq;
	if (x > seq.leftEnd())
	  return null;
	i--;
      }
      return null;
    }
  }

  //This method uses the marks as delimiters!
  public boolean inGap(double x) {
    int i = 0;
    while (i < sequences.size() 
	   && x >= ((Sequence)sequences.get(i)).leftEndMark())
      i++;
    if ( i == 0 || i == sequences.size())
      return false;
    Sequence res = (Sequence)sequences.get(i-1);
    return res.rightEndMark() < x;
  }

  public void fillGap(double x) {
    //First repeat inGap():
    int i = 0;
    while (i < sequences.size() 
	   && x >= ((Sequence)sequences.get(i)).leftEndMark())
      i++;
    if ( i == 0 || i == sequences.size())
      return;
    Sequence res = (Sequence)sequences.get(i-1);
    if (res.rightEndMark() >=x)
      return;
    //Forget the supplied x and compute left endpoint and length of gap:
    x = res.rightEndMark();
    double xprim = ((Sequence)sequences.get(i)).leftEndMark();
    double w = xprim - x;
    if (w>500) {
      controller.debugmessage("Gap too large (>500)");
      return;
    }
    Collection aaLists = 
      AminoAcid.getMatches(w, controller.seqParms.tolerance);
    Iterator iter = aaLists.iterator();
    StringBuffer deb = 
      new StringBuffer("Gap from "+ fmt3.format(x) + " to " +
		       fmt3.format(xprim) + " is matched by " +
		       aaLists.size() + " possible peptides:\n" +
		       "(numbers are: #matched by Y, #matched by B, "+
		       "#other matches\n");
    while (iter.hasNext())
      deb.append(spectrum.fillGap(x, (Collection)(iter.next()), 
				 controller.seqParms.tolerance, 
				 controller.seqParms.grass)+"\n");
    controller.debugmessage(deb.toString());
  }


  /**
   * Which sequences use p and for what?
   */
  public LinkedList peakUsage(Peak p) {
    Iterator iter = sequences.iterator();
    LinkedList res = new LinkedList();
    while (iter.hasNext()) {
      Mark mark = ((Sequence)iter.next()).getMark(p);
      if (mark != null)
	res.add(mark);
    }
    return res;
  }

  //Is p marked as start of a sequence? Return null if not, the
  //Triangle if so.
  public Mark isStart(Peak p) {
    Iterator iter = sequences.iterator();
    while (iter.hasNext()) {
      Mark mark = ((Sequence)iter.next()).getMark(p);
      if (mark != null)
	return mark;
    }
    return null;
  }

  //------------------------------------------------------------------------
  //Mirror support

  /**
   * Remove marks put on peaks with kind=MIRRORED
   */
  public void pruneMirroredPeaks() {
    Iterator iter = sequences.iterator();
    while (iter.hasNext()) { 
      Sequence seq = (Sequence)iter.next();
      seq.pruneMirroredPeaks();
      if (seq.isEmpty())
	iter.remove();
    }
  }

  /**
   *  Remove any marks set on peak p. (Should be implemented using
   *  peakUsage() method)
   */
  public void maybeRemovePeak(Peak p) {
    boolean found = false;
    Iterator iter = sequences.iterator();
    while (iter.hasNext()) { 
      Sequence seq = (Sequence)iter.next();
      found = seq.maybeRemovePeak(p) || found;  
      if (seq.isEmpty())
	iter.remove();
    }
  }


  //--------------------------------------------------------------------------
  //To start new sequences

  /**
   * Can one start a new sequence at p: possible obstructions are: (1)
   * p is inside an already constructed sequence; (2) p has been
   * marked as ignored; (3) the mirrored version of p is already used
   * and duplicates are not allowed.
   */
  public boolean canStartNew(Peak p) {
    return 
      inside(p.x) == null && 
      !p.ignore && 
      !(spectrum.isMirrored && 
	!controller.seqParms.allowDuplicatePeaks && 
	used_peaks.contains(p.orig_index));
  }

  /**
   * Start a new sequence at p
   */
  public Sequence startNew(Peak p) {
    Sequence seq = null;
    if (canStartNew(p)) {
      seq = new Sequence(used_peaks);
      sequences.add(indexOf(p.x), seq);
      seq.setStart(p);
    }
    return seq;
  }
  
  //-----------------------------------------------------------------------
  //To follow a computed candidate/proposal

  /**
   * May one follow the candidate from p in direction given by left?
   *
   * Allowable if p is in the candidate but either not inside a
   * sequence (and allowed for adding) or the correct endpoint of one
   */
  public boolean canFollow(Peak p, Proposal candidate, boolean left) {
    Sequence seq = inside(p.x);
    boolean inCandidate = false;
    if (canStartNew(p) 
	|| p.x == (left ? seq.leftEnd() : seq.rightEnd())) {
      if (candidate != null) {
	Iterator iter = candidate.iterator();
	while (!inCandidate && iter.hasNext()) 
	  if (((AcidWithPeak) iter.next()).peak == p)
	    inCandidate = true;
      }
    }
    return inCandidate;
  }

  /**
   * Follow the candidate from p in direction given by left! Do not go
   * further than the next sequence.
   *
   * Note that the followed sequence might not be constructible using
   * mouse clicks, if the total mass error is larger than tolerance.
   */
  public void follow(Peak p, Proposal candidate, boolean left) {
    //assert canFollow(p,candidate,left)
    Sequence seq = inside(p.x);
    if (seq==null)
      seq = startNew(p);
    double mass = left ? seq.leftEndMark() : seq.rightEndMark();
    //How far can we go?
    double limit = left ?  Double.NEGATIVE_INFINITY : Double.POSITIVE_INFINITY;
    Sequence nextseq = left ? seq.left() : seq.right();
    if (nextseq != null)
      limit = left ? nextseq.rightEnd() : nextseq.leftEnd();
    //Seek to the point in candidate where p is
    boolean found = false;
    Iterator iter = candidate.iterator();
    while (!found && iter.hasNext()) 
      if (((AcidWithPeak)iter.next()).peak == p)
	found = true;
    //Now, add the candidates
    Mark oldmark = (Mark)(left ? seq.acceptedL.getFirst() : 
			  seq.acceptedR.getFirst());
    Peak oldpeak = oldmark.peak;
    double oldscore = oldmark.score;
    while (iter.hasNext()) {
      AcidWithPeak awp = (AcidWithPeak) iter.next();
      Peak p2 = awp.peak;
      AminoAcid acid = awp.acid;
      mass += (left ? -1 : 1) * acid.monoW;
      if ((left && p2.x < limit) || (!left && p2.x > limit))
	break;
      AAMark nm = 
	new AAMark(acid, Mark.POTENTIAL, mass, p2.y, p2, seq, left);
      oldscore = nm.score = 
	oldscore + seq.seqScore.stepscore(acid, mass-p2.x, oldpeak, p2);
      nm.accept();
      oldpeak = p2;
    }
  }
  
  //------------------------------------------------------------------------  
  // View routines

  /**
   * paint just paints all sequences
   */
  public void paint(Graphics2D g, boolean isPrinting,
		    AbstractCoordinates coor) {
    Iterator iter = sequences.iterator();
    while (iter.hasNext()) 
      ((Sequence)iter.next()).paint(g, isPrinting, coor);
  }

  /**
   * show sequence in the seqOut textField and write to log
   */
  public void showSeq(JTextField seqOut, Log log) {
    final double epsilon = 0.1;
    double rightEnd = 0;
    int size = 0;
    double error = 0.0;
    StringList lst = new StringList();
    Iterator iter = sequences.iterator();
    while (iter.hasNext()) {
      Sequence seq = (Sequence)iter.next();
      int seqsize = seq.size();
      if (seqsize > 0) {
	double hole = seq.leftEndMark() - rightEnd;
	if (hole > epsilon || hole < -epsilon) 
	  lst.add("[" + fmt1.format(hole) + "]");
	rightEnd = seq.rightEndMark();
	lst.addAll(seq.getAccepted());
	size += seqsize;
	error += seq.error();
      }
    }
    StringBuffer sbuf = lst.toStringBuffer(controller.bwdSequence);
    if (size > 0) 
      sbuf.append(" (rms = " + fmt3.format(Math.sqrt(error/size)) + ")");
    String res = sbuf.toString();
    seqOut.setText(res);
    log.write(res);
  }

  /**
   * Find mass error of nearest potential aa mark
   */
  public double delta(int gx, int gy, AbstractCoordinates coor) {
    double delta = Double.NEGATIVE_INFINITY;
    Iterator iter = sequences.iterator();
    while (delta == Double.NEGATIVE_INFINITY && iter.hasNext()) 
      delta = ((Sequence)iter.next()).delta(gx, gy, coor);
    return delta;
  }

  /**
   * Write the contents of sequences to the debug window
   */
  public void debugout() {
    Iterator iter = sequences.iterator();
    StringBuffer out = new StringBuffer();
    while (iter.hasNext()) 
      ((Sequence)iter.next()).debugout(out);
    controller.debugmessage(out.toString());
  }

  //-------------------------------------------------------------------
  // 

  /**
   * Copy the current rendering parameters (called from SpectrumWindow)
   */
  public void setParams(double gTolerance, 
			boolean showErrors, 
			Spectrum spectrum) {
    this.gTolerance = gTolerance; 
    this.showErrors = showErrors;
    this.spectrum = spectrum;
  }

  /**
   * Recompute potential amino acids for all subsequences (a.o.)
   */
  public void setPotential() {
    fixup();
    //the real setPotential
    Iterator iter = sequences.iterator();
    while (iter.hasNext()) 
      ((Sequence)iter.next()).setPotential();
  }

  //--------------------------------------------------------------------
  //Support for finding

  /**
   * compare marks according to mass
   */
  private Comparator markComparator = new Comparator() {
      public int compare (Object o1, Object o2) {
	Mark m1 = (Mark)o1;
	Mark m2 = (Mark)o2;
	if (m1.x<m2.x)
	  return 1;
	else if (m1.x>m2.x)
	  return -1;
	else 
	  return 0;
      }
      public boolean equals (Object o1, Object o2) {
	return compare(o1,o2)==0;
      }
    };
  
  /**
   * Marks near a mouse position ordered by mass
   */
  public Mark[] nearbyMarks(int gx, int gy, AbstractCoordinates coor) {
    ArrayList out = new ArrayList();
    Iterator iter = sequences.iterator();
    while (iter.hasNext()) 
      out.addAll(((Sequence)iter.next()).nearbyMarks(gx, gy, coor));
    Mark[] res = new Mark[out.size()];
    out.toArray(res);
    Arrays.sort(res,markComparator);
    return res;
  }

  /**
   * Find nearest mark
   */
  public Mark nearestMark(int gx, int gy, AbstractCoordinates coor,
			  boolean left) {
    Mark nearest = null;
    //We iterate over all sequences instead of using inside() and
    //nearestSequence. The reason is that we have the coordinates as
    //screen coordinates.
    Iterator iter = sequences.iterator();
    while (nearest == null && iter.hasNext()) 
      nearest = ((Sequence)iter.next()).nearestMark(gx, gy, coor, left);
    return nearest;
  }
  /**
   * A sequence (should be SortedSet!) of acceptable marks for the
   * nearest sequence having p to the side given by parameter
   * left. Since it is only used in generating actions for the popup
   * menu handler, it might as well generate the actions, hiding the
   * methods of AAMark. 
   */
  public LinkedList /* of AAMark */ acceptableMarks(Peak p, boolean left) {
    Sequence seq = neighbourSequence(p.x, left);
    return seq == null ? new LinkedList() : seq.acceptableMarks(p,left);
  } 

  // ----------------------------------------------------------------------

  /**
   * Inner class: Sequences.Sequence One proposed sequence, with
   * accepted and potential aa
   */

  // A proposed sequence is represented by 
  //  * Two lists (really, stacks) of left-hand and right-hand accepted 
  //    marks; each is a red start mark or a green amino acid mark
  //  * Two sets of (blue) left-hand and right-hand AAMarks, each 
  //    denoting a potential amino acid

  //LinkedList list = left ? seq.acceptedL : seq.acceptedR;

  class Sequence implements AbstractSequence {
    // Accepted amino acids, incl start; ordered left-right and right-left
    private LinkedList /* of Mark */ acceptedL, acceptedR;
    
    public LinkedList getAccepted(boolean left) {
      return left ? acceptedL : acceptedR;
    }
    
    // Potential amino acids
    private HashSet /* of AAMark */ potentialL, potentialR;		
    // Invariants: 
    //  (1) acceptedL.isEmpty() == acceptedR.empty()
    //  (2) (acceptedL.contains(mark) && acceptedR.contains(mark)) 
    //      == (mark instanceof Triangle)

    //reference to the used_peaks set in our parent Sequences
    private IndexSet used_peaks;

    //For the moment, we let the score start at the sequence
    //start. One might as well just start anew all the time. The way
    //it is done now means that the score used to score potential
    //peaks is the one that was current when the endpoint was created
    //- not the score corresponding to current seqParms. This might be
    //considered to be a feature.
    private Score seqScore;

    /**
     * Only constructor
     */
    public Sequence(IndexSet used_peaks) {
      acceptedL = new LinkedList();
      acceptedR = new LinkedList();
      potentialL = new HashSet();
      potentialR = new HashSet();
      this.used_peaks = used_peaks;
    }


    //---------------------------------------------------------------------
    //Undo

    /**
     * Support for undo
     */
    LinkedList[] save() {
      LinkedList rest = new LinkedList();
      rest.add(seqScore);
      return new LinkedList[] {(LinkedList)acceptedL.clone(), 
				 (LinkedList)acceptedR.clone(),
				 rest };
    }
    /**
     * Support for undo
     */
    void restore(LinkedList[] lla) {
      acceptedL = restore(lla[0]);
      acceptedR = restore(lla[1]);
      seqScore = (Score)(lla[2].getFirst());
      //Caller will call setPotential etc.
    }
    /**
     * Fix the backreference in each mark:
     */
    private LinkedList restore(LinkedList ll) {
      Iterator iter = ll.iterator();
      while (iter.hasNext()) 
	((Mark)iter.next()).seq = this;
      return ll;
    }

    //-----------------------------------------------------------------------
    //Bookkeeping

    /**
     * Time to die (only public because it is called from
     * Triangle.remove(). One should probably only empty the sequence
     * in Triangle.remove() and then make sure Sequences.fixup() is
     * called by whatever invoked Triangle.remove(). In an ideal
     * world, class Mark would be invisible outside Sequences, and we
     * could package this more cleanly.
     */
    public void remove() {
      sequences.remove(this);
    }
    
    //This is called on every call to setPotential() i.e. all the
    //time. One might want to subclass LinkedList so that additions
    //and removals were directly reflected in the IndexSet. In that
    //case, IndexSet should contain reference counts (if a peak can
    //belong to more than one sequence) e.g. be a wrapper for TreeMap.
    //Or a heap?
    public void updateUsage() {
      //Assume this is only called after a total cleaup of used_peaks
      updateUsage(true);
      updateUsage(false);
    }
    private void updateUsage(boolean left) {
      LinkedList list = left ? acceptedL : acceptedR;
      if (!list.isEmpty()) {
	Iterator iter = list.iterator();
	while (iter.hasNext()) 
	  used_peaks.add(((Mark)iter.next()).peak.orig_index);
      }
    }

    //------------------------------------------------------------------------
    //Usage of peaks 

    /**
     * Return the mark if p is used, else return null
     */
    //Note that p can only be in both of acceptedX if it is the start
    public Mark getMark(Peak p) {
      Mark mark = getMark(p,true);
      if (mark == null)
	mark = getMark(p, false);
      return mark;
    }

    private Mark getMark(Peak p, boolean left) {
      LinkedList list = left ? acceptedL : acceptedR;
      Mark mark = null;
      boolean found = false;
      if (!list.isEmpty()) {
	Iterator iter = list.iterator();
	while (!found && iter.hasNext()) {
	  mark = (Mark)iter.next();
	  found = mark.peak == p;
	}
      }
      return found ? mark : null;
    }

    //------------------------------------------------------------
    //Mirror support

    /**
     * Called after removing the mirrored peaks from the spectrum
     */
    public void pruneMirroredPeaks() {
      pruneMirroredPeaks(true);
      pruneMirroredPeaks(false);
    }
    private void pruneMirroredPeaks(boolean left) {
      boolean found = false; 
      LinkedList list = left ? acceptedL : acceptedR;
      if (!list.isEmpty()) {
	//Find mirrored peak nearest to the start
	Iterator iter = list.iterator();
	Mark nearestmark = null;
	Mark mark;
	while (iter.hasNext()) {
	  mark = (Mark)iter.next();
	  if (mark.peak.kind == Peak.MIRRORED) 
	    nearestmark = mark;
	}
	//prune it 
	//
	//we do not just do mark.remove() because that would interfere
	//with the iterator in the caller.
	if (nearestmark != null)
	  do {
	    mark = (Mark)list.removeFirst();
	  } while (mark != nearestmark);
      }
    }
    
    /**
     * Called after removing a single peak (and adding its mirror)
     */
    public boolean maybeRemovePeak(Peak p) {
      Mark mark = getMark(p);
      if (mark != null) {
	if (mark.kind == Mark.START) {
	  acceptedL.clear();
	  acceptedR.clear();
	}
	else {
	  LinkedList list = ((AAMark)mark).left ? acceptedL : acceptedR;
	  Mark m;
	  do {
	    m = (Mark)list.removeFirst();
	  } while (m != mark);
	}
	return true;
      }
      return false;
    }


    // Expand potentialX to a list (ought to be SortedSet) of all
    // allowable markings of this peak 

    //PROBLEM: potentialX may avoid an acid if the "true" mass is
    //inside another sequence, even if a nearby peak is allowable for
    //that acid and outside! We could simply iterate over the acids,
    //computing mark.x etc.

    public LinkedList /* of AAMark */ acceptableMarks(Peak p, boolean left) {
      LinkedList res = new LinkedList();
      HashSet set = left ? potentialL : potentialR;
      Iterator iter = set.iterator();
      Peak olepeak = ((Mark)(left ? acceptedL.getFirst() : 
			    acceptedR.getFirst())).peak;
      double score = ((Mark)(left ? acceptedL.getFirst() : 
			    acceptedR.getFirst())).score;
      while (iter.hasNext()) { 
	AAMark mark = (AAMark)iter.next();
	if (Math.abs(mark.x - p.x) < controller.seqParms.tolerance) {
	  //NBNBNBNB Add test for mirrorred peak
	  AAMark m = new AAMark(p, mark);
	  //NBNBNB problems with acid index
	  m.score = score + seqScore.stepscore(mark.acid,mark.x-p.x,olepeak,p);
	  res.add(m);
	}
      }
      return res;
    } 

    
    public void setStart(Peak p) {
      Mark startPoint = new Triangle(p.x, p.y, p, this);
      acceptedL.add(startPoint);
      acceptedR.add(startPoint);
      seqScore = new SeqScore(spectrum.maxAbundance,
				 spectrum.totalMass,
				 controller.seqParms);
      startPoint.score = seqScore.peakscore(startPoint.peak);
    }

    /**
     * Set the left and right sets of potential amino acids
     */
    public void setPotential() {
      if (acceptedL.isEmpty())
	potentialL.clear();
      else
	setPotential(potentialL, -1, (Mark)acceptedL.getFirst());
      if (acceptedR.isEmpty())
	potentialR.clear();
      else
	setPotential(potentialR, +1, (Mark)acceptedR.getFirst());
    }

    /**
     * Find potential amino acid peaks (side=-1 if left, +1 if right)
     */
    private void setPotential(Set potential, int side, Mark m) {
      double x = m.x;
      double score = m.score;
      potential.clear();
      double dist = Double.POSITIVE_INFINITY;
      AAMark nearest = null;
      double limit;
      double tolerance = controller.seqParms.tolerance;
      double grass = controller.seqParms.grass;
      Peak olepeak = null; 

      if (side < 0) {		// add potential amino acids on the left
	Sequence left = left();
	limit = left != null ? left.rightEnd() : Double.NEGATIVE_INFINITY;
	olepeak = ((Mark)acceptedL.getFirst()).peak;
      } else {			// add potential amino acids on the right
	Sequence right = right();
	limit = right != null ? right.leftEnd() : Double.POSITIVE_INFINITY;
	olepeak = ((Mark)acceptedR.getFirst()).peak;
      }
      Iterator iter = AminoAcid.aaByMonoW.iterator();
      double globalMaxScore = Double.NEGATIVE_INFINITY;
      while (iter.hasNext()) {
	AminoAcid aa = (AminoAcid)iter.next();
	double weight = x + side * aa.monoW;
	double wmin = weight-tolerance;
	double wmax = weight+tolerance;
	if (side<0)
	  wmin = Math.max(wmin,limit);
	else
	  wmax = Math.min(wmax,limit);
	int imin = spectrum.getIndex(wmin);
	int imax = spectrum.getIndex(wmax);
	Peak[] obs = spectrum.obs;
	int nearestBar = -1;
	double maxScore = Double.NEGATIVE_INFINITY;
	double bestScore = Double.NEGATIVE_INFINITY;
	for (int i=imin; i<imax+1; i++) {
	  if (obs[i].y < grass * spectrum.maxAbundance ||
	      obs[i].ignore ||
	      obs[i].x <wmin || obs[i].x > wmax || //Superfluous
	      (spectrum.isMirrored && !controller.seqParms.allowDuplicatePeaks
	       && used_peaks.contains(obs[i].orig_index)))
	    continue;
	  double newScore = 
	    score + seqScore.stepscore(aa,weight-obs[i].x, olepeak, obs[i]);
	  if (newScore > maxScore) {
	    maxScore = newScore;
	    bestScore = newScore;
	    nearestBar = i;
	  }
	}
	if (nearestBar>=0
	    && (controller.seqParms.allowKR || //At end
		(x < Peak.MAX_END_OFFSET 
		 || obs[nearestBar].x < Peak.MAX_END_OFFSET 
		 ) ||
		!(aa.abbr.equals("K") || aa.abbr.equals("R")))) {
	  AAMark aamark = new AAMark(aa, Mark.POTENTIAL, weight, 
				     obs[nearestBar].y, obs[nearestBar],
				     this, side<0);
	  aamark.score = bestScore;
	  potential.add(aamark);
	  if (maxScore > globalMaxScore) {
	    globalMaxScore = maxScore;
	    nearest = aamark;
	  }
	}
      }
      // Mark best potential amino acid in red
      if (nearest != null) 
	nearest.setBest(true);
    }

    //----------------------------------------------------------------
    // View support

    /**
     * Return the sequence of accepted amino acids as a StringList 
     */
    public StringList getAccepted() {
      StringList lst = new StringList();
      Iterator iterL = acceptedL.iterator();
      while (iterL.hasNext()) {
	Mark mark = (Mark)iterL.next();
	lst.add(mark.getAbbr());
      }
      ListIterator iterR = acceptedR.listIterator(acceptedR.size());
      while (iterR.hasPrevious()) {
	Mark mark = (Mark)iterR.previous();
	lst.add(mark.getAbbr());
      }
      return lst;
    }

    /**
     * Compute the error of this sequence
     */
    public double error() {
      double distsum = 0.0;
      Iterator iterL = acceptedL.iterator();
      while (iterL.hasNext()) {
	Mark mark = (Mark)iterL.next();
	distsum += sqr(mark.error());
      }
      ListIterator iterR = acceptedR.listIterator(acceptedR.size());
      while (iterR.hasPrevious()) {
	Mark mark = (Mark)iterR.previous();
	distsum += sqr(mark.error());
      }
      return distsum;
    }

    /**
     * Compute the size (number of aa) of this sequence
     */
    public int size() {
      return acceptedL.size() + acceptedR.size() - 2;
    }

    /** 
     * Find the sequence to the left resp. right of this one, or null
     */
    private Sequence left() {
      int i = sequences.indexOf(this);
      return i > 0 ? (Sequence)sequences.get(i-1) : null;
    }

    private Sequence right() {
      int i = sequences.indexOf(this);
      return 0 <= i && i < sequences.size()-1 ? 
	(Sequence)sequences.get(i+1) 
	: null;
    }

    /**
     * Format the sequence for the debug window
     */
    public void debugout(StringBuffer out) {
      Iterator iterL = acceptedL.iterator();
      while (iterL.hasNext()) {
	Mark mark = (Mark)iterL.next();
	out.append(mark.getAbbr() + " " + 
		   fmt3.format(mark.x) + " " + 
		   fmt3.format(mark.error()) +"\n");
      }
      ListIterator iterR = acceptedR.listIterator(acceptedR.size());
      while (iterR.hasPrevious()) {
	Mark mark = (Mark)iterR.previous();
	out.append(mark.getAbbr() + " " + 
		   fmt3.format(mark.x) + " " + 
		   fmt3.format(mark.error()) +"\n");
      }
    }

    //--------------------------------------------------------------------
    //Position marks relative to mouse position

    /**
     * Marks near a given point in window coordinates
     */
    private void nearbyMarks(HashSet out, Collection coll, int gx, int gy, 
			    AbstractCoordinates coor) {
      Iterator iter = coll.iterator();
      while (iter.hasNext()) {
	Mark m = (Mark)iter.next();
	if (coor.dist(gx, gy, m.x, m.y) < gTolerance) 
	  out.add(m);
      }
    }
    /**
     * Marks near a given point in window coordinates
     */
    public HashSet nearbyMarks(int gx, int gy, 
			       AbstractCoordinates coor) {
      HashSet res = new HashSet();
      nearbyMarks(res, acceptedL, gx, gy, coor);
      nearbyMarks(res, acceptedR, gx, gy, coor);
      nearbyMarks(res, potentialL, gx, gy, coor);
      nearbyMarks(res, potentialR, gx, gy, coor);
      return res;
    }

    /**
     * The mark closest to a mouse position
     */
    private Mark nearestMark(Collection coll, int gx, int gy, 
			     AbstractCoordinates coor) {
      Mark nearest = null; 
      double dist = gTolerance;
      Iterator iter = coll.iterator();
      while (iter.hasNext()) {
	Mark mark = (Mark)iter.next();
	double idist = coor.dist(gx, gy, mark.x, mark.y);
	if (idist < dist) {
	  dist = idist;
	  nearest = mark;
	}
      }
      return nearest;
    }

    /**
     * Find nearest mark to the given side 
     */
    public Mark nearestMark(int gx, int gy, AbstractCoordinates coor, boolean left) {
      Mark nearest = null;
      if (left) {
	nearest= nearestMark(acceptedL, gx, gy, coor);
	if (nearest == null)
	  nearest = nearestMark(potentialL, gx, gy, coor);
      }
      else {
	nearest= nearestMark(acceptedR, gx, gy, coor);
	if (nearest == null)
	  nearest = nearestMark(potentialR, gx, gy, coor);
      }
      return nearest;
    }

    /**
     * Find mass error of nearest potential aa mark; negative if none
     */
    public double delta(int gx, int gy, AbstractCoordinates coor) {
      Mark nearest = nearestMark(potentialL, gx, gy, coor);
      if (nearest == null)
	nearest = nearestMark(potentialR, gx, gy, coor);
      if (nearest != null) 
	return nearest.x - nearest.peak.x;
      else
	return Double.NEGATIVE_INFINITY;
    }

    // Return true if sequence has no start aa nor accepted aa 
    
    public boolean isEmpty() {
      // assert acceptedL.isEmpty() == acceptedR.isEmpty();
      return acceptedL.isEmpty();
    }
    
    // Find mass of left and right end peak of sequence (+inf and -inf
    // if empty)
    
    public double leftEnd() {
      return ((Mark)acceptedL.getFirst()).peak.x;
    }
    
    public double rightEnd() {
      return ((Mark)acceptedR.getFirst()).peak.x;
    }
    
    // Find mass of left and right end mark of sequence (+inf and -inf
    // if empty)

    public double leftEndMark() {
      return ((Mark)acceptedL.getFirst()).x;
    }
    
    public double rightEndMark() {
      return ((Mark)acceptedR.getFirst()).x;
    }
    
    /**
     * Draw the accepted and potential amino acids
     */
    public void paint(Graphics2D g, boolean isPrinting,
		      AbstractCoordinates coor) {
      paint(acceptedL,  g, coor);
      paint(acceptedR,  g, coor);
      if (!isPrinting) {
	paint(potentialL, g, coor);
	paint(potentialR, g, coor);
      }
    }
    /**
     * Draw the amino acids of Collection coll
     */
    private void paint(Collection /* of Mark */ coll, Graphics2D g, 
		      AbstractCoordinates coor) {
      Iterator iter = coll.iterator();
      while (iter.hasNext()) 
	((Mark)iter.next()).draw(g, coor, controller.seqParms.tolerance,
				 showErrors);
    }
  }

  private static double sqr(double x) {
    return x * x;
  } 
}



/**
 * IndexSet 
 *
 * Utility wrapper for HashSet. 
 */

class IndexSet {
  //We could use a heap? (Cormen et al chap7) 
  //
  //For the used_peaks field of Sequences, a boolean[] indexed by
  //orig_index would be fine, but that would not be fine for the
  //dynamic programming, where we might want to attach this beast to
  //every TraceBackPeak
  private HashSet ts = new HashSet();
  boolean add(int i) {
    return ts.add(new Integer(i));
  }
  boolean contains(int i) {
    return ts.contains(new Integer(i));
  }
  boolean remove(int i) {
    return ts.remove(new Integer(i));
  }
  void clear() {
    ts.clear();
  }
  public String toString() {
    return ts.toString();
  }
}
