

import java.io.*;
import java.text.*;
import java.util.*;

/**
 * The fundamental data
 */
class Spectrum {
  private final static DecimalFormat fmt1 = new DecimalFormat("0.0");
  private final static DecimalFormat fmt3 = new DecimalFormat("0.000");

  String name;
  double totalMass; 
  int chargeState;
  Peak[] obs;
  int origPeakCount;
  double maxAbundance;
  boolean isMirrored;

  TreeSet /* of ITA */ ionTypeAlternatives;
  
  public Spectrum(String name, 
		  int origPeakCount,
		  double totalMass,
		  int chargeState, 
		  Collection obs) {
    this.name = name;
    this.totalMass = totalMass;
    this.chargeState = chargeState;
    this.origPeakCount = origPeakCount;
    //
    //We need the maxAbundance before we can add the bot and top
    //artificial peaks
    maxAbundance = Double.NEGATIVE_INFINITY;
    Iterator iter = obs.iterator();
    while (iter.hasNext()) {
      Peak p = (Peak) iter.next();
      maxAbundance = Math.max(maxAbundance, p.y);
    }
    //
    TreeSet sortedObs = new TreeSet(obs);

    double bottom_synthetic = IonType.Y.toMZ(0);
    double top_synthetic = IonType.B.mapToY(IonType.B.toMZ(0),totalMass);
    sortedObs.add(new Peak(bottom_synthetic,
			   bottom_synthetic,
			   maxAbundance/5,
			   Peak.BOTTOM_OI,
			   Peak.SYNTHETIC_BOTTOM));
    sortedObs.add(new Peak(top_synthetic,
			   top_synthetic,
			   maxAbundance/5,
			   Peak.TOP_OI,
			   Peak.SYNTHETIC_TOP));

    this.obs = (Peak[])sortedObs.toArray(new Peak[] { });
  
    //
    setLocalAbundanceLevels(30);

    //Finally, the ionType alternatives: 
    setITAs();
  }

  public String getDescription() {
    StringBuffer res = 
      new StringBuffer("<html><head>>" +
		       "<style>td {font-size : 12pt;}</style>" +
		       "</head><body bgcolor=white text=black><tt>" +
		       "<table border=0 cellspacing=1 cellpadding=1>");
    
    res.append("<tr><td>File name<td>&nbsp;<td>" + name);
    res.append("<tr><td>Total mass<td><td>" + fmt3.format(totalMass));
    res.append("<tr><td>Charge state<td><td>" + chargeState);
    res.append("<tr><td>Peaks in file<td><td>" + origPeakCount);
    res.append("<tr><td>Max abundance <td><td>" + maxAbundance);
    
    res.append("</table></tt></body></html>");
    return res.toString();
  }
  


  private void setITAs() {
    //This code typically takes 100-200ms on my workstation 
    ionTypeAlternatives = new TreeSet();
    Iterator iter = IonType.iterator();
    while (iter.hasNext()) {
      IonType iontype = (IonType) iter.next();
      for (int i = 0; i< this.obs.length; i++) {
	if (this.obs[i].orig_index >=0) {
	  double w = iontype.mapToY(this.obs[i].x_orig, totalMass);
	  ionTypeAlternatives.add(new ITA(w, iontype, this.obs[i]));
	}
      }
    }

  }


  //This was an attempt at estimating totalMass more precisely from
  //the spectrum
  public void tmtest() {
    int i = -1;
    int highindex = obs.length-1; 
    int count = 0;
    double tol = 0.7; 
    double[] z = new double[1001];
    while (++i <= highindex) {
      if (obs[i].orig_index<0) 
	continue;
      double limit = totalMass+2*ChemicalFormula.H.monoW-obs[i].x+tol;
      while (obs[highindex].x > limit)
	highindex--;
      if (i>highindex)
	break;
      limit = totalMass+2*ChemicalFormula.H.monoW-obs[i].x-tol;
      int j = highindex;
      while (obs[j].x >= limit) {
	int x = (int)(z.length*((obs[j].x+obs[i].x-2*ChemicalFormula.H.monoW)-
				(totalMass-tol))/(2*tol));
	if (x<0 || x>=1000) {
	  //System.out.println("Line 3581: "+x+" "+i+" "+j);
	}
	else
	  z[x] += obs[j].y*obs[i].y;//Math.max(0,Math.log(obs[j].y*obs[i].y));
	//out.println((obs[j].x+obs[i].x)+" "+Math.log(obs[j].y*obs[i].y));
	j--; //limit will never be less than totalMass/2-tol
      }
    }
    //
    System.out.println(""+totalMass);
    try {
      PrintWriter out = new PrintWriter(new FileOutputStream("a.dat"));
      for (int x=0; x<z.length; x++) 
	out.println((x*((2*tol)/z.length) + totalMass-tol)+" "+z[x]);
      out.close();
    }
    catch(FileNotFoundException ex) {};
  }


  //getIonTypeAlts() returns the ITAs that are within tolerance of w.
  //
  //Note: the check for grass here there and everywhere should be
  //changed to a "boolean significant_abundance" field on Peak that
  //can involve grass and the local abundance/background levels.
  //
  //Perhaps there should be a version taking the already constructed
  //sequences into account: in fact utilizing the used_peaks field in
  //Sequences.
  public Collection /*of ITA*/ getIonTypeAlts(double w, 
					      double tolerance,
					      double grass) {
    LinkedList res = new LinkedList();
    Peak fakelow = new Peak(0,0,0,-1,0);
    Peak fakehigh = new Peak(0,0,0,obs.length,0);
    SortedSet abe = 
      ionTypeAlternatives.subSet(new ITA(w-tolerance,null,fakelow),
				 new ITA(w+tolerance,null,fakehigh));
    Iterator iter = abe.iterator();
    while (iter.hasNext()) {
      ITA ita = (ITA)(iter.next());
      double rel_a = ita.p.y/maxAbundance;
      if (rel_a >= grass) 
	res.add(ita);
    }
    return res;
  }
  public String getIonTypeAltsFmt(double w, double tolerance, double grass) {
    StringBuffer res = new StringBuffer();
    Collection coll = getIonTypeAlts(w, tolerance, grass);
    Iterator iter = coll.iterator();
    boolean foundSome = false;
    double rms = 0.0;
    double total_abundance = 0.0;
    //Prune duplicate names?
    while (iter.hasNext()) {
      ITA ita = (ITA)(iter.next());
      double rel_a = ita.p.y/maxAbundance;
      if (rel_a >= grass) {
	if (foundSome)
	  res.append(",");
	else 
	  res.append(" (");
	res.append(ita.iontype.abbr);
	rms += (w-ita.w)*(w-ita.w);
	total_abundance += rel_a;
	foundSome = true;
      }
    }
    if (foundSome)
      res.append(")"+ // "rms="+fmt3.format(rms)+
		 " sum(abundance)="+fmt1.format(100*total_abundance)+ "%");
    return res.toString();
  }

  /**
   * Compare two peaks wrt. abundance
   */

  public Comparator abundanceComparator = new Comparator() {
      public int compare (Object o1, Object o2) {
	Peak p1 = (Peak)o1;
	Peak p2 = (Peak)o2;
	if (p1.y<p2.y)
	  return 1;
	else if (p1.y>p2.y)
	  return -1;
	else 
	  return p1.compareTo(p2);
      }
      public boolean equals (Object o1, Object o2) {
	return compare(o1,o2)==0;
      }
    };
  

  /**
   * lal set by taking max of abundances within window around peak.
   */
  //This should perhaps be run only on file reading
  private void setLocalAbundanceLevels(double window) {
    double[] la = new double[origPeakCount];
    double[] lb = new double[origPeakCount];
    int lowindex = 0;
    int index = -1;
    int highindex = 0;
    final int RANK = 10;
    TreeSet treeSet = new TreeSet(abundanceComparator);
    while (++index < obs.length) {
      if (obs[index].orig_index < 0)
	continue;
      double w = obs[index].x;
      while (highindex < obs.length && obs[highindex].x < w+window) {
	if (obs[highindex].orig_index >= 0)
	  treeSet.add(obs[highindex]);
	highindex++;
      }
      while (lowindex < obs.length && obs[lowindex].x < w-window) {
	if (obs[lowindex].orig_index >= 0) 
	  treeSet.remove(obs[lowindex]);
	lowindex++;
      }
      try {
	la[obs[index].orig_index] = ((Peak)treeSet.first()).y;
      }
      catch (NoSuchElementException e) {
	System.out.println(lowindex+" "+index+" "+highindex+" "+
			   obs.length +" "+e);
      }
      Iterator iter = treeSet.iterator();
      int rank = 0;
      Object foo = null;
      while (iter.hasNext() && rank++<RANK)
	foo= iter.next();
      try {
	lb[obs[index].orig_index] = ((Peak)foo).y;
      }
      catch (NullPointerException e) {
	System.out.println(lowindex+" "+index+" "+highindex+" "+
			   obs.length+" "+e);
      }
    }
    for (int i=0; i<obs.length; i++)
      if (obs[i].orig_index >=0) {
	obs[i].la = la[obs[i].orig_index];
	obs[i].lb = lb[obs[i].orig_index];
      }
  }

  /**
   * Remove a single peak and insert a reinterpreted peak
   *
   * Reuse old peaks in order to easily preserve sequences! Note that
   * we only do so when isMirrored is false!
   * 
   */
  boolean reinterpretOne(Peak p, IonType iontype, Sequences sequences) {
    if (p != null && !isMirrored) {
      //Only reinterpret if ORIGINAL or TRANSFORMED:
      if (p.kind == Peak.ORIGINAL || p.kind == Peak.TRANSFORMED) {
	sequences.maybeRemovePeak(p);
	//We have remembered the original mass:
	p.x = iontype.mapToY(p.x_orig, totalMass);
	p.kind = iontype==IonType.Y ? Peak.ORIGINAL : Peak.TRANSFORMED;
	p.iontype = iontype==IonType.Y ? null : iontype;
	Arrays.sort(obs);
	return true;
      } 
      return false;
    }
    return false;
  }

  //Reinterpret a whole list
  public boolean reinterpretList(Collection pullList, Peak avoidMe, 
				 Sequences sequences) { 
    boolean wasChanged = false;
    if (!isMirrored) {
      Iterator iter = pullList.iterator();
      while (iter.hasNext()) {
	ITA ita = (ITA)(iter.next());
	if (ita.p != avoidMe && ita.iontype != ita.p.iontype) 
	  wasChanged |= reinterpretOne(ita.p,ita.iontype, sequences);
      }
    }
    return wasChanged;
  }

  public boolean removeReinterpretations(Sequences sequences) {
    boolean wasChanged = false;
    if (!isMirrored) {
      for (int i=0; i<obs.length; i++) {
	Peak p = obs[i];
	if (p.kind == Peak.TRANSFORMED) {
	  sequences.maybeRemovePeak(p);
	  p.x = p.x_orig;
	  p.kind = Peak.ORIGINAL;
	  p.iontype = null;
	  wasChanged = true;
	}
      }	
      Arrays.sort(obs);
    }    
    return wasChanged;
  }
  
  /**
   * Transform the spectrum by mirroring (Y -> B) all non-synthetic
   * peaks and show it, or remove all mirrored peaks if this was a
   * mirrored spectrum
   */  
  //retain sequences!
  void mirror(Sequences sequences) {
    TreeSet newobs = new TreeSet();
    //We do not have a direct copy of the original spectrum, but we
    //can recreate it:
    for (int i=0; i<obs.length; i++) {
      Peak p = obs[i];
      if (p.kind == Peak.SYNTHETIC_BOTTOM ||p.kind == Peak.SYNTHETIC_TOP)
	newobs.add(p);
      else if (isMirrored) {
	switch (p.kind) {
	case Peak.ORIGINAL: 
	  newobs.add(p);
	  break;
	case Peak.MIRRORED:
	  break;
	default:
	  throw new IllegalArgumentException();
	}
      }
      else { //not mirrored yet
	switch (p.kind) {
	case Peak.ORIGINAL: 
	  newobs.add(p);
	  double newmass = IonType.B.mapFromY(p.x,totalMass);
	  Peak np = new Peak(newmass,p.x_orig,p.y,p.orig_index, Peak.MIRRORED);
	  newobs.add(np);
	  break;
	case Peak.TRANSFORMED:
	  sequences.maybeRemovePeak(p);
	  np = new Peak(p.x_orig,p.x_orig,p.y,p.orig_index, Peak.ORIGINAL);
	  newobs.add(np);
	  newmass = IonType.B.mapFromY(np.x,totalMass);
	  np = new Peak(newmass,p.x_orig,p.y,p.orig_index, Peak.MIRRORED);
	  newobs.add(np);
	  break;
	default:
	  throw new IllegalArgumentException();
	}
      }
    }
    obs = (Peak[])newobs.toArray(new Peak[] { });
    isMirrored = !isMirrored; 
  }

  
  /** 
   * The first i with obs[i].x >= x (used when clipping). If no such i
   * exists, return obs.length-1.
   */
  public int getIndex(double x) {
    int bot = 0; 
    if (x<= obs[bot].x)
      return bot;
    int top = obs.length-1;
    if (x>= obs[top].x)
      return top;
    while (top-bot>1) {
      int mid =(top+bot)/2;
      if (obs[mid].x < x)
	bot = mid;
      else
	top = mid;
    }
    return top; 
  }

  //Let w be the left endpoint and aaList be a list of AmonoAcids, so
  //(w,aaList) represent a putative chunk of a Peptide. Method
  //fillGap() returns a score for how well this chunk matches the
  //Spectrum at hand: we create all types of ions from all possible
  //peptide bonds and "count how many" of these ions match the
  //spectrum within (tolerance,grass). One would imagine taking
  //abundance and mass_error into account.
  //
  // Moreover, one could imagine normalizing the result according to
  // aaList.size(). Consider having to choose between G+G and N: G+G
  // will generate exactly the same ions at positions 0 and 2 as N
  // will at positions 0 and 1. 
  //
  // If we "just count", "fillGap(G+G)-fillGap(N)" will be the sum of
  // score for matches of ions of G+G at position 1 and even a slight
  // random match of a rare iontype would decide in favor of G+G - in
  // fact N would never be preferred in favor of G+G. 
  //
  // OTH if we divide by aaList.length, G+G would only be preferred to
  // N if the midpoint scored better than the endpoints. If, say,
  // there were both B and Y peaks for the endpoints but only a Y peak
  // for the midpoint, N would be preferred, which is quite wrong in
  // this case.
  //
  // The G+G vs N example seems to indicate that the return value of
  // fillGap should be something Comparable but more complicated than
  // just a double value. Perhaps sth. like (#Ymatches, #Bmatches,
  // %matches)
  public String fillGap(double w, Collection aaList, 
		       double tolerance, double grass) {
    Iterator iter = aaList.iterator();
    int matchY = 0;
    int matchB = 0;
    int match = 0;
    while (iter.hasNext()) {
      AminoAcid aa = (AminoAcid)(iter.next());
      w += aa.monoW;
      boolean foundY = false;
      boolean foundB = false;
      //Since we have precomputed ionTypeAlternatives we do not have
      //to loop over iontype and match to the spectrum:
      Iterator itas = getIonTypeAlts(w, tolerance, grass).iterator();
      while (itas.hasNext()) {
	ITA ita = (ITA)(itas.next());
	foundY |= ita.iontype == IonType.Y;
	foundB |= ita.iontype == IonType.B;
	match++;
      }
      if (foundY)
	matchY++;
      if (foundB)
	matchB++;
    }
    return aaList+": "+matchY+" "+matchB+" "+match;
    //Disgusting:
    //return matchY*100.0+matchB+match/(1.0*aaList.size()*IonType.getCount());
  }


  /**
   * Read a file in (very restricted) Mascot format, or Sequest DTA format
   */
  static Spectrum readFile(String filename) throws IOException {
    Reader rd = new BufferedReader(new FileReader(filename));
    StreamTokenizer stok = new StreamTokenizer(rd);
    boolean dtaformat = true;
    boolean nomascotchargetoken = true;
    int chargeState = 1;
    double totalMass = -1;
    stok.parseNumbers();
    stok.eolIsSignificant(true);
    stok.ordinaryChars('=', '=');
    stok.nextToken();
    while (stok.ttype == StreamTokenizer.TT_WORD) {
      // Apparently not a Sequest DTA file; so guess it is Mascot MS/MS Ions
      dtaformat = false;
      String kw = stok.sval.toUpperCase();
      // System.out.println(kw);
      if (kw.equals("BEGIN"))
	skipLine(stok);
      else if (kw.equals("TITLE")) 
	skipLine(stok);
      else if (kw.equals("PEPMASS")) 
	totalMass = skipEqAndReadNumber(stok);
      else if (kw.equals("CHARGE")) {
	chargeState = (int)(0.5 + skipEqAndReadNumber(stok));
	nomascotchargetoken = false;
      }
      else 
	throw new IOException("Unrecognized keyword " + stok.sval 
	 		      + " in line " + stok.lineno());
    }
    if (!dtaformat) {
      //totalMass = (totalMass - chargeState) * chargeState + 1;
    }
    else { 
      if (stok.ttype == StreamTokenizer.TT_NUMBER) {
	totalMass = stok.nval;	
	stok.nextToken();
      } else 
 	throw new IOException("Number expected in line " + stok.lineno());
      if (stok.ttype == StreamTokenizer.TT_NUMBER) {
 	chargeState = (int)(0.5 + stok.nval);
 	skipLine(stok);
      } else 
 	throw new IOException("Number expected in line " + stok.lineno());
    }
    ArrayList /* of Peak */ obs = new ArrayList();
    double maxX=-1;
    while (stok.ttype == StreamTokenizer.TT_NUMBER) {
      double x = stok.nval; // chargeState * stok.nval;
      maxX = x; //So we could have declared x outside this loop
      stok.nextToken();
      double y = stok.nval;
      stok.nextToken();
      obs.add(new Peak(x, x, y, obs.size(), Peak.ORIGINAL));
      if (stok.ttype == StreamTokenizer.TT_EOL)
	stok.nextToken();
      else
	if (stok.ttype != StreamTokenizer.TT_EOF)
	  throw new IOException("Newline expected in line " + stok.lineno());
    }
    if (dtaformat) {
      //This seems to work, but we have very few dtaformat files to test on
      totalMass -= ChemicalFormula.H.monoW; 
    }
    else {
      if (nomascotchargetoken) { //guess chargestate 
	chargeState = (int)(maxX/totalMass+1.0);
	System.out.println("Guessing chargestate from: " + maxX/totalMass);
      }
      totalMass = (totalMass - ChemicalFormula.H.monoW ) * chargeState;
    }
    return new Spectrum(filename, obs.size(),totalMass , 
			chargeState, obs);
  }

  static void skipLine(StreamTokenizer stok) throws IOException {
    while (stok.ttype != StreamTokenizer.TT_EOF 
	   && stok.ttype != StreamTokenizer.TT_EOL) 
      stok.nextToken();
    stok.nextToken();
  }

  static double skipEqAndReadNumber(StreamTokenizer stok) throws IOException {
    while (stok.ttype != StreamTokenizer.TT_EOF && stok.ttype != '=') 
      stok.nextToken();
    stok.nextToken();
    if (stok.ttype == StreamTokenizer.TT_NUMBER) {
      double res = stok.nval;
      skipLine(stok);
      return res;
    } else
      throw new IOException("Number expected in line " + stok.lineno());
  }

  //experimental deconvolution stuff is here at the end

  public void remapDoubleCharged(double grass) {
    //assert spectrum has not been remapped or mirrored
    final double halfProton = 0.5;
    final double halftol = 0.04;
    //We assume that three or more consequtive, above grass peaks
    //0.5+-0.03 Da apart constitutes an isotopic group of a
    //doubly charged ion. 

    // Copy the peak(references) to a scratch //TreeSet
    TreeSet oldobs = new TreeSet(Arrays.asList(obs));
    //This is where the remapped spectrum will end up
    TreeSet newobs = new TreeSet();
    while (!oldobs.isEmpty()) {
      Peak p = (Peak)(oldobs.first());
      oldobs.remove(p);
      //grass is not considered as a possible start of an isotopic
      //peak group:
      if (p.y/maxAbundance<grass) {
	newobs.add(p);
	continue;
      }
      Iterator iter = oldobs.iterator();
      long lastRank = 0;
      LinkedList potentialRemap = null;
      //Some of the following code is a little silly, trying to cope
      //with situations where two distinct peaks fit into the same
      //p.x+ir*halfProton+-halftol interval:
      while (iter.hasNext()) {
	Peak q = (Peak)(iter.next());
	if (q.kind != Peak.ORIGINAL)
	  continue;
	long ir = Math.round((q.x-p.x)/halfProton);
	if (ir>=lastRank+2) 
	  //q.x >= p.x+(lastRank+1)*halfProton+halfProton/2 
	  break;
	if (Math.abs(p.x+ir*halfProton-q.x) < halftol) {
	  //ir is lastRank or lastRank+1
	  if (potentialRemap==null)
	    potentialRemap = new LinkedList();
	  potentialRemap.add(q);
	  if (q.y/maxAbundance>grass && q.y > p.y/(2*ir)) //quite ad hoc
	    lastRank = ir;
	}
      }
      if (lastRank>=2) {
	double newmass = 2*p.x-Element.H.monoW;
	newobs.add(new Peak(newmass, newmass, p.y, p.orig_index, 
			    Peak.ORIGINAL)); 
	iter = potentialRemap.iterator();
	while (iter.hasNext()) {
	  Peak q = (Peak)(iter.next());
	  oldobs.remove(q);
	  //Note: this is *not* a IonType reinterpretation
	  newmass = 2*q.x-Element.H.monoW;
	  newobs.add(new Peak(newmass, newmass, q.y, q.orig_index, 
			      Peak.ORIGINAL)); 
	}
      }
      else {
	newobs.add(p);
      }
    }
    obs = (Peak[])newobs.toArray(new Peak[] { });
    setITAs();
  }

  public void remapIsotopeVariants(double grass) {
    //assert spectrum has not been remapped or mirrored
    final double isoDiff = 1.0;
    final double isoTol = 0.1;
    final double isoAbu = Element.C.isotopes[1].abundance; //C13
    final double isoFact = isoAbu/(1-isoAbu);
    //For each peak (in an upwards sweep) assume it has
    //nC=Math.round(x*0.04) C-atoms and remove from the following
    //peaks at x+i*isoDiff abundance corresponding to
    //y*binomial(nC,0.0111,i). If the new abundance is below grass,
    //remove the peak.

    // Copy the peak(references) to a scratch //TreeSet
    TreeSet oldobs = new TreeSet(Arrays.asList(obs));
    //This is where the remapped spectrum will end up
    TreeSet newobs = new TreeSet();
    while (!oldobs.isEmpty()) {
      Peak p = (Peak)(oldobs.first());
      oldobs.remove(p);
      //grass is not considered as a possible start of an isotopic
      //peak group:
      if (p.y/maxAbundance<grass) {
	newobs.add(p);
	continue;
      }
      Iterator iter = oldobs.iterator();
      long nCest = Math.round(p.x*0.04);
      int lastRank = 0;
      double binom = 1.0;
      double newAbundance = p.y;
      double toBeRemapped = 0;
      while (iter.hasNext()) { //And binom*p.y>=grass*maxAbundance
	Peak q = (Peak)(iter.next());
	if (q.kind != Peak.ORIGINAL)
	  continue;
	int rank = (int)(Math.round((q.x-p.x)/isoDiff));
	//
	if (rank>lastRank) {
	  for (int i=lastRank+1; i<=rank;i++) {
	    binom *= (nCest-i+1)*isoFact/i;
	  }
	  lastRank = rank;
	  toBeRemapped = binom*p.y;
	  if (toBeRemapped/maxAbundance<grass) 
	    break;
	}
	//
	if (Math.abs(p.x+rank*isoDiff-q.x) < isoTol) {
	  double newy = q.y-toBeRemapped;
	  newAbundance += Math.min(q.y, toBeRemapped);
	  if (newy>grass*maxAbundance) {
	    //newobs.add(new Peak(q.x, q.x, newy, q.orig_index, 
	    //		Peak.ORIGINAL)); 
	    q.y = newy;
	  }
	  else 
	    iter.remove();
	  toBeRemapped = Math.max(0.0, -newy);
	}
      }
      newAbundance = Math.min(newAbundance, maxAbundance);
      newobs.add(new Peak(p.x, p.x, newAbundance, p.orig_index, 
			  Peak.ORIGINAL)); 
    }
    obs = (Peak[])newobs.toArray(new Peak[] { });
    setITAs();
  }


} //class Spectrum


//---------------------------------------------------------------------------
//Helper class ITA
class ITA implements Comparable {
  double w;
  IonType iontype;
  Peak p;
  ITA(double w, IonType iontype, Peak p) {
    this.w = w;
    this.iontype = iontype;
    this.p = p;
  }
  //If two ITAs have the same iontype and p.orig_index they should
  //be considered the same. Instead of comparing the IonTypes, which
  //we are not particularly keen on, we simply (as a quick hack)
  //compare two int references to the originating peaks. This is
  //correct since different iontypes would give different w for the
  //same originating peak. This makes it easy to create fake ITAs
  //with the right comparison properties in getIonTypeAltsFormatted
  //below.
  public int compareTo(Object o) {
    if (!(o instanceof ITA)) 
      throw new IllegalArgumentException("ITA comparation");
    ITA other = (ITA)o;
    if (w < other.w)
      return -1;
    else if (w > other.w)
      return 1;
    else 
      return p.orig_index-other.p.orig_index;
  }
}

