
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.awt.print.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;

// The spectrum window 

// Notation: 
//   (x, y)            are true (mass, abundance) coordinates
//   (xoff, yoff)      is the spectrum window origo, in true coordinates
//   (width, height)   is the displayed spectrum width, in true coordinates

//   (gx, gy)          are spectrum window coordinates, in pixels 
//   (gxoff, gyoff)    is the spectrum window origo, in pixels
//   (gwidth, gheight) are spectrum window dimensions, in pixels

//   (xscale, yscale)  is the scale, pixels per true unit, at zoom=1
//   zoom              is the zoom factor, so the actual number of
//                     pixels per true unit is xscale*zoom, yscale

// Computing the scales (for linear y axis)
// --------------------
//   xscale = gwidth/width
//   yscale = gheight/height

// From true coordinates (x, y) to window coordinates (gx, gy):
//   gx = (int)((x - xoff) * xscale * zoom + gxoff)
//   gy = (int)(gheight - (y - yoff) * yscale * zoom - gyoff)

// From window coordinates (gx, gy) to true coordinates (x, y):
//   x = (gx - gxoff) / xscale / zoom + xoff
//   y = (-gy - gyoff + gheight) / yscale / zoom + yoff

// For logarithmic y axis substitute:
//   yscale = (gheight-gyoff-TOPOFFSET)/Math.log(LIMIT/seqParms.grass);
//   gy = TOPOFFSET+yscale*Math.log(spectrum.maxAbundance/y
//   y  = spectrum.maxAbundance*Math.exp((TOPOFFSET-gy)/yscale);


/**
 * SpectrumWindow
 */

class SpectrumWindow extends JPanel implements Printable /*, KeyListener*/ {
  private final static double gTolerance = 49.0;   // in pixels squared
  private final static int gxoff = 10, gyoff = 20; // axis offsets
  //private final double aDiscount = 0.00001;       // abundance match discount 

  private Dimension initialSize;	// Initial size of the canvas
  private double width, height;		// In true coordinate units
  private double xoff, yoff;		// In true coordinate units
  private double zoom;			// > 1 = enlarge

  //A few fields maintained and used in the mousehandler but also
  //needed in the keyhandler:
  private boolean mouseinside;          //Remember whether mouse is inside
  private int mousegx, mousegy;		// Most recent mouse coordinates

  private Log log;			// The log of attempted sequences

  private Spectrum spectrum;		// The spectrum being shown

  private Sequences sequences;		// The (sub)sequences being built

  //Some fields determining the appearance of this window
  private boolean showErrors;		// show Errors of accepted peaks
  private boolean showLocalLevels;      // show local levels
  private boolean logScale;             // Use log scale for y axis
  private boolean showBackwards;	// Show also backward spectrum
  private String showExtra;             // Show also reinterpreted spectrum

  private Peak[] sortedAbundances;	// Used for labelling
  private int peaksLabelled;		// Number of mass-labelled peaks

  private final static DecimalFormat fmt1 = new DecimalFormat("0.0");
  private final static DecimalFormat fmt2 = new DecimalFormat("0.00"); 
  private final static DecimalFormat fmt3 = new DecimalFormat("0.000");

  private AbstractController controller; //SIC
  
  private Line2D zoomLine;
  
  // Construct a new empty spectrum window

  public SpectrumWindow(Dimension initialSize, 
			final AbstractController controller) {
    this.initialSize = initialSize;
    this.controller = controller;
    controller.registerSpectrumWindow(this);
    this.log = new Log();

    MouseInputListener mousie = new ChooseListener();
    addMouseListener(mousie);
    addMouseMotionListener(mousie);

    //The following is a test for using the jdk1.3 interface instead
    //of the KeyListener inteface
    InputMap im = getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_R,0),"resetZoom");
    getActionMap().put("resetZoom", resetZoomAction);
    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_Z,0),"zoomIn");
    getActionMap().put("zoomIn", zoomInAction);
    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_O,0),"zoomOut");
    getActionMap().put("zoomOut", zoomOutAction);
    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_C,0),"recenter");
    getActionMap().put("recenter", recenterAction);

    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_M,0),"moreMasses");
    getActionMap().put("moreMasses", showMassesAction);
    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_N,0),"hideMasses");
    getActionMap().put("hideMasses", hideMassesAction);
    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN,0),"stepLeft");
    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT,0),"stepLeft");
    getActionMap().put("stepLeft", stepLeftAction);
    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP,0),"stepRight");
    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT,0),"stepRight");
    getActionMap().put("stepRight", stepRightAction);
    im.put(KeyStroke.getKeyStroke(KeyEvent.VK_D,0),"debug");
    getActionMap().put("debug", debugAction);

  }

  /**
   * Set and show a (new) spectrum
   */

  void resetVariousVariables() {
    sequences = controller.newSequences();
    showBackwards = false;
    showExtra = "None";
    controller.resetProposal();
    crop(0, 0, 1.0);
    controller.mousePosition();
  }
  
  /**
   * Set and show a (new) spectrum
   */

  public void setSpectrum(Spectrum spectrum, boolean reset) {
    this.width = spectrum.totalMass*1.10; 
    this.height = spectrum.maxAbundance*1.15;
    
    this.spectrum = spectrum;
    // Create sorted table of abundances to assist mass labelling
    int obsCount = spectrum.obs.length;
    this.sortedAbundances = new Peak[obsCount];
    for (int i=0; i<obsCount; i++)
      sortedAbundances[i] = spectrum.obs[i];
    Arrays.sort(sortedAbundances, spectrum.abundanceComparator );
    if (reset)
      resetVariousVariables();
    controller.showSeq();
    controller.clearUndoStack();
    controller.checkPoint();
    enableMirrorAction();
    printAction.setEnabled(true);
    peaksLabelled = 0;
    controller.setSpectrum(spectrum);
    controller.message("");
    repaint();
  }

  /**
   * Forward various varibale values to sequences. 
   */

  public void setParams() {
    if (sequences!=null)
      sequences.setParams(gTolerance,showErrors, spectrum);
  }

  /**
   * call sequences.setPotential()
   */

  public void setPotential() {
    if (sequences != null)
      sequences.setPotential();
  }

  //--------------------------------------------------------------------------
  //Call Spectrum.mirror()

  private void mirror() {
    if (spectrum != null) {
      spectrum.mirror(sequences);
      //setSpectrum(newspectrum, false);
      sequences.pruneMirroredPeaks();
      setParams();
      setPotential(); //Should clean out missing peaks
      controller.setSpectrumChanged();
      controller.showSeq();
      mirrorAction.putValue(Action.NAME, spectrum.isMirrored ? 
			    "Remove mirrored spectrum" : 
			    "Add mirrored spectrum");
      repaint();
    }
  }
  private Action mirrorAction = new AbstractAction("Add mirrored spectrum") {
      {
	this.setEnabled(false);
      }
      public void actionPerformed(ActionEvent e) {
	mirror();
      }
    };
  public Action getMirrorAction() {
    return mirrorAction;
  }
  public void enableMirrorAction() {
    mirrorAction.putValue(Action.NAME, spectrum.isMirrored ? 
			  "Remove mirrored spectrum" : 
			  "Add mirrored spectrum");
    mirrorAction.setEnabled(true);
  }
  
  //--------------------------------------------------------------------------
  // React to mouse clicks

  //Note: nearestPeak and corresponding creation of Actions should be
  //a method in Spectrum!
  private int nearestPeak(int gx, int gy, Coordinates coor) {
    //assert spectrum != null && spectrum.obs != null && spectrum.obs.length>0;
    Peak[] obs = spectrum.obs;
    int nearestPeak = -1; 
    double dist = gTolerance;
    int imin = spectrum.getIndex(coor.toX(gx-(int)gTolerance));
    int imax = spectrum.getIndex(coor.toX(gx+(int)gTolerance));
    for (int i=imin; i<=imax; i++) {
      double idist = coor.dist(gx, gy, obs[i].x, obs[i].y);
      if (idist < dist) {
	dist = idist;
	nearestPeak = i;
      }
    }
    return nearestPeak;
  }

  //
  // Note: we avoid using the middle button

  /**
   * Inner class SpectrumWindow.ChooseListener
   *
   * Handles Mouse Events in the spectrum window
   */

  class ChooseListener extends MouseInputAdapter {
    //Remember whether we did popup on keyPressed/Released
    private boolean didPopup;      

    //Fields for controlling zoom by dragging
    private int pressed_x;      //For drag support
    private int pressed_y;      //For drag support
    private int zoomLineY;      //At which height to put the zoom feedback line
    private boolean dozoom;     //      
    private int old_zoom_x;     //For efficient repainting
    private final static int zoomthreshold = 30;
    private Cursor zoomCursor;
    {
      Toolkit tk = Toolkit.getDefaultToolkit();
      Dimension bestsize = tk.getBestCursorSize(1,1);
      if (bestsize.width != 0) {
	int w = bestsize.width;
	int h = bestsize.height;
	BufferedImage img = 
	  new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
	Graphics2D g = img.createGraphics();
	Font f = new Font("Monospaced",Font.PLAIN,10);
	g.setFont(f);
	g.setColor(Color.blue);
	g.drawString("ZOOM",0,10);
	zoomCursor = tk.createCustomCursor(img, new Point(0,0), "MagCursor");
      }
      else
	zoomCursor = new Cursor(Cursor.E_RESIZE_CURSOR);
    }

    private Coordinates coor;

    private void position(MouseEvent e) {
      if (!mouseinside || spectrum == null)
	controller.mousePosition();
      else {
	coor = new Coordinates(getSize());
	mousegx = e.getX();
	mousegy = e.getY();
	double x = coor.toX(mousegx);
	double rel_y = coor.toY(e.getY())/spectrum.maxAbundance;
	controller.mousePosition(x,rel_y);
      }
    }

    public void mouseEntered(MouseEvent e) {
      mouseinside = true;
      position(e);
    }

    public void mouseExited(MouseEvent e) {
      mouseinside = false;
      position(e);
      if (spectrum != null) 
	repaint(0,coor.gheight-12,coor.gwidth,coor.gheight);
    }
    
    public void mouseClicked(MouseEvent e) {
      if (!didPopup) 
	mouseClick(e);
    }

    public void mousePressed(MouseEvent e) {
      old_zoom_x = pressed_x = e.getX();
      pressed_y = e.getY();
      Coordinates coor = new Coordinates(getSize());
      zoomLineY = pressed_y;
      if (e.isPopupTrigger()) { 
	maybeShowPopup(e);
	didPopup = true;
      }
      else 
	//This is to reset a previous didPopup=true
	didPopup = false;
    }
    
    public void mouseReleased(MouseEvent e) {
      setCursor(Cursor.getDefaultCursor()); 
      zoomLine = null;
      if (dozoom)
	  repaint (Math.min(pressed_x,old_zoom_x),
		   zoomLineY-1,
		   Math.max(pressed_x,old_zoom_x), 
		   zoomLineY+1);
      if ( dozoom &&
	   spectrum!=null &&
	   Math.abs(pressed_x-e.getX()) > zoomthreshold) {
	Coordinates coor = new Coordinates(getSize());
	double newxoff = coor.toX(Math.min(pressed_x,e.getX()));
	double newotherend = coor.toX(Math.max(pressed_x,e.getX()));
	double newzoom = spectrum.totalMass/(newotherend-newxoff);
	crop(newxoff, 0, newzoom); 
      }
      else {
      	if (e.isPopupTrigger()) { 
	  maybeShowPopup(e);
	  didPopup = true;
	}
      }
      dozoom = false;
    }
    
    public void mouseMoved(MouseEvent e) {
      if (spectrum != null) {
	position(e);
	double delta = sequences.delta(mousegx, mousegy, coor);
	if (delta != Double.NEGATIVE_INFINITY)
	  controller.setDelta(fmt3.format(delta));
	else
	  controller.setDelta("");
	//This should only be done if viewing alternative interprtations:
	repaint(0,coor.gheight-12,coor.gwidth,coor.gheight);
      }
    }

    public void mouseDragged(MouseEvent e) {
      if (mouseinside && !didPopup &&
	  Math.abs(pressed_x-e.getX()) > zoomthreshold) {
	dozoom = true;
	zoomLine = new Line2D.Float(pressed_x,zoomLineY,e.getX(),zoomLineY);
	repaint (Math.min(old_zoom_x,e.getX()),
		 zoomLineY-1,
		 Math.max(old_zoom_x,e.getX()), 
		 zoomLineY+1);
	old_zoom_x = e.getX();
	setCursor(zoomCursor);
      }
      else {
	zoomLine = null;
	if (dozoom)
	  repaint (Math.min(pressed_x,old_zoom_x),
		   zoomLineY-1,
		   Math.max(pressed_x,old_zoom_x), 
		   zoomLineY+1);
	dozoom = false;
	old_zoom_x = pressed_x; 
	setCursor(Cursor.getDefaultCursor()); 
      }
    }

    //The utility functions called above
    private void mouseClick(MouseEvent e) {
      if (spectrum != null && spectrum.obs != null) {
	setParams();
	position(e);
	Mark lm = sequences.nearestMark(mousegx, mousegy, coor, true);
	Mark rm = sequences.nearestMark(mousegx, mousegy, coor, false);
	if (lm == null && rm == null) {
	  int npi = nearestPeak(mousegx, mousegy, coor);
	  if (npi>-1) {
	    Peak clickedPeak = spectrum.obs[npi];
	    if (sequences.canStartNew(clickedPeak)) {
	      sequences.startNew(clickedPeak);
	      controller.checkPoint();
	      setPotential();
	      controller.showSeq();
	      repaint();
	    }
	  }
	}
	else {
	  if (lm != null) 
	    if (lm.kind == Mark.POTENTIAL) 
	      ((AAMark)lm).accept();
	    else
	      lm.remove(); 
	  if (rm != null) 
	    if (rm.kind == Mark.POTENTIAL) 
	      ((AAMark)rm).accept();
	    else
	      rm.remove(); 
	  controller.checkPoint();
	  setPotential();
	  controller.showSeq();
	  repaint();
	}
      }
    }

    private void maybeShowPopup(MouseEvent e) {
      if (spectrum != null) {
	setParams(); //silly, but needed for getNearestPeakIndex() //NB! still?
	mousegx = e.getX(); mousegy = e.getY();
	final Coordinates coor = new Coordinates(getSize());
	JPopupMenu popup = new JPopupMenu();
	JMenuItem menuItem;
	//
	Mark[] nearbyMarks = sequences.nearbyMarks(mousegx, mousegy, coor);
	if (nearbyMarks.length > 0) {
	  popup.add("Marks:");
	  for (int j=0; j< nearbyMarks.length; j++) {
	    final Mark m = nearbyMarks[j];
	    if (m.kind == Mark.START) 
	      menuItem = new JMenuItem("Remove sequence at "+fmt1.format(m.x));
	    else {
	      AAMark am = (AAMark) m;
	      String text = am.acid.abbr+" mark at "+fmt1.format(m.x)+ 
		" (" + (am.left?"left":"right") + " branch)";
	      if (m.kind == Mark.ACCEPTED) 
		menuItem = new JMenuItem("Remove " + text);
	      else 
		menuItem = new JMenuItem("Accept " + text);
	    }
	    menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		  if (m.kind == Mark.POTENTIAL) 
		    ((AAMark)m).accept();
		  else
		    m.remove(); 
		  controller.checkPoint();
		  setPotential();
		  controller.showSeq();
		  repaint();
		}
	      });
	  popup.add(menuItem);
	  }
	  popup.addSeparator();
	}
	final int npi = nearestPeak(mousegx, mousegy, coor);
	final Peak clickedPeak = npi>-1 ? spectrum.obs[npi] : null;
	if (npi>-1) {
	  if (clickedPeak.ignore) {
	    popup.add("Peak at " + fmt1.format(clickedPeak.x) + " is ignored");
	      menuItem = new JMenuItem("Reconsider peak");
	      menuItem.addActionListener(new ActionListener() {
		  public void actionPerformed(ActionEvent e) {
		    clickedPeak.ignore = false;
		    //Should reset candidate here
		    setPotential();
		    controller.showSeq();
		    repaint();
		  }
		});
	      popup.add(menuItem);
	  }
	  else {
	    String interpretation = 
	      clickedPeak.iontype==null ||
	      clickedPeak.iontype==IonType.Y ? 
	      "assumed Y-ion" : 
	      "reinterpreted "+clickedPeak.iontype.abbr+"-ion";
	    popup.add("Peak at " + fmt1.format(clickedPeak.x) + 
		      " ("+interpretation+"):");
	    //We should not start a new sequence if there is a mark nearby!?
	    if (nearbyMarks.length==0 && sequences.canStartNew(clickedPeak)) {
	      menuItem = new JMenuItem("Start new sequence here");
	      menuItem.addActionListener(new ActionListener() {
		  public void actionPerformed(ActionEvent e) {
		    sequences.startNew(clickedPeak);
		    controller.checkPoint();
		    setPotential();
		    controller.showSeq();
		    repaint();
		  }
		});
	      popup.add(menuItem);
	    }
	    LinkedList accL = sequences.acceptableMarks(clickedPeak, true);
	    LinkedList accR = sequences.acceptableMarks(clickedPeak, false);
	    accL.add(null);
	    accR.add(null);
	    final Mark clickedStart = sequences.isStart(clickedPeak);
	    Iterator liter = accL.iterator();
	    while (liter.hasNext()) {
	      Iterator riter = accR.iterator();
	      AAMark lmark = (AAMark) liter.next();
	      while (riter.hasNext()) {
		AAMark rmark = (AAMark) riter.next();
		if (lmark !=null || rmark !=null) {
		  menuItem = new JMenuItem("Add peak as " +
					   (rmark==null?"-":rmark.acid.abbr) + 
					   "," +
					   (lmark==null?"-":lmark.acid.abbr));
		  final AAMark lm = lmark;
		  final AAMark rm  = rmark;
		  menuItem.addActionListener(new ActionListener() {
		      public void actionPerformed(ActionEvent e) {
			if (lm != null)
			  lm.accept();
			if (rm != null)
			  rm.accept();
			if (lm != null && rm != null && clickedStart != null) {
			  clickedStart.remove();
			}
			controller.checkPoint();
			setPotential();
			controller.showSeq();
			repaint();
		      }
		    });
		  popup.add(menuItem);
		}
	      }
	    }
	    //
	    if (controller.proposalValid()) {
	      menuItem = new JMenuItem("Show proposal centered here");
	      menuItem.addActionListener(new ActionListener() {
		  public void actionPerformed(ActionEvent e) {
		    controller.showProposal(npi);
		    repaint();
		  }
		});
	      popup.add(menuItem);
	      //
	      for (int side=-1; side < 2; side +=2) {
		final boolean left = side < 0;
		final Proposal candidate = controller.currentProposal(left);
		if (candidate != null &&
		    sequences.canFollow(clickedPeak,candidate,left)) {
		  menuItem = 
		    new JMenuItem("Follow proposal "+(left?"left":"right"));
		  menuItem.addActionListener(new ActionListener() {
		      public void actionPerformed(ActionEvent e) {
			sequences.follow(clickedPeak,candidate,left);
			controller.checkPoint();
			setPotential();
			controller.showSeq();
			repaint();
		      }		 
		    });
		  popup.add(menuItem);
		  popup.addSeparator();
		}
	      }
	    }
	    //
	    //Ignorance:
	    if (sequences.peakUsage(clickedPeak).isEmpty()) {
	      menuItem = new JMenuItem("Ignore this peak");
	      menuItem.addActionListener(new ActionListener() {
		  public void actionPerformed(ActionEvent e) {
		    clickedPeak.ignore = true;
		    setParams();
		    setPotential();
		    controller.setSpectrumChanged();
		    repaint();
		  }
		});
	      popup.add(menuItem);
	    }
	    //
	    //Reinterpretations:
	    if (!spectrum.isMirrored && clickedPeak.orig_index > -1) {
	      Iterator iter = IonType.iterator();
	      JMenu subMenu = new JMenu("Reinterpret original as iontype");
	      while (iter.hasNext()) { 
		final IonType iontype = (IonType)(iter.next());
		menuItem = new JMenuItem(iontype.abbr);
		menuItem.addActionListener(new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
		      sequences.maybeRemovePeak(clickedPeak);
		      if (spectrum.reinterpretOne(clickedPeak,
						  iontype,
						  sequences)) {
			setParams();
			setPotential();
			controller.setSpectrumChanged();
			repaint();
		      }
		    }
		  });
		subMenu.add(menuItem);
	      }
	      popup.add(subMenu);
	    }
	  }
	}
	//
	if (nearbyMarks.length==0 && npi == -1) {
	  popup.add(controller.getUndoAction());
	  popup.add(controller.getRedoAction());
	  popup.addSeparator();
	  final double thisx = coor.toX(mousegx);
	  if (true && sequences.inGap(thisx)) {
	    menuItem = new JMenuItem("List possible peptides matching gap");
	    menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		  sequences.fillGap(thisx);
		}
	      });
	    popup.add(menuItem);
	  }
	  if (!sequences.isEmpty()) {
	    menuItem = new JMenuItem("Remove all sequences");
	    menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		  sequences.clear();
		  setParams();
		  setPotential();
		  controller.showSeq();
		  repaint();
		}
	      });
	    popup.add(menuItem);
	  }
	  if (!spectrum.isMirrored) {
	    //This action should also be present in menubar
	    popup.add(new AbstractAction("Remove all reinterpretations") {
		public void actionPerformed(ActionEvent e) {
		  if (spectrum.removeReinterpretations(sequences)) {
		    setParams();
		    setPotential();
		    controller.showSeq();
		    controller.setSpectrumChanged();
		    repaint();
		  }
		}
	      });
	  }
	  if (!spectrum.isMirrored) {
	    //This action should also be present in menubar
	    popup.add(new AbstractAction("Move doublecharged peaks") {
		public void actionPerformed(ActionEvent e) {
		  spectrum.removeReinterpretations(sequences);
		  sequences.clear();
		  controller.clearUndoStack();
		  spectrum.remapDoubleCharged(controller.seqParms.grass);
		  setParams();
		  setPotential();
		  controller.showSeq();
		  controller.setSpectrumChanged();
		  repaint();
		}
	      });
	    popup.add(new AbstractAction("Move isotope variants") {
		public void actionPerformed(ActionEvent e) {
		  spectrum.removeReinterpretations(sequences);
		  sequences.clear();
		  controller.clearUndoStack();
		  spectrum.remapIsotopeVariants(controller.seqParms.grass);
		  setParams();
		  setPotential();
		  controller.showSeq();
		  controller.setSpectrumChanged();
		  repaint();
		}
	      });
	  }
	  //NBNBNB
	  if (false) {
	    menuItem = new JMenuItem("NBNBNB!!!!!");
	    menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		  spectrum.tmtest();
		}
	      });
	    popup.add(menuItem);
	  }
	  popup.add(resetZoomAction);
	  popup.add(printAction);
	  popup.add(mirrorAction);
	}
	//
	final double thisx = coor.toX(mousegx);
	final Collection pullList = 
	    spectrum.getIonTypeAlts(thisx,
				    controller.seqParms.tolerance,
				    controller.seqParms.grass);
	if (!spectrum.isMirrored && !pullList.isEmpty()) {
	  popup.addSeparator();
	  menuItem = new JMenuItem("Collect "+
				   pullList.size()+
				   " peaks here ("+
				   fmt3.format(thisx)+
				   ") by reinterpretation");
	  menuItem.addActionListener(new ActionListener() {
	      public void actionPerformed(ActionEvent e) {
		if (spectrum.reinterpretList(pullList, clickedPeak, sequences)) {
		  setPotential();
		  controller.showSeq();
		  controller.setSpectrumChanged();
		  repaint();
		}
	      }
	    });
	  popup.add(menuItem);
	}
	popup.show(e.getComponent(), mousegx, mousegy);
      }
    }
  }

  private Action resetZoomAction = new AbstractAction("Reset zoom") {
      public void actionPerformed(ActionEvent e) {
 	crop(0.0, 0.0, 1.0);
      }
    };

  private Action zoomInAction =  new AbstractAction("Zoom in") {
      public void actionPerformed(ActionEvent e) {
	Coordinates coor = new Coordinates(getSize());
	double x = coor.toX(mousegx);
	double y = coor.toY(mousegy);
	xoff = (x + xoff) / 2;
	yoff = (y + yoff) / 2;
	crop(xoff, yoff, zoom * 2.0);
      }
    };

  private Action zoomOutAction =  new AbstractAction("Zoom out") {
      public void actionPerformed(ActionEvent e) {
	Coordinates coor = new Coordinates(getSize());
	double x = coor.toX(mousegx);
	double y = coor.toY(mousegy);
	if (zoom > 1.0) {
	  xoff = 2 * xoff - x;
	  yoff = 2 * yoff - y;
	  crop(xoff, yoff, zoom / 2.0);
	}
      }
    };

  private Action recenterAction =  new AbstractAction("Recenter") {
      public void actionPerformed(ActionEvent e) {
	Coordinates coor = new Coordinates(getSize());
	double x = coor.toX(mousegx);
	double y = coor.toY(mousegy);
	xoff = x - width/zoom/2;
	yoff = y - height/2;
	crop(xoff, yoff, zoom);
      }
    };

  private Action showMassesAction =  new AbstractAction("Show more masses") {
      public void actionPerformed(ActionEvent e) {
	if (peaksLabelled < sortedAbundances.length) 
	  peaksLabelled += 10;
	repaint();
      }
    };

  private Action hideMassesAction =  new AbstractAction("Hide more masses") {
      public void actionPerformed(ActionEvent e) {
	if (peaksLabelled > 0) 
	  peaksLabelled -= 10;
	repaint();
      }
    };

  private Action stepLeftAction =  new AbstractAction("Step Left") {
      public void actionPerformed(ActionEvent e) {
	Coordinates coor = new Coordinates(getSize());
	double x = coor.toX(mousegx);
	double y = coor.toY(mousegy);
	xoff = xoff - width/zoom/10.0;
	crop(xoff, yoff, zoom);
	if (mouseinside)
	  controller.mousePosition(x-width/zoom/10.0,y/spectrum.maxAbundance);
      }
    };

  private Action stepRightAction =  new AbstractAction("Step right") {
      public void actionPerformed(ActionEvent e) {
	Coordinates coor = new Coordinates(getSize());
	double x = coor.toX(mousegx);
	double y = coor.toY(mousegy);
	xoff = xoff + width/zoom/10.0;
	crop(xoff, yoff, zoom);
	if (mouseinside)
	  controller.mousePosition(x+width/zoom/10.0,y/spectrum.maxAbundance);
      }
    };

  private Action debugAction =  new AbstractAction("Debugout") {
      public void actionPerformed(ActionEvent e) {
	sequences.debugout();
      }
    };



  // React to keyboard inputs. Remember that we are a KeyListener()

  //This seems to be necessary only on the MS Windows platform -
  //without it VK_LEFT anfd VK_RIGHT are not delivered as keyPressed
  //to sw! NOte that this is deprecated in java 1.4!!!
  //public boolean isFocusable() {
  //  return true;
  //}

  /*
  public void keyTyped(KeyEvent e) {
  }
  public void keyReleased(KeyEvent e) {
  }
  public void keyPressed(KeyEvent e) {
    if (spectrum != null && !e.isControlDown()) {
      int key = e.getKeyCode();
      Coordinates coor = new Coordinates(getSize());
      double x = coor.toX(mousegx);
      double y = coor.toY(mousegy);
      switch (key) {
      case KeyEvent.VK_C:	 //Recenter around (x, y)
	xoff = x - width/zoom/2;
	yoff = y - height/2;
	crop(xoff, yoff, zoom);
	break;
	case KeyEvent.VK_Z:	// Zoom in around (x, y) 
	xoff = (x + xoff) / 2;
	yoff = (y + yoff) / 2;
	crop(xoff, yoff, zoom * 2.0);
	break;
	case KeyEvent.VK_O:	// Zoom out around (x, y) 
	if (zoom > 1.0) {
	  xoff = 2 * xoff - x;
	  yoff = 2 * yoff - y;
	  crop(xoff, yoff, zoom / 2.0);
	}
	break;
	//NB: the VK_R action has been put in the ActionMap/InputMap
	case KeyEvent.VK_R:	// Reset to origo and set zoom to 1.0
 	crop(0.0, 0.0, 1.0);
	break;
      case KeyEvent.VK_M:	// Show (more) peak mass numbers
	// assert sortedAbundances != null;
	if (peaksLabelled < sortedAbundances.length) 
	  peaksLabelled += 10;
	repaint();
	break;
      case KeyEvent.VK_N:	// Hide (more) peak mass numbers
	// assert sortedAbundances != null;
	if (peaksLabelled > 0) 
	  peaksLabelled -= 10;
	repaint();
	break;
      case KeyEvent.VK_LEFT: case KeyEvent.VK_KP_LEFT:   // Step 10% left
	xoff = xoff - width/zoom/10.0;
	crop(xoff, yoff, zoom);
	if (mouseinside)
	  controller.mousePosition(x-width/zoom/10.0,y/spectrum.maxAbundance);
	break;
      case KeyEvent.VK_RIGHT: case KeyEvent.VK_KP_RIGHT: // Step 10% right
	xoff = xoff + width/zoom/10.0;
	crop(xoff, yoff, zoom);
	if (mouseinside)
	  controller.mousePosition(x+width/zoom/10.0,y/spectrum.maxAbundance);
	break;
      case KeyEvent.VK_D:
	sequences.debugout();
	break;
      }
    }
  }
  */

  /**
   * change viewArea
   */

  public void crop(double xoff, double yoff, double zoom) {
    this.xoff = Math.max(0, xoff); 
    this.yoff = 0; 
    this.zoom = zoom;
    repaint();
  }

  //--------------------------------------------------------------------------
  // Functions to be called by Listeners for the JMenuBar items:

  // Allow or disallow the display of K and R

  public void setAllowKR(boolean allowKR) {
    controller.seqParms.allowKR = allowKR;
    if (allowKR) 
      controller.message("Allow K and R");
    else
      controller.message("Disallow K and R");
    controller.setSeqParmsChanged(SpectrumWindow.this);
  }

  // Allow or disallow the display of error texts

  public void setShowErrors(boolean showErrors) {
    this.showErrors = showErrors;
    if (sequences != null) {
      setParams();
      repaint();
    }
  }

  // Show sequence forwards or backwards?

  public void setBwdSequence(boolean bwdSequence) {
    controller.bwdSequence = bwdSequence;
    if (sequences != null) {
      setParams();
      controller.showSeq();
    }
    //NBNBNB Shouldn't we update the propP text too?
  }

  // Allow or disallow the display of local levels
  public void setShowLocalLevels(boolean showLocalLevels) {
    this.showLocalLevels = showLocalLevels;
    if (sequences != null) {
      setParams();
      repaint();
    }
  }

  // Show or don't show the backwards spectrum

  public void setBwdSpectrum(boolean backwards) {
    this.showBackwards = backwards;
    repaint();
  }
 
  // Show or don't show the extra spectrum

  public void setExtra(String extra) {
    this.showExtra = extra;
    repaint();
  }

  // Use or don't use log scale on Y axis

  public void setLogScale(boolean logScale) {
    this.logScale = logScale;
    repaint();
  }

  // Change the match tolerance for potential amino acids (in amu)

  public void decrTolerance() {
    if (controller.seqParms.tolerance >= 0.1999999999) 
      controller.seqParms.tolerance -= 0.1;
    controller.message("Tolerance = " + 
		       fmt1.format(controller.seqParms.tolerance) + " amu");
    controller.setSeqParmsChanged(SpectrumWindow.this);
  }

  public void incrTolerance() {
    if (controller.seqParms.tolerance <= 1.5000000001) 
      controller.seqParms.tolerance += 0.1;
    controller.message("Tolerance = " + 
		       fmt1.format(controller.seqParms.tolerance) + " amu");
    controller.setSeqParmsChanged(SpectrumWindow.this);
  }

  /**
   * Draw the spectrum, masses, and sequences. Calls 
   */

  public void paintComponent(Graphics g) {
    super.paintComponent(g);
    paintComponent(g,false,getSize());
  }

  /**
   * Draw the spectrum in a rectangle of given dimensions - to be used
   * both for the screen and for printing
   */

  public void paintComponent(Graphics g, boolean isPrinting, Dimension sz) {
    Graphics2D g2d = (Graphics2D) g;
    g.setColor(Color.white);
    g.fillRect(0, 0, sz.width, sz.height);
    Coordinates coor = new Coordinates(sz);
    g.setColor(Color.blue);
    if (zoomLine != null)
      g2d.draw(zoomLine);
    g2d.draw(new Line2D.Float(coor.fromXLim(xoff), 
			      coor.fromYLim(0),
			      coor.fromXLim(xoff+0.99*width), 
			      coor.fromYLim(0)));
    g2d.draw(new Line2D.Float(coor.fromXLim(0), 
			      coor.fromYLim(yoff),
			      coor.fromXLim(0), 
			      coor.fromYLim(yoff+0.99*height)));
    {
      //All this work just to paint a few tick marks
      double xtop = coor.toX(sz.width-20);
      double xbot = coor.toX(0);
      double xdiff = xtop - xbot;
      double a = Math.log(xdiff);
      double b = a-Math.log(10)*Math.floor(a/Math.log(10));
      int c = b<Math.log(2) ? 2 : b<Math.log(5) ? 5 : 10;
      double tickDiff = Math.exp(a-b)*c/10;
      int tickCount = (int)(xdiff/tickDiff);
      double tickStart = tickDiff*((int)(xbot/tickDiff)+1);
      Font oldfont = g.getFont();
      g.setFont(new Font("Monospaced",Font.PLAIN,10));
      FontMetrics fontmetrics = g.getFontMetrics();
      for (int j=0; j<tickCount; j++) {
	double x = tickStart+j*tickDiff;
	float gx = coor.fromX(x);
	float gy = coor.fromY(0);
	//note: DecimalFormat should be scaled too! 
	String s = fmt1.format(x);
	int gwidth = fontmetrics.stringWidth(s);
	g2d.drawString(s, gx-gwidth/2, gy+10);
	g2d.draw(new Line2D.Float(gx,gy,gx,gy+4));
      }
      g.setFont(oldfont);
    }
    //Show positions of other iontypes corresponding to a Y ion at the
    //mouseposition. Change it.mapFromY to it.mapToY to change the
    //sense and show where a reinterpretation of a Y peak at the mouse
    //position would end up.
    if (mouseinside && spectrum != null && !isPrinting) {
      Font oldfont = g.getFont();
      g.setFont(new Font("Monospaced",Font.PLAIN,10));
      Iterator iter = IonType.iterator();
      while (iter.hasNext()) {
	IonType it = (IonType)iter.next();
	String txt = it.abbr;
	//Too much to show all kinds of ions:
	if (txt.endsWith("*"))
	  continue;
	double mz = it.mapFromY(coor.toX(mousegx), spectrum.totalMass);
	float newgx = coor.fromX(mz);
	g2d.drawString(txt, newgx, sz.height);
	g2d.draw(new Line2D.Float(newgx,sz.height-9,newgx,sz.height-12));
      }
      g.setFont(oldfont);
    }
    if (spectrum != null && spectrum.obs != null) { 
      Rectangle clip = g.getClipBounds();
      double xmin = coor.toX(clip.x);
      double xmax = coor.toX(clip.width+clip.x);
      int imin = spectrum.getIndex(xmin);
      int imax = spectrum.getIndex(xmax);
      double ymin = coor.toY(clip.height+clip.y);
      double ymax = coor.toY(clip.y);
      Peak[] obs = spectrum.obs;
      // Show forward spectrum
      g.setColor(Color.black);
      float oldlx = imin>0 ? coor.fromX(obs[imin-1].x) : -1;
      float olday = imin>0 ? coor.fromYLim(obs[imin-1].la) :  0;
      float oldby = imin>0 ? coor.fromYLim(obs[imin-1].lb) :  0;
      //Move ymax test out here!! Or even one level higher as
      //clip.y<=sz.height-gyoff
      for (int i=imin; i<=imax; i++) {
	float gxi = coor.fromX(obs[i].x);
	if ( ymax >= 0 && ymin <= obs[i].y) {
	  float gy0 = coor.fromYLim(0);
	  float gyi = coor.fromYLim(obs[i].y);
	  if (obs[i].kind != Peak.ORIGINAL)
	    g.setColor(obs[i].kind == Peak.MIRRORED ||
		       obs[i].kind == Peak.TRANSFORMED ? 
		       Color.orange : Color.darkGray);
	  if (obs[i].ignore)
	    g.setColor(Color.lightGray);
	  g2d.draw(new Line2D.Float(gxi, gy0, gxi, gyi));
	  if (obs[i].kind != Peak.ORIGINAL || obs[i].ignore)
	    g.setColor(Color.black);
	}
	//show local abundancelevel: note that it is not correct to
	//interpolate linearly here in the case of the sliding window
	//order statistics based definition of the local levels, but
	//the error will only be seen if the spectrum has large holes.
	if (showLocalLevels) {
	  Color cs = g.getColor();
	  g.setColor(Color.red);
	  float gyla = coor.fromYLim(obs[i].la);
	  float gylb = coor.fromYLim(obs[i].lb);
	  if (oldlx>-1 /*&& ymax >= gylb && ymin <= gyla*/) {
	    g2d.draw(new Line2D.Float(oldlx, olday, gxi, gyla));
	    g2d.draw(new Line2D.Float(oldlx, oldby, gxi, gylb));
	  }
	  oldlx = gxi;
	  olday = gyla;
	  oldby = gylb;
	  g.setColor(cs);
	}
      }
      if (peaksLabelled>0) {
	//double peakLimit = peakLimit(peaksLabelled);
	FontMetrics fontmetrics = g.getFontMetrics();
	double xtop = coor.toX(sz.width);
	for (int j=0; j<peaksLabelled; j++) {
	  Peak p = sortedAbundances[j];
	  //Only draw string if peak is visible
	  if (p.x>=xoff && p.x<=xtop) {
	    float gxi = coor.fromX(p.x);
	    float gyi = coor.fromY(p.y);
	    String s = fmt1.format(p.x);
	    int gwidth = fontmetrics.stringWidth(s);
	    g2d.drawString(s, gxi-gwidth/2, gyi-2);
	  }
	}
      }
      if (controller.proposalValid()) { 
	// Show DP candidate

	// This should be a draw method of Proposal - like for
	// Sequences below, perhaps even called via controller
	g.setColor(Color.magenta);
	for (int side = -1; side <2; side +=2) {
	  Proposal candidate = controller.currentProposal(side<0);
	  if (candidate != null) {
	    Iterator iter = candidate.iterator();
	    while (iter.hasNext()) {
	      Peak p = ((AcidWithPeak) iter.next()).peak;
	      float gxi = coor.fromX(p.x);
	      if (0 <= gxi && gxi < sz.width) {
		float gyi = coor.fromYLim(p.y);
		g2d.draw(new Line2D.Float(gxi,gyi,gxi+side*5,gyi-5));
	      }
	    }
	  }
	}
      }
      if (showBackwards) {
	// Show backward spectrum 
	g.setColor(Color.yellow);
	IonType ion = IonType.B;
	int i1 = spectrum.getIndex(ion.mapToY(xmax,spectrum.totalMass));
	int i2 = spectrum.getIndex(ion.mapToY(xmin,spectrum.totalMass));
	imax = Math.max(i1,i2);
	imin = Math.min(i1,i2);
	float gy0 = coor.fromYLim(0);
	for (int i=imin; i<=imax; i++) {
	  float gxi = coor.fromX(ion.mapFromY(obs[i].x,spectrum.totalMass));
	  if (ymax >= 0 && ymin <= obs[i].y && 
	      0 <= gxi && gxi < sz.width &&
	      obs[i].orig_index >= 0) {
	    float gyi = coor.fromYLim(obs[i].y);
	    g2d.draw(new Line2D.Float(gxi, gy0, gxi, gyi));
	  }
	}
      }
      if (!showExtra.equals("None")) {
	// Show extra spectrum
	g.setColor(Color.pink);
	IonType ion = IonType.get(showExtra);
	//assert ion != null
	int i1 = spectrum.getIndex(ion.mapToY(xmax,spectrum.totalMass));
	int i2 = spectrum.getIndex(ion.mapToY(xmin,spectrum.totalMass));
	imax = Math.max(i1,i2);
	imin = Math.min(i1,i2);
	float gy0 = coor.fromYLim(0);
	for (int i=imin; i<=imax; i++) {
	  float gxi = coor.fromX(ion.mapFromY(obs[i].x,spectrum.totalMass));
	  if (ymax >= 0 && ymin <= obs[i].y && 
	      0 <= gxi && gxi < sz.width &&
	      obs[i].orig_index >= 0) {
	    float gyi = coor.fromYLim(obs[i].y);
	    g2d.draw(new Line2D.Float(gxi, gy0, gxi, gyi));
	  }
	}
      }
      g.setColor(Color.black);
    }
    if (sequences != null) {
      Font oldfont = g.getFont();
      if (isPrinting) 
	g.setFont(new Font("Monospaced",Font.PLAIN,10));
      sequences.paint(g2d, isPrinting, coor);
      if (isPrinting)
	g.setFont(oldfont);
    }
  }

  //----------------------------------------------------------------------
  //Printing support

  public Action getPrintAction() {
    return printAction;
  }
  private Action printAction = new AbstractAction("Print") {
      PrinterJob pj;
      PageFormat pf;
      {
	this.setEnabled(false);
      }
      public void actionPerformed(ActionEvent e) {
	if (pj == null) {
	  pj = PrinterJob.getPrinterJob();
	  pf = pj.defaultPage();
	  //Bad, bad Sun that forces one to do this. Is there a
	  //way to set default paper = A4 in a properties file?
	  Paper a4 = new Paper();
	  a4.setSize(595.0,842.0);
	  pf.setPaper(a4);
	}
	PageFormat newpf=pj.pageDialog(pf);
	if (newpf != pf) {
	  pf = newpf;
	  pj.setPrintable(SpectrumWindow.this,newpf);
	  if(pj.printDialog()){
	    try { 
	      pj.print(); 
	    } 
	    catch (PrinterException pex ) { 
	      controller.debugmessage(pex.toString());
	    }
	  }
	}
      }
    };
  
  public int print(Graphics g, PageFormat pf, int pi)
    throws PrinterException {
    if (pi >= 1 || spectrum==null) {
      return Printable.NO_SUCH_PAGE;
    }
    Graphics2D g2 = (Graphics2D) g;
    g2.translate(pf.getImageableX(), pf.getImageableY());
    g2.setStroke(new BasicStroke(0.12f)); //1 pixel at 600dpi
    Font f = new Font("Monospaced",Font.PLAIN,10);
    g2.setFont(f);
    
    String styl = "<style>td {font-size : 7pt;}</style>";

    final StringBuffer parmsText = 
      new StringBuffer("<html><head>" +
		       styl +
		       "</head><body>" + 
 		       "<table border=0 cellpadding=0 cellspacing=1>" +
		       "<tr><td><b>Printed:</b><td>" + (new Date()) + 
		       "<tr><td><b>Spectrum</b><td>" + spectrum.name + 
		       "<tr><td><b>Sequence</b><td>" + controller.getSeqOut()+ 
		       controller.seqParms.toHTML() +
		       "</table></body></html>");
    JLabel parms = new JLabel(parmsText.toString());
    parms.setSize(parms.getPreferredSize());
    parms.paint(g2);
    
    int h = parms.getHeight()+5; 
    g2.translate(0,h);
    paintComponent(g2,true,new Dimension((int)pf.getImageableWidth(),
					 (int)pf.getImageableHeight()-h));
    return Printable.PAGE_EXISTS;
  }

  //-------------------------------------------------------------------------

  public Dimension getPreferredSize() { return initialSize; }

  public Dimension getMinimumSize() { return initialSize; }

  public Spectrum getSpectrum() { return spectrum; }

  public Log getLog() { return log; }

  /**
   * Inner class SpectrumWindow.Coordinates
   *
   * Convert between window (pixel) coordinates, and true coordinates
   * 
   */

  class Coordinates implements AbstractCoordinates {
    int gwidth, gheight;
    double xscale, yscale;
    final private static double LIMIT = 2.1;
    final private static int TOPOFFSET = 40;

    Coordinates(Dimension sz) {
      gwidth = sz.width; gheight = sz.height;
      xscale = gwidth / width * zoom; 
      if (logScale)
	yscale = (gheight-gyoff-TOPOFFSET)/
	  Math.log(LIMIT/controller.seqParms.grass); 
      else
	yscale = gheight / height;
    }

    // From window (pixel) coordinates to true coordinates

    public double toX(int gx) 
    { return (gx - gxoff) / xscale + xoff; }

    public double toY(int gy) {
      if (logScale) 
	return spectrum.maxAbundance*Math.exp((TOPOFFSET-gy)/yscale);
      else
	return (gheight - gy - gyoff) / yscale + yoff; 
    }
    // From true coordinates to window (pixel) coordinates; same with limits

    public float fromX(double x) 
    { return (float)((x - xoff) * xscale + gxoff); }

    public float fromY(double y) {
      if (logScale)
	return (float)(y>spectrum.maxAbundance*controller.seqParms.grass/LIMIT ? 
		       TOPOFFSET+yscale*Math.log(spectrum.maxAbundance/y): 
		       gheight-gyoff);
      else
	return (float)(gheight - (y - yoff) * yscale - gyoff); 
    } 

    public float fromXLim(double x) 
    { return Math.max(0, Math.min(gwidth, fromX(x))); }

    public float fromYLim(double y) 
    { return Math.max(0, Math.min(gheight, fromY(y))); }

    //distance from (gx,gy) to (fromX(x), fromY(y))
    public double dist(int gx, int gy, double x, double y) {
      return sqr(fromX(x)-gx) + sqr(fromY(y)-gy);
    }
  }
  private static double sqr(double x) {
    return x * x;
  } 

}
