import java.io.*;
import java.util.*;


// The current implementation of class SeqScore really bites
// here. Each object has fields of 6 doubles, 1 int and 1 object
// reference. Each SequencePair object has 2 of those, while 1 double
// should suffice(*). This seems to give a factor about 20 in memory
// consumption, roughly corresponding to a factor 5 in the number of
// peaks that may be handled. 
//
// (*) if we were to let independent_errors be false, we might have to
// remember two doubles containing the predicted masses given by
// adding together the monoW.


// We have never really needed the full power of the SeqScore. What we
// should have would be an interface that SequencingParameters would
// implement:

interface ScoreMethod {
  //peakscore is the mass-independent penalty for the abundance
  double peakscore(Peak p);
  //acidscore contains the premium for adding an acid (1 if beta=0)
  //and the penalties for mass misfit and the KRcorrection at the end.
  //We need p1 or p2 for checking the end condition. If
  //independent_errors is false we need mass_error (or a predicted
  //mass) else we could do with p1 and p2.
  double acidscore(AminoAcid a, double mass_error, Peak p1, Peak p2);
  //We would only call acidscore to plumb the whole in a SequencePair,
  //while the rest of chenlike() and denovodp would use
  double stepscore(AminoAcid a, double mass_error, Peak p1, Peak p2);
  //that simply does peakscore(p1)+acidscore(...)
}



/* 

 Let a spectrum be given. p_bot and p_top are the synthetic peaks at
 the bottom and top corresponding to imagined empty and full Y ions.

 Assume that all Y-ions are present and most B-ions too, possibly
 because the spectrum has been created by Spectrum.mirror().

 Also assume we have chosen a procedure to recognize "duplicates" in
 the form of a function "boolean~avoid(Peak~p1,Peak~p2)" that returns
 true if p1 and p2 correspond (probably) to the same position int the
 peptide. The avoid() function could e.g. be implemented by comparing
 p1.orig_index to p2.orig_index (relevant if spectrum.isMirrored), or
 it could be implemented as a check if p2.x is within (some) tolerance
 of the mirror of p1.x. Assume that avoid(p1,p2) is true at least when
 p1.x and p2.x are exact mirrors of each other.

 We want to find the Sequence from p_bot to p_top that has maximal
 score among Sequence's for which all pairs (p1,p2) of peaks satisfy
 !avoid(p1,p2).

 It turns out (Chen et al) that one should work inwards from both ends
 simultaneously instead of working leftwards or rightwards from either
 end.

 The basic idea is the following: Consider the maximizing Sequence. If
 we split it in two by removing the amino acid that bridges, x_c, the
 centre of the spectrum (x_c is its own mirror), we get a pair of
 sequences, (s1:p_bot->pi,s2:p_top->pj), with the property that every
 peak p1 from s1 and every peak p2 from s2 satisfies !avoid(p1,p2).
 Say that s1 and s2 are compatible if they have this property.
 Moreover the pair (s1,s2) has maximal sum of scores among compatible
 pairs of sequences with the same endpoints. This looks like the basis
 of a dynamic programming algorithm.

 So, for each pair of peaks pi<x_c<pj we want to find the pair
 (s1:p_bot->pi,s2:p_top->pj)=SP(pi,pj) - if any - that maximizes the
 sum of scores for compatible pairs of sequences ending at pi,pj. Let
 d(pi,pj)=min(x_c-pi,pj-x_c), the distance from the centre of the
 spectrum to {pi,pj}. Since !avoid(pi,pj) the distances are different:
 x_c-pi!=pj-x_c.  Assume that we know SP(pi',pj') for every pair with
 larger distance from the centre. Assume, without loss of generality,
 that pi is closer to the centre than pj. For every pi'<pi which is a
 monoW away from pi within tolerance, we already know SP(pi',pj). We
 may add pi to each such SP(pi',pj) and select the highest scoring
 sequence pair.

 We now start with the pair (p_bot,p_top) and iteratively compute the
 map {(pi,pj)->SP(pi,pj):d(pi,pj)>=t} for smaller and smaller t
 running through all values of the distance from a peak to x_c. 

 Now, for each amino acid, and each compatible pair pi<x_c<pj that is
 monoW apart within tolerance, we may connect the two parts of
 SP(pi,pj) (if it exists) to get a candidate for a global
 Sequence. Comparing all such candidates will give the maximizer (if
 any).



 The exact algorithm for computing the full {(pi,pj)->SP(pi,pj)} might
 use the following idea to avoid considering a lot of impossible
 choices of pi or pj. Maintain at every time two (sorted) lists,
 Pot_bot and Pot_top, of potential, but as yet not considered choices
 for pi and pj. Start with Pot_bot={p_bot} and Pot_top={p_top}. We
 also maintain sorted lists, C_bot and C_top, of peaks already
 considered - thus we know SP(pi,pj) whenever (pi,pj) is in
 (C_bot)x(C_top) and mutually compatible. At each step, take the peak,
 p, in (Pot_bot)union(Pot_top) farest away from x_c. Assume without
 loss of generality that p is from Pot_bot. Compute SP(p,pj) for every
 pj in C_top and add all peaks less than x_c and a monoW+-tolerance
 above p to Pot_bot. Note that for efficiency purposes, P_bot and
 P_top should contain not only peaks, but tracebacks to the peaks that
 nominated them. When the full {(pi,pj)->SP(pi,pj)} has been computed,
 we may search through C_top and C_bot for differences within
 tolerance of a monoW to find potential bridges.


 Note the above seems to be quite dependent on a good value for totalMass!


*/

class ProposalQ {
  private TreeSet tree;
  private int maxRank;
  ProposalQ(int N) {
    maxRank = N;
    tree = new TreeSet();
  }
  void add(Proposal tb) {
    if (tree.size() == maxRank) {
      Proposal last = (Proposal) tree.last();
      if (tb.compareTo(last) < 0) 
	tree.remove(last);
      else 
	return;
    }
    tree.add(tb);
  } 
  void addAll(Proposal[] tbs) {
    if (tbs !=null)
      for (int i=0; i<tbs.length; i++)
	add(tbs[i]);
  }
  Proposal[] toArray() {
    if (tree.size() == 0) 
      return new Proposal[0];
    Proposal[] res = new Proposal[tree.size()];
    tree.toArray(res);
    return res;
  }
  Iterator iterator() {
    return tree.iterator();
  }
  int size() {
    return tree.size();
  }
}
  


abstract class chenlike implements ProposalAlgo {
  private final double x_c;
  private final double tolerance;
  private final double grass;
  
  private ProposalQ proposalQ = new ProposalQ(10);
  
  //This is the only method lacking implementation:
  abstract boolean avoid(Peak pi, Peak pj);
  
  class PeakAcidPair {
    Peak p;
    AminoAcid acid;
    PeakAcidPair(Peak p, AminoAcid acid) {
      this.p = p;
      this.acid = acid;
    }
  }

  //This is the generalization of Endpoint from DenovoDP
  static class SequencePair implements Comparable {
    private static int lastSPTId;
    private int id;               //To make ordering total
    Peak pi;                   // always < x_c
    Peak pj;                   // always > x_c
    Score bot_score;
    Score top_score;
    boolean left;              // last side added
    SequencePair prev;// null means this is the start of the pair
    AminoAcid acid;            // null if prev is null
    //
    SequencePair(Peak pi, Peak pj, Score bot_score, Score top_score, 
			  boolean left,
			  SequencePair prev, AminoAcid acid) {
      id = lastSPTId++;
      this.pi = pi;
      this.pj = pj;
      this.bot_score = bot_score;
      this.top_score = top_score;
      this.left = left;
      this.prev = prev;
      this.acid = acid;
    }
    //Note that we cannot utilize the compareTo method of class Score,
    //but must use the sum of the numeric score() values of the Score
    //objects
    public int compareTo(Object o) {
      SequencePair it = (SequencePair) o;
      double my_score = score();
      double its_score = it.score();
      //if the scores happens to be equal, prefer (arbitrarily) the
      //element created first
      return 
	my_score>its_score ? -1 : 
	my_score<its_score ? 1 :
	id - it.id;
    }
    double score() {
      return bot_score.score()+top_score.score();
    }
    public String toPeptide(String c) {
      StringList res = new StringList();
      res.add("<"+c+">");
      SequencePair spp = this;
      while (spp.prev != null) {
	if (spp.left)
	  res.addFirst(spp.acid.abbr);
	else
	  res.addLast(spp.acid.abbr);
	spp = spp.prev; 
      }
      return res.toStringBuffer(false).toString();
    }
    public Proposal toProposal(AminoAcid bridging_acid) {
      LinkedList acids = new LinkedList();
      LinkedList peaks = new LinkedList();
      acids.add(bridging_acid);
      SequencePair spp = this;
      while (spp.prev != null) {
	if (spp.left) {
	  acids.addFirst(spp.acid);
	  peaks.addFirst(spp.pi);
	}
	else {
	  acids.addLast(spp.acid);
	  peaks.addLast(spp.pj);
	}
	spp = spp.prev; 
      }
      peaks.addLast(spp.pj);
      LinkedList proposal = new LinkedList();
      proposal.add(new AcidWithPeak(spp.pi, null));
      while (!acids.isEmpty()) 
	proposal.add(new AcidWithPeak((Peak)(peaks.removeFirst()),
				      (AminoAcid)(acids.removeFirst())));
      Score pi_score = bot_score.makenew(pi);
      double bridging_error = pi.x+bridging_acid.monoW-pj.x;
      double bridging_score = 
	pi_score.add(pj, bridging_acid, bridging_error).score() 
	-pi_score.score() 
	-bot_score.makenew(pj).score();
      double totalscore = score() + bridging_score;
      return new Proposal(proposal, totalscore);
    }
  }

  //This is a finite-max-length priority queue of
  //SequencePairs. Just a utility wrapper for TreeSet.
  static class SequencePairQ {
    private TreeSet tree;
    private int maxRank;
    SequencePairQ(int N) {
      maxRank = N;
      tree = new TreeSet();
    }
    void add(SequencePair tb) {
      if (tree.size() == maxRank) {
	SequencePair last = (SequencePair) tree.last();
	if (tb.compareTo(last) < 0) 
	  tree.remove(last);
	else 
	  return;
      }
      tree.add(tb);
    } 
    void addAll(SequencePair[] tbs) {
      if (tbs !=null)
	for (int i=0; i<tbs.length; i++)
	  add(tbs[i]);
    }
    SequencePair[] toArray() {
      if (tree.size() == 0) 
      return new SequencePair[0];
      SequencePair[] res = new SequencePair[tree.size()];
      tree.toArray(res);
      return res;
    }
    Iterator iterator() {
      return tree.iterator();
    }
    int size() {
      return tree.size();
    }
  } 

  

  //To compare two peaks wrt closeness to center of spectrum. We will
  //only compare objects that are peaks belonging to
  //spectrum.obs. When distance from center equal, let the ordinary
  //peak comparison decide.
  private Comparator peakComparator = new Comparator() {
      public int compare(Object o1, Object o2) {
	Peak p1 = (Peak) o1;
	Peak p2 = (Peak) o2;
	double d1 = Math.abs(p1.x-x_c);
	double d2 = Math.abs(p2.x-x_c);
	int retval;
	if (d1<d2)
	  retval = 1;
	else if (d1>d2)
	  retval = -1;
	else 
	  retval = p1.compareTo(p2);
	return retval;
      }
    };

  private TreeMap potential_combined; //Peak to traceback
  private TreeMap considered_bot;   //Peak to TreeMap defining SP: 
  //C_bot{p1}{p2} is a SequencePairQ
  private TreeMap considered_top; //C_top{p2}{p1}==C_bot{p1}{p2}

  //After having implemented a variant of the algorithm, one will be
  //able to count how many peak C_top and C_bot would contain, and
  //from this estimate whether TreeMaps or just extra fields on Peak
  //(as PotPeak above) would be best. One could have
  //potential_combined be new~LinkedList[obs.length] and run through
  //it using two indices. It seems plausible that a few hundred amu
  //away from the endpoints, every peak will be in potential_combined
  //- this should be tested. One could have considered be
  //SequencePairQ[obs.length][obs.length]

  //Amino Acid data
  private static AminoAcid[] sortedAcids;
  private static double maxMonoW;
  static {
    sortedAcids = 
      (AminoAcid[]) AminoAcid.aaByMonoW.toArray(new AminoAcid[0]);
    maxMonoW = sortedAcids[sortedAcids.length-1].monoW;
  }
  
  chenlike(Spectrum spectrum, Score score, Feedback feedback) {
    x_c = spectrum.totalMass/2+Element.H.monoW;
    tolerance = score.getTolerance();
    grass = score.getGrass();
    double a_limit = spectrum.maxAbundance*score.getGrass();
    //
    
    potential_combined = new TreeMap(peakComparator);
    considered_bot = new TreeMap(peakComparator);
    considered_top = new TreeMap(peakComparator);

    //Silly: we have to search for p_bot and p_top here, so we might
    //want to add them as fields in Spectrum.
    //OR: we could supply those as input to the constructor
    Peak[] obs = spectrum.obs;
    Peak p_bot=null, p_top=null;
    for (int i=0; i<obs.length; i++) {
      Peak p = obs[i];
      if (p.orig_index == Peak.BOTTOM_OI)
	p_bot = p;
      if (p.orig_index == Peak.TOP_OI)
	p_top = p;
    }
    //
    potential_combined.put(p_bot,null);
    potential_combined.put(p_top,null);
    //
    SequencePair startSPT = 
      new SequencePair(p_bot, p_top, 
		       score.makenew(p_bot), 
		       score.makenew(p_top),
		       true, null, null);
    SequencePairQ botQ = new SequencePairQ(10);
    SequencePairQ topQ = new SequencePairQ(10);
    botQ.add(startSPT);
    topQ.add(startSPT);
    TreeMap botMap = new TreeMap();
    botMap.put(p_top,botQ);
    TreeMap topMap = new TreeMap();
    topMap.put(p_bot,topQ);
    considered_bot.put(p_bot,botMap);
    considered_top.put(p_top,topMap);

    //First part of the DP: we work our way form the endpoints to the
    //middle, hoping we will be able to get all the way:
    while (!potential_combined.isEmpty()) {
      Peak ppppp = (Peak) potential_combined.firstKey();
      //Usual sign convention in AminoSequencer for side(?):
      int side = ppppp.x < x_c ? -1 : 1;
      TreeMap considered = side < 0 ? considered_bot : considered_top; 
      TreeMap considered_other = side > 0 ? considered_bot : considered_top; 
      LinkedList papList = (LinkedList) potential_combined.remove(ppppp);

      double partnerLimit = 2*x_c-ppppp.x-side*(maxMonoW+tolerance);

      //For each listed peak-acid pair, get all the SP's, add pppp to
      //all of them if compatible and add the results to
      //considered.
      //
      //This block is entered for every ppppp other than p_bot and p_top
      if (papList !=null) {
	//the partners->SPs to be attached to ppppp:
	TreeMap pppppMap = new TreeMap();
	//the possible (predecessor,acid) pairs:
	Iterator iter = papList.iterator();
	while (iter.hasNext()) {
	  PeakAcidPair pa = (PeakAcidPair) (iter.next());
	  Peak prevPeak = pa.p;
	  AminoAcid acid = pa.acid;
	  //assert considered.get(prevPeak) != null; //else 
	  TreeMap spMap = (TreeMap)(considered.get(prevPeak));
	  if (spMap == null) {
	    System.out.println("spMap null at: "+prevPeak.x);
	    continue;
	  }
	  Iterator partner_iter = spMap.keySet().iterator();
	  while (partner_iter.hasNext()) {
	    Peak partnerPeak = (Peak)(partner_iter.next());
	    if (avoid(ppppp, partnerPeak))
	      continue;
	    if (side*(partnerPeak.x - partnerLimit) < 0) {
	      continue;
	    }
	    SequencePairQ foo = (SequencePairQ)spMap.get(partnerPeak);
	    TreeMap partnerMap = (TreeMap)(considered_other.get(partnerPeak));
	    //System.out.println("  "+partnerPeak.x+" "+foo.size());
	    Iterator sp_iter = foo.iterator();
	    while (sp_iter.hasNext()) {
	      SequencePair sptb = (SequencePair)sp_iter.next();
	      //assert {sptb.pi, sptb.pj} == {prevPeak, partnerPeak}
	      //System.out.println("        "+sptb.pi.x+" "+sptb.pj.x);
	      //
	      //Temporarily, we only consider independent mass errors:
	      double mass_error = ppppp.x-prevPeak.x+side*acid.monoW;
	      Score tempscore = side<0?sptb.bot_score:sptb.top_score;
	      Score newscore = tempscore.add(ppppp, acid, mass_error);
	      SequencePair newsptb = 
		new SequencePair((side<0?ppppp:sptb.pi),(side<0?sptb.pj:ppppp),
				 (side<0?newscore:sptb.bot_score),
				 (side<0?sptb.top_score:newscore),
				 side<0,sptb,acid);
	      //Add it to both considered_xxx!!!
	      SequencePairQ foo1 =(SequencePairQ)pppppMap.get(partnerPeak);
	      if (foo1 == null) {
		foo1 = new SequencePairQ(10);
		pppppMap.put(partnerPeak,foo1);
	      }
	      foo1.add(newsptb);
	      //Other way round:
	      foo1 = (SequencePairQ)partnerMap.get(ppppp);
	      if (foo1 == null) {
		foo1 = new SequencePairQ(10);
		partnerMap.put(ppppp,foo1);
	      }
	      foo1.add(newsptb);
	    }
	  }
	}
	//This should never happen: there will always be compatible
	//partners for ppppp. NONONO: after we have optimized by
	//requiring that the partner may at most be maxMonoW further
	//away from x_c than ppppp this may well happen.
	if (pppppMap.isEmpty()) {
	  //System.out.println("Dead end: "+ppppp.x);
	  continue;
	}
	//else:
	considered.put(ppppp,pppppMap);
	
      }
      //For all possible successors of ppppp, add it to potential_combined:

      //This loop over peaks and acid should be optimized by the same
      //methods as DenovoDP
      for (int a=0; a<sortedAcids.length; a++) {
	//Optimal order depends on side
	double monoW = sortedAcids[a].monoW;
	double nextMass = ppppp.x - side*monoW;
	//If nextMass is on the other side of or very near centre:
	if ((nextMass - x_c)*side < tolerance)
	  continue;
	int imin = spectrum.getIndex(nextMass-tolerance);
	int imax = spectrum.getIndex(nextMass+tolerance);
	for (int i=imin; i<=imax; i++) {
	  Peak nextPeak = obs[i];
	  //This is really just because of the possibility that
	  //imin==imax==obs.length-1
	  if (Math.abs(nextPeak.x-nextMass)>tolerance)
	    continue;
	  //Check for grass condition!!
	  if (nextPeak.y <= a_limit /*0.1*nextPeak.la*/ )
	    continue;
	  LinkedList nextPapL = (LinkedList) potential_combined.get(nextPeak);
	  if (nextPapL == null) {
	    nextPapL = new LinkedList();
	    potential_combined.put(nextPeak, nextPapL);
	  }
	  nextPapL.add(new PeakAcidPair(ppppp,sortedAcids[a]));
	}
      }
      


    }
    //
    // Now we have computed SP(pi,pj) in the form of considered_xxx
    
    //
    double bot_th = x_c - maxMonoW - tolerance; 
    SortedMap relevant_bot = considered_bot.tailMap(new Peak(bot_th,0,0,-3,-3));

    Iterator pi_iter = relevant_bot.keySet().iterator();
    while (pi_iter.hasNext()) {
      Peak pi = (Peak)(pi_iter.next());
      TreeMap partners = (TreeMap) (considered_bot.get(pi));
      double limit = pi.x + maxMonoW+tolerance; 
      SortedMap relevant_partners=partners.headMap(new Peak(limit,0,0,-3,-3));
      Iterator pj_iter = relevant_partners.keySet().iterator();
      while (pj_iter.hasNext()) {
	Peak pj = (Peak)(pj_iter.next());
	for (int a=0; a<sortedAcids.length; a++) 
	  if (Math.abs(pi.x+sortedAcids[a].monoW-pj.x)<tolerance) {
	    SequencePairQ sptbq = (SequencePairQ)(partners.get(pj));
	    Iterator pair_iter = sptbq.iterator();
	    while (pair_iter.hasNext()) {
	      SequencePair sptb = (SequencePair)(pair_iter.next());
	      proposalQ.add(sptb.toProposal(sortedAcids[a]));
	    }
	  }
      }
    }
    System.out.println("Proposals: "+proposalQ.size());
    Iterator pqiter = proposalQ.iterator();
    while (pqiter.hasNext()) {
      System.out.println(((Proposal)(pqiter.next())).toString());
    }
    

  } 


  public static void main(String[] args) throws IOException {
    String filename = 
      args.length >0 ? args[0] : "1317.txt";//"QTof_ETYGDMADCCEK.dta";
    Spectrum s = Spectrum.readFile(filename);
    s.mirror(null);
    SequencingParameters seqParms = new SequencingParameters();
    seqParms.grass = 0.02;
    Score score = new SeqScore(s.maxAbundance,s.totalMass,seqParms);    
    //System.out.println((new DenovoDP(s,score,null)).toString());
    new chenlike(s, score, null) {
	boolean avoid(Peak pi, Peak pj) {
	  return pi.orig_index == pj.orig_index;
	}
      };
  }
  
  //The ProposalAlgo interface
  public Proposal getProposal(int i, boolean left, int rank) {
    return null;
  }
  public String getSequence(int i, boolean left, int rank, boolean bwdSequence) {
    return "FOO";
  }
  public String getSequence(int i, int lr, int rr,boolean bwdSequence) {
    return "BAR";
  }
  public int getProposalCount(int i, boolean left) {
    return 0;
  }

  
}
