
import java.io.*;
import java.util.*;


/**
 * class Peptide only used for generating spectra
 */
class Peptide {
  private AminoAcid[] acids;
  Peptide(String[] abbrs) {
    acids = new AminoAcid[abbrs.length];
    for (int i=0; i<abbrs.length; i++)
      acids[i] = (AminoAcid)AminoAcid.getAA(abbrs[i]);
  }
  Peptide(String peptide) {
    LinkedList acidlist = new LinkedList();
    int i=0; 
    while (i<peptide.length()) {
      int j = i+1;
      while (j<peptide.length() && !Character.isLetter(peptide.charAt(j)))
	j++;
      AminoAcid aa = AminoAcid.getAA(peptide.substring(i,j));
      if (aa == null)
	throw new IllegalArgumentException("Bad acid abbr: " +
					   peptide.substring(i,j));
      acidlist.add(aa);
      i=j;
    }
    acids = (AminoAcid[]) acidlist.toArray(new AminoAcid[] {});
  }
  public int length() {
    return acids.length;
  }
  public double totalMass() {
    return residueMass(acids.length,true)+IonType.terminalGroupMasses;
  }
  private double residueMass(int i, boolean fromStart) {
    //assert i<=acids.length
    double res = 0.0;
    for (int j=0; j<i; j++) {
      res += acids[fromStart ? j : acids.length-1-j ].monoW;
    }
    return res;
  }
  public double ion(int i, IonType it) {
    return it.toMZ(residueMass(i,it.cterminal));
  }
  public double ion(int i, String iontype) {
    return ion(i,IonType.get(iontype));
  }
  //Another road to the computation of the mass that allows computing
  //isotope dependence!
  public ChemicalFormula ionFormula(int i, IonType it) {
    //assert i<=acids.length
    ChemicalFormula[] residues = new ChemicalFormula[i];
    boolean fromStart = it.cterminal;
    for (int j=0; j<i; j++) {
      residues[j] = acids[fromStart ? j : acids.length-1-j ].formula;
    }
    return it.formula.add(residues);
  }
  public ChemicalFormula ionFormula(int i, String it) {
    return ionFormula(i, IonType.get(it));
  }
  //Generate a spectrum. Imagine adding other peak types, random mass
  //errors, random isotopes!


  /*
    KVALNQALESLAES
    KPGEQADTAESAEGQAAN
    KLDDLDEGE
    
    RDFQMGEVVDLETK
    
    RTANLADELLL
  */


  public static void main(String[] args)
    throws FileNotFoundException {
    String aaString;
    int charge;
    //Rounding is really just for testing the effects of 
    double rounding;
    //
    
    if (args.length == 1) {
      System.out.println("usage: java Peptide <PEPTIDE> <CHARGE> [decimals]");
      return;
    }
    else if (args.length == 0) {
      aaString = "KAWGGTY";
      charge = 2;
      rounding = 1000;
    } 
    else {
      aaString = args[0];
      charge = Integer.parseInt(args[1]);
      rounding = args.length>2 ? Math.pow(10,Integer.parseInt(args[2])): 1000;
    } 
       //
    double sigma = 0.150;
    Peptide peptide = new Peptide(aaString);
    PrintStream out = 
      new PrintStream(new FileOutputStream("Synthetic_"+aaString+".txt"));
    out.println("BEGIN IONS");
    out.println("TITLE=Synthetic "+aaString+" spectrum");
    out.println("PEPMASS="+(peptide.totalMass()/charge+ChemicalFormula.H.monoW));
    out.println("CHARGE="+charge+"+");
    //
    TreeSet obs = new TreeSet();
    int oi = 1;
    double tm = peptide.totalMass();
    Random rnd = new Random(2511348);
    for (int i=1; i<peptide.length(); i++) {
      ChemicalFormula formula = peptide.ionFormula(i,"Y");
      double rel_mass = formula.monoW/tm;
      double yf = 1/(1+0*(rel_mass-0.4)*(rel_mass-0.4));
      double ylevel = 1.0+100.0*rnd.nextDouble()*yf;
      double[][] mass = formula.toIsotopicSpectrum(100,rnd);
      double del = sigma*rnd.nextGaussian();
      for (int j=0; j<mass.length; j++) {
	double x = mass[j][0]+del;
	x = Math.round(rounding*x)/rounding;
	obs.add(new Peak(x, x, ylevel*mass[j][1]/100 , oi++, Peak.ORIGINAL));
      }
      //
      formula = peptide.ionFormula(i,"B");
      rel_mass = formula.monoW/tm;
      yf = 1/(1+0*(rel_mass-0.4)*(rel_mass-0.4));
      ylevel = 1.0+40.0*rnd.nextDouble()*yf;
      mass = formula.toIsotopicSpectrum(100,rnd);
      del = sigma*rnd.nextGaussian();
      for (int j=0; j<mass.length; j++) {
	double x = mass[j][0]+del;
	x = Math.round(rounding*x)/rounding;
	obs.add(new Peak(x, x, ylevel*mass[j][1]/100 , oi++, Peak.ORIGINAL));
      }
      //
      formula = peptide.ionFormula(i,"A");
      rel_mass = formula.monoW/tm;
      yf = 1/(1+0*(rel_mass-0.4)*(rel_mass-0.4));
      ylevel = 1.0+20.0*rnd.nextDouble()*yf;
      mass = formula.toIsotopicSpectrum(100,rnd);
      del = sigma*rnd.nextGaussian();
      for (int j=0; j<mass.length; j++) {
	double x = mass[j][0]+del;
	x = Math.round(rounding*x)/rounding;
	obs.add(new Peak(x, x, ylevel*mass[j][1]/100 , oi++, Peak.ORIGINAL));
      }
    }
    // Random grass
    for (int i=1; i<1000; i++) {
      double rel_mass = 0.4+0.2*rnd.nextGaussian();
      double mass = tm*rel_mass;
      double x = Math.round(rounding*mass)/rounding;
      if (x<19)
	continue;
      double ran = rnd.nextGaussian();
      double y = ran*ran/(2+20*(rel_mass-0.4)*(rel_mass-0.4))+0.001;
      obs.add(new Peak(x, x, y , oi++, Peak.ORIGINAL));
    }
    Iterator iter = obs.iterator();
    while (iter.hasNext()) {
      Peak p = (Peak)(iter.next());
      out.println(p.x+" "+p.y);
    }
  }
}

