// KVL AminoSequencer, a program to support manual de novo sequencing
// of proteins by interpretation of tandem mass spectrometry data (MS/MS).
// Version 1.0 of 2002-01-04

// Authors: Niels Jrgen Kokholm (kokholm@dina.kvl.dk) 
// Peter Sestoft (sestoft@dina.kvl.dk)
// Department of Mathematics and Physics
// Royal Veterinary and Agricultural University
// Thorvaldsensvej 40, DK-1871 Frederiksberg C, Denmark

// This software is under the GNU General Public License, 
// http://www.gnu.org/licenses/

// Thanks to Natascha Helena Beyer <nhbeyer@mail.dk> and Gerhard
// Saalbach <g.saalbach@risoe.dk>

/*
Problems:




*/


/*

To do, ideas, questions:

 - adjust print scaling, if possible
 - are there widely recognized one-letter codes for Mox, Wox, Cys*, Cys**?
 - make the tolerance adjustment more transparent (in the menu)
 - multiply by charge state or not? (lutefisk examples should not be multipl.)
 - overlaps between the backwards and forwards spectra may be good 
   candidates for significant peaks?  Hence, start by toggling the backwards
   spectrum on and off, and look for partially disappearing peaks
 - the main problem seems to be the choice of peaks that represent
   potential amino acids; maybe the tolerance should depend on the
   distribution of abundances, so that tolerance is smaller for low
   peaks and higher for high peaks.  I.e. a low peak is chosen only if
   it matches extremely well, whereas a high peak is 
   Idea: Let h(a) be the fraction of peaks with y >= a; then make the
   tolerance inversely proportional to (1 + hDisc * h(a)).
 - allow user to override the choice of potential aa peaks: e.g. right 
   click on a peak to see what aas are closest, then click again to 
   accept one of them
 - ignore very low peaks when looking for potential amino acids? yes.
 - weight the distance with the abundance, so that high peaks are
   preferred over low ones?  When there are very many (mostly low)
   peaks, there's a high probability that a low peak will match an aa
   mass better than a high one, which is probably misleading.  Hence ,
   e.g. wdist = dist / (1 + aDiscount * abundance)?
 - when there are several potential peaks that are almost equally good, 
   preferably mark the highest one
 - show the masses at some user-selected peaks; how?
 - reasonable tolerances when matching an amino acid to a peak:
   Helena: 0.1, 0.5, 0.8; Lutefisk (Johnson) seems to use 0.45
 - when quitting with an unsaved spectrum, ask user whether he wants to 
   save it
 - how better exploit the reverse spectrum?  
 - would it be useful to log also the masses between the aa?
 - what exactly is it that masslynx can do?  interactive interpretation 
   of the spectrum as well as database searches
 - get more example files (Helena, Gerhard?, Kristian?)
 - how exploit the natural isotope 34S (in the M and C aa): highlight 
   likely double 32S/34S peaks?
 - highlight peaks that are 1 or 2 Daltons apart, as the correspond 
   to isotopes --- how useful is this?
 - why is the drawing of mass numbers so slow in Windows?  
   Apparently it is drawString itself, not the stringWidth thing
 - create an applet version that can run under JDK 1.1 (write own 
   LinkedList and HashSet implementations, possibly using HashTable);
   make provision for loading a DTA file; trim down the AminoAcids 
   class for this purpose
 - isotopes, thanks to Eva Danielsen:
   C  98.89% 12.0000; 1.11% 13.00335
   N  99.63% 14.00307; 0.37% 15.00011
   O 99.759% 15.99491; 0.204% O-18
   S  95% 31.97207; 4.22% 33.96786; 0.76% 32.97146
   H - 99.985% 1.007825 - 

   concerning abundances cf also
   http://www.physics.curtin.edu.au/iupac/docs.html

   concerning isotopic masses cf also
   http://www.nea.fr/html/dbdata/mass_rmd_mas95.txt and
   http://iaeand.iaea.or.at/ndspub/masses/
*/

/* USER MANUAL
   -----------

   This program supports manual de novo sequencing of proteins by
   interpretation of tandem mass spectrometry data (MS/MS).

   Start the program on file myspectrum.dta by executing 
	java Sequencer myspectrum.dta
   or
	java Sequencer 

   This command must be executed on the Unix or MS DOS command line. 


   The spectrum window
   -------------------
  
   The amino acid sequence is found by marking peaks in the spectrum.
   At any point during this process, the spectrum will be decorated
   with zero or more subsequences, each consisting of:

     * a red mark: the starting point
     * zero or more green marks: accepted amino acids
     * zero or more blue marks: potential amino acids

   with green marks on either side of the red mark, and blue marks
   only on the left and right fringes.  The accepted (green) amino
   acids of two subsequences may meet but cannot overlap.  The
   potential (blue) amino acids belonging to one subsequence may
   overlap with potential (blue) amino acids of another sequence.  If
   one peak is decorated with potential amino acids from two different
   sequences, then clicking on it will accept both and make the
   sequences meet in that peak.

   Building a sequence
   -------------------

   (1) Click on some peak to select it as the starting point; it will
       be shown in red.

       Every neighbouring peak whose distance from the selected one
       corresponds to the mass of an amino acid is a potential new
       peak to select, and will be labelled in blue.

   (2) Clicking on a potential (blue) amino acid will color it green,
       remove all the blue ones relating to the same previously
       accepted (green) amino acid, and add new potential (blue) amino
       acids corresponding to the recently accepted (green) one.  The
       best-matching potential amino acid is highlighted with a red
       ring.

   Changing one's mind
   -------------------

   (3) Clicking on an accepted (green) amino acid will backtrack to
       the time just before is was accepted.  More precisely, all blue
       and green marks further out will be removed, and blue marks
       will be added corresponding to the mark most recently accepted
       before this one.
   
   (4) Clicking on the red mark will remove all markings altogether,
       on both sides, and will allow you to select a new starting
       point.

   Keystroke commands
   ------------------

   z	   zoom in around the cursor
   o	   zoom out around the cursor
   c       recenter around the cursor
   r       reset to full view of the spectrum
   arrows  step left, right
   m       show masses of 10 more peaks; may be repeated
   n       hide masses of 10 more peaks; may be repeated

   Menu commands
   -------------

   File | Open		           ^O   open file and read a new spectrum
   File | Save log	           ^S   save log to file
   File | Print		           ^P   print the spectrum window
   File | Quit		           ^Q   quit the program
   Settings | Backwards            ^B   show spectrum backwards also
   Settings | Allow K and R        ^A   allow/disallow K and R
   Settings | Decrease tolerance   ^D   decrease match tolerance (in amu)
   Settings | Increase tolerance   ^T   increase match tolerance (in amu)
   Help | Help                     ^F1  show brief help
   Help | About                         show brief program information
   Help | Amino Acids
   Help | Ion Types

   The program uses the monoisotopic masses, obtained 
   from http://molbiol.soton.ac.uk/mass-spec/amino2.htm 
   and  http://www.rmsb.u-bordeaux2.fr/rmsb/ms/AA-masses.html; 
   but some differ by 0.00001 Dalton from those of Protana's AminoCalc.  
*/


import java.util.*;
import java.text.*;
import java.io.*;
import java.awt.print.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.lang.reflect.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;

// ----------------------------------------------------------------------
/** 
 * Sequencer contains the main() entry point and constructs the (main)
 * user interface components. 
 */

class Sequencer {
  public static void main(String[] args) throws IOException {
    //try {
      //String plafName = "com.sun.java.swing.plaf.motif.MotifLookAndFeel";
      //String plafName = "com.sun.java.swing.plaf.metal.MetalLookAndFeel";
      //String plafName = "com.sun.java.swing.plaf.windows.WindowsLookAndFeel";
      //UIManager.setLookAndFeel(plafName);
    //} catch (Exception e) { System.out.println(e); }
    final String progname = "KVL AminoSequencer";
    final LockableJFrame f = new LockableJFrame(progname);
    f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

    JTextField mDelta = new JTextField(10);
    JTextField lDist = new JTextField(10);
    JTextField rDist = new JTextField(10);
    JTextField xCoor = new JTextField(8);
    JTextField yCoor = new JTextField(8);
    JTextField ionTypeMatch = new JTextField(40);
    mDelta.setMaximumSize(new Dimension(8,20));
    lDist.setMaximumSize(new Dimension(10,20));
    rDist.setMaximumSize(new Dimension(10,20));
    xCoor.setMaximumSize(new Dimension(8,20));
    yCoor.setMaximumSize(new Dimension(8,20));
    ionTypeMatch.setMaximumSize(new Dimension(40,20));
    mDelta.setEditable(false); 
    lDist.setEditable(false); 
    rDist.setEditable(false); 
    xCoor.setEditable(false);
    yCoor.setEditable(false);
    ionTypeMatch.setEditable(false);
    final JTextField seqOut = new JTextField(40);
    seqOut.setEditable(false);
    final JTextField msgField = new JTextField(60);
    msgField.setEditable(false);
    //

    final Observable fileObservable = new Observable() {
	public void notifyObservers() {
	  setChanged();
	  super.notifyObservers();
	}
      };
    //
    final Controller controller =
      new Controller(lDist,rDist,xCoor, yCoor, msgField, ionTypeMatch, 
		     seqOut, mDelta, fileObservable, f);
    final SpectrumWindow sw = 
      new SpectrumWindow(new Dimension(600, 400), controller);
    //    f.addKeyListener(sw);
    final ProposalPanel propP = new ProposalPanel(controller);
    final DPParmsFrame dpparmsFrame = new DPParmsFrame(controller);
    if (args.length > 0) {
      f.setTitle(progname + " [" + args[0] + "]");      
      sw.setSpectrum(Spectrum.readFile(args[0]), true);
      fileObservable.notifyObservers();
    }
    JMenuBar mb = new JMenuBar();
    f.setJMenuBar(mb);
    { // The File menu
      JMenu fileMenu = new JMenu("File");
      // File | Open
      JMenuItem open = new JMenuItem("Open file");
      open.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, 
						    ActionEvent.CTRL_MASK));
      fileMenu.add(open);
      open.addActionListener(new ActionListener() {
	  JFileChooser fd;
	  String fs = System.getProperty("file.separator");
	  public void actionPerformed(ActionEvent e) {
	    if (fd == null) {
	      fd = new JFileChooser("."); 
	      fd.setDialogTitle("Open file");
	      fd.setFileFilter(new javax.swing.filechooser.FileFilter () {
		  public boolean accept(File fil) {
		    return 
		      fil.isDirectory()
		      || fil.getName().endsWith(".txt") 
		      || fil.getName().endsWith(".dta");
		  }
		  public String getDescription() {
		    return "*txt, *dta";
		  }
		});
	    }
	    if(fd.showOpenDialog(f) == JFileChooser.APPROVE_OPTION) {
	      String filename = fd.getSelectedFile().getName();
	      String ffname = fd.getCurrentDirectory() +fs+ filename;
	      try {
		sw.setSpectrum(Spectrum.readFile(ffname), true);
		fileObservable.notifyObservers();
		f.setTitle(progname + " [" + filename + "]");
	      } 
	      catch (IOException exn) { 
		controller.debugmessage(exn.toString()); 
	      }
	    }
	  }
	});
      // File | Save log
      JMenuItem save = new JMenuItem("Save log");
      save.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, 
						    ActionEvent.CTRL_MASK));
      fileMenu.add(save);
      save.addActionListener(new ActionListener() {
	  JFileChooser fd;
	  public void actionPerformed(ActionEvent e) {
	    if (sw.getSpectrum() != null) {
	      if (fd==null) {
		fd = new JFileChooser("."); 
		//Should have been the current dir of the file that was
		//opened? Well, if the file is on a CD it might be
		//reasonable to get the directory from where we started
		//the program.
		fd.setDialogTitle("Save log");
	      }
	      fd.setSelectedFile(new File(sw.getSpectrum().name + ".log"));
	      if(fd.showSaveDialog(f) == JFileChooser.APPROVE_OPTION) {
		String filename = fd.getSelectedFile().getName();
		sw.getLog().writeFile(filename, controller);
	      }
	    }
	  }
	});
      // File | Print
      JMenuItem print = new JMenuItem(sw.getPrintAction());
      print.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_P, 
						  ActionEvent.CTRL_MASK));
      fileMenu.add(print);
      // File | Quit
      JMenuItem quit = new JMenuItem("Quit");
      quit.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, 
						    ActionEvent.CTRL_MASK));
      fileMenu.add(quit);
      quit.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent e) {
	    System.exit(0);
	  }
	});
      mb.add(fileMenu);
    }
    { // The edit menu
      JMenu editMenu = new JMenu("Edit");
      // Edit | Undo
      JMenuItem undoItem = new JMenuItem(controller.getUndoAction());
      editMenu.add(undoItem);
      undoItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Z, 
						     ActionEvent.CTRL_MASK));

      // Edit | Redo
      JMenuItem redoItem = new JMenuItem(controller.getRedoAction());
      editMenu.add(redoItem);
      redoItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Y, 
						    ActionEvent.CTRL_MASK));
      // Edit | Mirror
      JMenuItem mirrorSpec = new JMenuItem(sw.getMirrorAction());
      editMenu.add(mirrorSpec);
      mirrorSpec.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_M, 
						    ActionEvent.CTRL_MASK));
      mb.add(editMenu);
    }
    { // The Settings menu
      JMenu settingsMenu = new JMenu("Settings");
      // Settings | Show errors on accepted peaks
      final JCheckBoxMenuItem showErrors = 
	new JCheckBoxMenuItem("Show errors", false);
      settingsMenu.add(showErrors);
      showErrors.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_E, 
						       ActionEvent.CTRL_MASK));
      showErrors.addItemListener(new ItemListener() {
	  public void itemStateChanged(ItemEvent e) {
	    sw.setShowErrors(showErrors.isSelected());
	  }
	});
      // Settings | Show sequence backwards
      final JCheckBoxMenuItem bwdSequence = 
	new JCheckBoxMenuItem("Show sequence backwards", false);
      bwdSequence.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_B, 
						    ActionEvent.CTRL_MASK));
      settingsMenu.add(bwdSequence);
      bwdSequence.addItemListener(new ItemListener() {
	  public void itemStateChanged(ItemEvent e) {
	    sw.setBwdSequence(bwdSequence.isSelected());
	    propP.updateTexts();
	  }
	});
      // Settings | Show local levels
      final JCheckBoxMenuItem showLocalLevels = 
	new JCheckBoxMenuItem("Show local levels", false);
      settingsMenu.add(showLocalLevels);
      showLocalLevels.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_L, 
						    ActionEvent.CTRL_MASK));
      showLocalLevels.addItemListener(new ItemListener() {
	  public void itemStateChanged(ItemEvent e) {
	    sw.setShowLocalLevels(showLocalLevels.isSelected());
	  }
	});
      // Settings | LogScale
      final JCheckBoxMenuItem logScale = 
	new JCheckBoxMenuItem("Use log scale", false);
      logScale.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_W, 
						    ActionEvent.CTRL_MASK));
      settingsMenu.add(logScale);
      logScale.addItemListener(new ItemListener() {
	  public void itemStateChanged(ItemEvent e) {
	    sw.setLogScale(logScale.isSelected());
	  }
	});
      settingsMenu.addSeparator();
      // Settings | Show backwards spectrum
      final JCheckBoxMenuItem bwdSpectrum = 
	new JCheckBoxMenuItem("Show backwards spectrum", false);
      bwdSpectrum.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, 
						    ActionEvent.CTRL_MASK));
      settingsMenu.add(bwdSpectrum);
      bwdSpectrum.addItemListener(new ItemListener() {
	  public void itemStateChanged(ItemEvent e) {
	    sw.setBwdSpectrum(bwdSpectrum.isSelected());
	  }
	});
      // Settings | Extra
      JMenu extra = new JMenu("Show reinterpreted spectrum");
      final JRadioButtonMenuItem jrbmiNone;
      {
	Iterator iter = IonType.iterator();
	ButtonGroup extraBG = new ButtonGroup();
	ActionListener myListener = new ActionListener() {
	    public void actionPerformed(ActionEvent e) {
	      sw.setExtra(e.getActionCommand());
	    }
	  };
	jrbmiNone = new JRadioButtonMenuItem("None", true);
	jrbmiNone.addActionListener(myListener);
	extra.add(jrbmiNone);
	extraBG.add(jrbmiNone);
	while (iter.hasNext()) {
	  IonType iontype = (IonType)(iter.next());
	  if (iontype == IonType.Y)
	    continue;
	  JRadioButtonMenuItem jrbmi = new JRadioButtonMenuItem(iontype.abbr);
	  jrbmi.addActionListener(myListener);
	  extra.add(jrbmi);
	  extraBG.add(jrbmi);
	}
      }
      settingsMenu.add(extra);
      settingsMenu.addSeparator();
      // Settings | Allow K and R
      final JCheckBoxMenuItem allowKR =
	new JCheckBoxMenuItem("Allow K and R", false);
      settingsMenu.add(allowKR);
      controller.addSeqParmsObserver(new Observer() {
	  public void update(Observable o, Object arg) {
	    if (allowKR.getState() != controller.seqParms.allowKR)
	      allowKR.setSelected(controller.seqParms.allowKR);
	  }
	});
      
      allowKR.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_A, 
						    ActionEvent.CTRL_MASK));
      allowKR.addItemListener(new ItemListener() {
	  public void itemStateChanged(ItemEvent e) {
	    if (allowKR.isSelected() != controller.seqParms.allowKR)
	      sw.setAllowKR(allowKR.isSelected());
	  }
	});
      // Settings | Decrease tolerance
      JMenuItem decrTol = new JMenuItem("Decrease tolerance");
      settingsMenu.add(decrTol);
      decrTol.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_D, 
						    ActionEvent.CTRL_MASK));
      decrTol.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent e) {
	    sw.decrTolerance();
	  }
	});
      // Settings | Increase tolerance
      JMenuItem incrTol = new JMenuItem("Increase tolerance");
      settingsMenu.add(incrTol);
      incrTol.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_T, 
						    ActionEvent.CTRL_MASK));
      incrTol.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent e) {
	    sw.incrTolerance();
	  }
	});
      // Settings | DP data
      JMenuItem dpData = new JMenuItem("Sequencing parameters...");
      settingsMenu.add(dpData);
      dpData.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent e) {
	    //This should not be necessary or even to do each time:
	    dpparmsFrame.pack();
	    dpparmsFrame.setVisible(true);
	    dpparmsFrame.setState(Frame.NORMAL);
	  }
	});
      //
      mb.add(settingsMenu);
      
      fileObservable.addObserver(new Observer() {
	  public void update(Observable o, Object arg) {
	    jrbmiNone.setSelected(true);
	    bwdSpectrum.setSelected(false);
	    //logScale?, mirroring txt?
	  }
	});
    }

    /**
     * TextDialog
     */

    { // The Help menu
      JMenu helpMenu = new JMenu("Help");
      // Help | Help
      JMenuItem help = new JMenuItem("Help");
      help.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F1, 
						    ActionEvent.CTRL_MASK));
      helpMenu.add(help);
      help.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent e) {
	    JDialog dia = 
	      new TextDialog(f,progname + " Help", 8,
		 "<html><body bgcolor=white text=black>"
		 + "Keystroke commands:<br>"
		 + "z           zoom in around the cursor<br>"
		 + "o           zoom out around the cursor<br>"
		 + "c           recenter around the cursor<br>"
		 + "r           reset to full view of the spectrum<br>"
		 + "m         show masses of 10 more high peaks<br>"
		 + "n          hide masses of 10 more low peaks<br><br>"
		 + "Use arrows to scroll left, right</body></html>");
	  }});
      // Help | About
      JMenuItem about = new JMenuItem("About");
      helpMenu.add(about);
      about.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent e) {
	    String html = "<html><head>" +
	      "<style>td {font-size : 12pt;}</style>" +
	      "</head><body bgcolor=white text=black><tt>" 
	      + "KVL AminoSequencer version 1.0 of 2002-01-04<br>"
	      + "Peter Sestoft (sestoft@dina.kvl.dk)<br>"
	      + "Niels Jrgen Kokholm (kokholm@dina.kvl.dk)<br>"
	      + "See http://www.dina.kvl.dk/~sestoft/aminoseq/"
	      + "</tt></body></html>";
	    JDialog dia = 
	      new TextDialog(f,"About " + progname, 6, html);
	  }
	});
      // Help | Amino acids
      JMenuItem aa = new JMenuItem("Amino acids");
      helpMenu.add(aa);
      aa.addActionListener(new ActionListener() {
	  String list;
	  int h;
	  {
	    StringBuffer l = 
	      new StringBuffer("<html><head>" +
			       "<style>td {font-size : 12pt;}</style>" +
			       "</head><body bgcolor=white text=black><tt>" +
			       "<table border=0 cellspacing=1 cellpadding=0>"+ 
			       "<tr><th>Abbr&nbsp;<th halign=right>monoW");
	    Iterator iter = AminoAcid.aaByMonoW.iterator();
	    while (iter.hasNext()) {
	      AminoAcid aa = (AminoAcid)iter.next();
	      l.append("<tr><td>"+aa.abbr + "<td halign=right>" + aa.monoW);
	      h++;
	    }
	    l.append("</table></tt></body></html>");
	    list = l.toString();
	  };
	  public void actionPerformed(ActionEvent e) {
	    JDialog dia = new TextDialog(f,"Amino acids " + progname, h, list);
	  }
	});
      // Help | Ion Types
      JMenuItem it = new JMenuItem("Ion types");
      helpMenu.add(it);
      it.addActionListener(new ActionListener() {
	  String list;
	  int h;
	  {
	    StringBuffer l = 
	      new StringBuffer("<html><head>>" +
			       "<style>td {font-size : 12pt;}</style>" +
			       "</head><body bgcolor=white text=black><tt>" +
			       "<table border=0 cellspacing=1 cellpadding=0>" +
			       "<tr><th>Name<th>&nbsp;<th halign=left>m/z");
	    Iterator iter = IonType.iterator();
	    while (iter.hasNext()) {
	      IonType iontype = (IonType)iter.next();
	      l.append("<tr><td>"+iontype.abbr + "<td><td>" + iontype.txt);
	      h++;
	    }
	    l.append("</table></tt></body></html>");
	    list = l.toString();
	  };
	  public void actionPerformed(ActionEvent e) {
	    JDialog dia = new TextDialog(f,"Ion types " + progname, h, list);
	  }
	});
      // Help | About Spectrum
      JMenuItem as = new JMenuItem("About spectrum");
      helpMenu.add(as);
      as.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent e) {
	    JDialog dia = new TextDialog(f,"About spectrum " + progname, 1, 
					 controller.getSpectrumDescription());
	  }
	});
       mb.add(helpMenu);
    }

    Container contentPane = f.getContentPane();
    contentPane.setLayout(new BorderLayout());
    contentPane.add(msgField, BorderLayout.NORTH);
    contentPane.add(sw, BorderLayout.CENTER);
    //
    Box boxW = new Box(BoxLayout.Y_AXIS);
    boxW.add(Box.createVerticalStrut(5));
    JPanel mousePositionPanel = new JPanel();
    mousePositionPanel.setLayout(new BoxLayout(mousePositionPanel, 
					       BoxLayout.X_AXIS));
    mousePositionPanel.add(new JLabel("x = "));
    mousePositionPanel.add(xCoor);
    mousePositionPanel.add(Box.createHorizontalStrut(10));
    mousePositionPanel.add(new JLabel("y = "));
    mousePositionPanel.add(yCoor);
    mousePositionPanel.add(Box.createHorizontalStrut(10));
    mousePositionPanel.add(new JLabel("ion types = "));
    mousePositionPanel.add(ionTypeMatch);
    mousePositionPanel.add(Box.createGlue());
    javax.swing.border.Border b = BorderFactory.createEtchedBorder();
    b = BorderFactory.createTitledBorder(b,"Mouse position");
    mousePositionPanel.setBorder(b);
    /*
    */
    boxW.add(mousePositionPanel);
    //
    boxW.add(Box.createVerticalStrut(5));
    JPanel sequencesPanel = new JPanel();
    sequencesPanel.setLayout(new BoxLayout(sequencesPanel,BoxLayout.Y_AXIS));
    Box boxW2 = new Box(BoxLayout.X_AXIS);
    boxW2.add(seqOut);
    boxW2.add(Box.createHorizontalStrut(10));
    boxW2.add(new JLabel("delta = "));
    boxW2.add(mDelta);
    boxW2.add(Box.createGlue());
    sequencesPanel.add(boxW2);
    Box boxW1 = new Box(BoxLayout.X_AXIS);
    boxW1.add(new JLabel("\u0394 Left = "));
    boxW1.add(rDist);
    boxW1.add(Box.createHorizontalStrut(10));
    boxW1.add(new JLabel("\u0394 Right = "));
    boxW1.add(lDist);
    boxW1.add(Box.createGlue());
    sequencesPanel.add(boxW1);
    b = BorderFactory.createEtchedBorder();
    b = BorderFactory.createTitledBorder(b,"Sequences");
    sequencesPanel.setBorder(b);
    boxW.add(sequencesPanel); 
    //
    boxW.add(Box.createVerticalStrut(5));
    b = BorderFactory.createEtchedBorder();
    b = BorderFactory.createTitledBorder(b,"Proposals");
    propP.setBorder(b);
    boxW.add(propP);
    //
    contentPane.add(boxW, BorderLayout.SOUTH);
    //
    f.pack(); f.setVisible(true);
  }


}

    class TextDialog extends JDialog {
      public TextDialog(JFrame f, String title, int lines, String text) { 
	super(f, title); 
	//	JTextArea t = new JTextArea(text);//, lines, 40);
	//t.setEditable(false);
	JLabel t = new JLabel(text);
	getContentPane().add(t, BorderLayout.CENTER);
	JButton ok = new JButton("OK");
	JPanel pS = new JPanel();
	pS.add(ok);
	ok.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) {
	      hide();
	    }
	  });
	getContentPane().add(pS, BorderLayout.SOUTH);
	pack(); setVisible(true);	      
      }
    }
